@@ -27,14 +27,13 @@
 #	     and parse_ prepended.
 #	16 - Password free text
 
-# generate_config(&config, info-file, [module], [&can-config], [checkbox-name])
+# generate_config(&config, info-file, [module], [&can-config], [checkbox-name],
+#		  [only-section])
 # Prints HTML for 
 sub generate_config
 {
-local %config = %{$_[0]};
-local $file = $_[1];
-local $canconfig = $_[3];
-local $cbox = $_[4];
+local ($configref, $file, $module, $canconfig, $cbox, $section) = @_;
+local %config = %$configref;
 
 # Read the .info file in the right language
 local (%info, @info_order, %einfo, $o);
@@ -45,6 +44,11 @@ foreach $o (@lang_order_list) {
 	}
 @info_order = &unique(@info_order);
 
+if ($section) {
+	# Limit to settings in one section
+	@info_order = &config_in_section($section, \@info_order, \%info);
+	}
+
 # Show the parameter editors
 local $c;
 foreach $c (@info_order) {
@@ -63,9 +67,9 @@ foreach $c (@info_order) {
 		push(@p, @ep[scalar(@p) .. @ep-1]);
 		}
 	if ($p[1] == 14) {
-		$_[2] || &error($text{'config_ewebmin'});
-		&foreign_require($_[2], "config_info.pl");
-		local @newp = &foreign_call($_[2], $p[2], @p);
+		$module || &error($text{'config_ewebmin'});
+		&foreign_require($module, "config_info.pl");
+		local @newp = &foreign_call($module, $p[2], @p);
 		$newp[0] ||= $p[0];
 		@p = @newp;
 		}
@@ -74,13 +78,13 @@ foreach $c (@info_order) {
 		print "\t<b>" . $p[0] . "</b>\n</td></tr>\n" ;
 		next;
 		}
-	if ($p[1] == 16 && $config{'config_16_insecure'}) {
+	if ($p[1] == 16 && $gconfig{'config_16_insecure'}) {
 		# Don't allow mode 16
 		$p[1] = 12;
 		}
-	if ($_[2] && -r &help_file($_[2], "config_$c")) {
+	if ($module && -r &help_file($module, "config_$c")) {
 		print "<tr> $checkhtml <td valign=top>",
-		      &hlink("<b>$p[0]</b>", "config_$c", $_[2]),"</td>\n";
+		      &hlink("<b>$p[0]</b>", "config_$c", $module),"</td>\n";
 		}
 	else {
 		print "<tr> $checkhtml <td valign=top><b>$p[0]</b></td>\n";
@@ -142,6 +146,7 @@ foreach $c (@info_order) {
 		print "</select>\n";
 		}
 	elsif ($p[1] == 5) {
+		# User chooser
 		if ($p[2]) {
 			printf
 			   "<input type=radio name=${c}_def value=1 %s>$p[2]\n",
@@ -149,9 +154,16 @@ foreach $c (@info_order) {
 			printf "<input type=radio name=${c}_def value=0 %s>\n",
 			   $config{$c} eq "" ? "" : "checked";
 			}
-		print &unix_user_input($c, $config{$c});
+		if ($p[3]) {
+			print "<input name=$c size=30 value=\"$config{$c}\"> ".
+			      &user_chooser_button($c, 1);
+			}
+		else {
+			print &unix_user_input($c, $config{$c});
+			}
 		}
 	elsif ($p[1] == 6) {
+		# Group chooser
 		if ($p[2]) {
 			printf
 			   "<input type=radio name=${c}_def value=1 %s>$p[2]\n",
@@ -159,7 +171,13 @@ foreach $c (@info_order) {
 			printf "<input type=radio name=${c}_def value=0 %s>\n",
 			   $config{$c} eq "" ? "" : "checked";
 			}
-		print &unix_group_input($c, $config{$c});
+		if ($p[3]) {
+			print "<input name=$c size=30 value=\"$config{$c}\"> ".
+			      &group_chooser_button($c, 1);
+			}
+		else {
+			print &unix_group_input($c, $config{$c});
+			}
 		}
 	elsif ($p[1] == 7) {
 		print "<input name=\"$c\" size=40 value=\"",
@@ -229,9 +247,9 @@ foreach $c (@info_order) {
 		print "</select>\n";
 		}
 	elsif ($p[1] == 15) {
-		$_[2] || &error($text{'config_ewebmin'});
-		&foreign_require($_[2], "config_info.pl");
-		local $html = &foreign_call($_[2], "show_".$p[2],
+		$module || &error($text{'config_ewebmin'});
+		&foreign_require($module, "config_info.pl");
+		local $html = &foreign_call($module, "show_".$p[2],
 					    $config{$c}, @p);
 		print $html;
 		}
@@ -245,11 +263,11 @@ foreach $c (@info_order) {
 	}
 }
 
-# parse_config(&config, info-file, [module], [&canconfig])
+# parse_config(&config, info-file, [module], [&canconfig], [section])
 # Updates the specified configuration with values from %in
 sub parse_config
 {
-local ($config, $file, $module, $canconfig) = @_;
+local ($config, $file, $module, $canconfig, $section) = @_;
 
 # Read the .info file
 local (%info, @info_order, $o);
@@ -259,6 +277,11 @@ foreach $o (@lang_order_list) {
 	}
 @info_order = &unique(@info_order);
 
+if ($section) {
+	# Limit to settings in one section
+	@info_order = &config_in_section($section, \@info_order, \%info);
+	}
+
 # Actually parse the inputs
 local $c;
 foreach $c (@info_order) {
@@ -271,7 +294,7 @@ foreach $c (@info_order) {
 		$newp[0] ||= $p[0];
 		@p = @newp;
 		}
-	if ($p[1] == 16 && $config{'config_16_insecure'}) {
+	if ($p[1] == 16 && $gconfig{'config_16_insecure'}) {
 		# Don't allow mode 16
 		$p[1] = 12;
 		}
@@ -327,11 +350,35 @@ foreach $c (@info_order) {
 		local $pkg = $_[2];
 		$pkg =~ s/[^A-Za-z0-9]/_/g;
 		eval "\%${pkg}::in = \%in";
-		$config{$c} = &foreign_call($_[2], "parse_".$p[2],
-					    $config{$c}, @p);
+		$config->{$c} = &foreign_call($_[2], "parse_".$p[2],
+					    $config->{$c}, @p);
 		}
 	}
 }
 
+# config_in_section(&section, &order, &config-info)
+# Returns a list of config names that are in some section
+sub config_in_section
+{
+local ($section, $info_order, $info) = @_;
+my @new_order = ( );
+my $in_section = 0;
+foreach my $c (@$info_order) {
+	local @p = split(/,/, $info->{$c});
+	if ($p[1] == 11 && $c eq $section) {
+		$in_section = 1;
+		}
+	elsif ($p[1] == 11 && $c ne $section) {
+		$in_section = 0;
+		}
+	elsif ($in_section) {
+		push(@new_order, $c);
+		}
+	}
+return @new_order;
+}
+
+
+
 1;
 
