@@ -622,7 +622,7 @@ else {
 sub show_times_input
 {
 local $job = $_[0];
-if ($config{'vixie_cron'} && !$_[1]) {
+if ($config{'vixie_cron'} && (!$_[1] || $_[0]->{'special'})) {
 	# Allow selection of special @ times
 	print "<tr $cb> <td colspan=6>\n";
 	printf "<input type=radio name=special_def value=1 %s> %s\n",
@@ -1039,6 +1039,30 @@ if ($job->{'start'}) {
 	delete($job->{'start'});
 	delete($job->{'end'});
 	&create_wrapper($range_cmd, $module_name, "range.pl");
+	&set_ownership_permissions(undef, undef, 0755, $range_cmd);
+	return 1;
+	}
+return 0;
+}
+
+# convert_comment(&job)
+# Given a cron job with a # command after the command, sets the comment field
+sub convert_comment
+{
+if ($job->{'command'} =~ /^(.*)\s*#([^#]*)$/) {
+	$job->{'command'} = $1;
+	$job->{'comment'} = $2;
+	return 1;
+	}
+return 0;
+}
+
+# unconvert_comment(&job)
+# Adds an comment back to the command in a cron job
+sub unconvert_comment
+{
+if ($job->{'comment'} =~ /\S/) {
+	$job->{'command'} .= " #".$job->{'comment'};
 	return 1;
 	}
 return 0;
