@@ -260,34 +260,34 @@ return $out =~ /fetchmail\s+release\s+(\S+)/ ? $1 : undef;
 # show_polls(&polls, file, user)
 sub show_polls
 {
-print "<table border width=100%>\n";
-print "<tr $tb> <td><b>$text{'index_poll'}</b></td> ",
-      "<td><b>$text{'index_active'}</b></td> ",
-      "<td><b>$text{'index_proto'}</b></td> ",
-      "<td><b>$text{'index_users'}</b></td> </tr>\n" if (@{$_[0]});
-foreach $p (@{$_[0]}) {
-	print "<tr $cb>\n";
-	print "<td valign=top><a href='edit_poll.cgi?file=$_[1]&",
-	      "idx=$p->{'index'}&user=$_[2]'>",&html_escape($p->{'poll'}),
-	      "</a></td>\n";
-	print "<td>",$p->{'skip'} ? "<font color=#ff0000>$text{'no'}</font>"
-                                  : $text{'yes'},"</td>\n";
-	print "<td>",$p->{'proto'} ? &html_escape(uc($p->{'proto'}))
-				   : $text{'default'},"</td>\n";
-	print "<td>\n";
-	foreach $u (@{$p->{'users'}}) {
-		printf "%s -> %s<br>\n",
-			&html_escape($u->{'user'}),
-			&html_escape(@{$u->{'is'}} ? join(" ", @{$u->{'is'}})
-						   : $_[2]);
+if (@{$_[0]}) {
+	print &ui_columns_start([ $text{'index_poll'},
+				  $text{'index_active'},
+				  $text{'index_proto'},
+				  $text{'index_users'} ], 100);
+	foreach $p (@{$_[0]}) {
+		local @cols;
+		push(@cols, "<a href='edit_poll.cgi?file=$_[1]&".
+		      	    "idx=$p->{'index'}&user=$_[2]'>".
+			    &html_escape($p->{'poll'})."</a>");
+		push(@cols, $p->{'skip'} ?
+		    "<font color=#ff0000>$text{'no'}</font>" : $text{'yes'});
+		push(@cols, $p->{'proto'} ? &html_escape(uc($p->{'proto'}))
+					  : $text{'default'});
+		local $ulist;
+		foreach $u (@{$p->{'users'}}) {
+			$ulist .= sprintf "%s -> %s<br>\n",
+				&html_escape($u->{'user'}),
+				&html_escape(@{$u->{'is'}} ?
+				   join(" ", @{$u->{'is'}}) : $_[2]);
+			}
+		push(@cols, $ulist);
+		print &ui_columns_row(\@cols);
 		}
-	print "</td> </tr>\n";
+	print &ui_columns_end();
 	}
-print "</table>\n";
 print "<a href='edit_poll.cgi?new=1&file=$_[1]&user=$_[2]'>$text{'index_add'}</a>\n";
-print "&nbsp;&nbsp;\n";
 print "<a href='edit_global.cgi?file=$_[1]&user=$_[2]'>$text{'index_global'}</a>\n";
-print "&nbsp;&nbsp;\n";
 if (@{$_[0]}) {
 	print "<a href='check.cgi?file=$_[1]&user=$_[2]'>$text{'index_run'}</a>\n";
 	}
