@@ -261,6 +261,11 @@ elsif ($_[0] =~ /^:include:(.*)$/ && $config{'mail_system'} == 0) {
 elsif ($_[0] =~ /^\\($remote_user)$/ && $config{'mail_system'} == 0) {
 	@rv = (7, undef);
 	}
+elsif ($remote_user =~ /^(\S+)\@(\S+)$/ &&
+       $_[0] eq "\\$1-$2" && $config{'mail_system'} == 0) {
+	# Virtualmin created user for Postfix when username contains @
+	@rv = (7, undef);
+	}
 elsif ($_[0] =~ /^&(.*)$/) {
 	@rv = (1, $1);
 	}
@@ -350,27 +355,8 @@ foreach my $a (@aliases) {
 			# Usermin autoreply program
 			return undef if ($simple->{'autoreply'});
 			$simple->{'autoreply'} = &make_absolute($aval);
-			local @lines;
-			open(FILE, $simple->{'autoreply'});
-			while(<FILE>) {
-				if (/^Reply-Tracking:\s*(.*)/) {
-					$simple->{'replies'} = $1;
-					}
-				elsif (/^Reply-Period:\s*(.*)/) {
-					$simple->{'period'} = $1;
-					}
-				elsif (/^No-Autoreply:\s*(.*)/) {
-					$simple->{'no_autoreply'} = $1;
-					}
-				elsif (/^From:\s*(.*)/) {
-					$simple->{'from'} = $1;
-					}
-				else {
-					push(@lines, $_);
-					}
-				}
-			close(FILE);
-			$simple->{'autotext'} = join("", @lines);
+			$simple->{'auto'} = 1;
+			&read_autoreply($simple->{'autoreply'}, $simple);
 			}
 		else {
 			# Some un-supported rule
@@ -379,15 +365,53 @@ foreach my $a (@aliases) {
 		}
 	}
 $simple ||= { 'local' => 1 };	# if no settings, assume local delivery
+if (!$simple->{'autoreply'}) {
+	# Get autoreply message from default file
+	$simple->{'autoreply'} = &make_absolute("autoreply.txt");
+	&read_autoreply($simple->{'autoreply'}, $simple);
+	}
 return $simple;
 }
 
+# read_autoreply(file, &simple)
+sub read_autoreply
+{
+local ($file, $simple) = @_;
+local @lines;
+open(FILE, $file);
+while(<FILE>) {
+	if (/^Reply-Tracking:\s*(.*)/) {
+		$simple->{'replies'} = $1;
+		}
+	elsif (/^Reply-Period:\s*(.*)/) {
+		$simple->{'period'} = $1;
+		}
+	elsif (/^No-Autoreply:\s*(.*)/) {
+		$simple->{'no_autoreply'} = $1;
+		}
+	elsif (/^No-Autoreply-Regexp:\s*(.*)/) {
+		push(@{$simple->{'no_autoreply_regexp'}}, $1);
+		}
+	elsif (/^Autoreply-File:\s*(.*)/) {
+		push(@{$simple->{'autoreply_file'}}, $1);
+		}
+	elsif (/^From:\s*(.*)/) {
+		$simple->{'from'} = $1;
+		}
+	else {
+		push(@lines, $_);
+		}
+	}
+close(FILE);
+$simple->{'autotext'} = join("", @lines);
+}
+
 # save_simple(&simple)
 # Creates a .forward file with the given simple settings
 sub save_simple
 {
 local ($simple) = @_;
-if ($simple->{'local'} && !$simple->{'forward'} && !$simple->{'autotext'}) {
+if ($simple->{'local'} && !$simple->{'forward'} && !$simple->{'auto'}) {
 	# If doing only local delivery, just remove .forward
 	&unlink_file($forward_file);
 	}
@@ -395,41 +419,59 @@ else {
 	# Need to create .forward
 	&open_tempfile(FORWARD, ">$forward_file");
 	if ($simple->{'local'}) {
-		&print_tempfile(FORWARD, "\\$remote_user\n");
+		if ($remote_user =~ /^(\S+)\@(\S+)$/ &&
+		    defined(getpwnam("$1-$2"))) {
+			# Handle users created by Virtualmin for Postfix
+			# with @ in names
+			&print_tempfile(FORWARD, "\\$1-$2\n");
+			}
+		else {
+			&print_tempfile(FORWARD, "\\$remote_user\n");
+			}
 		}
 	if ($simple->{'forward'}) {
 		&print_tempfile(FORWARD, $simple->{'forward'},"\n");
 		}
-	if ($simple->{'autotext'}) {
-		if (!$simple->{'autoreply'}) {
-			# Create autoreply file
-			$simple->{'autoreply'} =
-				"$remote_user_info[7]/autoreply.txt";
-			}
-		&open_tempfile(AUTO, ">$simple->{'autoreply'}");
-		if ($simple->{'replies'}) {
-			&print_tempfile(AUTO,
-				"Reply-Tracking: $simple->{'replies'}\n");
-			}
-		if ($simple->{'period'}) {
-			&print_tempfile(AUTO,
-				"Reply-Period: $simple->{'period'}\n");
-			}
-		if ($simple->{'no_autoreply'}) {
-			&print_tempfile(AUTO,
-				"No-Autoreply: $simple->{'no_autoreply'}\n");
-			}
-		if ($simple->{'from'}) {
-			&print_tempfile(AUTO, "From: $simple->{'from'}\n");
-			}
-		&print_tempfile(AUTO, $simple->{'autotext'});
-		&close_tempfile(AUTO);
-		local $afile = &make_relative($simple->{'autoreply'});
+	if ($simple->{'auto'}) {
+		local $afile = $simple->{'autoreply'};
 		&print_tempfile(FORWARD, "\"|$module_config_directory/autoreply.pl $afile $remote_user\"\n");
 		}
 	&close_tempfile(FORWARD);
 	&set_forward_perms();
 	}
+
+if ($simple->{'autotext'}) {
+	# Save autoreply text
+	if (!$simple->{'autoreply'}) {
+		# Create autoreply file
+		$simple->{'autoreply'} =
+			"$remote_user_info[7]/autoreply.txt";
+		}
+	&open_tempfile(AUTO, ">$simple->{'autoreply'}");
+	if ($simple->{'replies'}) {
+		&print_tempfile(AUTO,
+			"Reply-Tracking: $simple->{'replies'}\n");
+		}
+	if ($simple->{'period'}) {
+		&print_tempfile(AUTO,
+			"Reply-Period: $simple->{'period'}\n");
+		}
+	if ($simple->{'no_autoreply'}) {
+		&print_tempfile(AUTO,
+			"No-Autoreply: $simple->{'no_autoreply'}\n");
+		}
+	foreach my $r (@{$simple->{'no_autoreply_regexp'}}) {
+		&print_tempfile(AUTO, "No-Autoreply-Regexp: $r\n");
+		}
+	foreach my $f (@{$simple->{'autoreply_file'}}) {
+		&print_tempfile(AUTO, "Autoreply-File: $f\n");
+		}
+	if ($simple->{'from'}) {
+		&print_tempfile(AUTO, "From: $simple->{'from'}\n");
+		}
+	&print_tempfile(AUTO, $simple->{'autotext'});
+	&close_tempfile(AUTO);
+	}
 }
 
 1;
