@@ -25,8 +25,8 @@ while(<GPG>) {
 		local $k = { 'size' => $1,
 			     'key' => $2,
 			     'date' => $3,
-			     'name' => [ $4 ],
-			     'email' => [ $5 ],
+			     'name' => $4 ? [ $4 ] : [ ],
+			     'email' => $5 ? [ $5 ] : $4 ? [ "" ] : [ ],
 			     'index' => scalar(@rv) };
 		$kmap{$k->{'key'}} = $k;
 		while(1) {
@@ -91,9 +91,9 @@ return $pass;
 # put_passphrase(pass, &key)
 sub put_passphrase
 {
-open(PASS, ">$user_module_config_directory/pass.$_[1]->{'key'}");
-print PASS $_[0],"\n";
-close(PASS);
+&open_tempfile(PASS, ">$user_module_config_directory/pass.$_[1]->{'key'}");
+&print_tempfile(PASS, $_[0],"\n");
+&close_tempfile(PASS);
 chmod(0700, "$user_module_config_directory/pass.$_[1]->{'key'}");
 }
 
@@ -102,11 +102,11 @@ chmod(0700, "$user_module_config_directory/pass.$_[1]->{'key'}");
 # returns an error message or undef on failure
 sub encrypt_data
 {
-local $srcfile = &tempname();
+local $srcfile = &transname();
 local @keys = ref($_[2]) eq 'ARRAY' ? @{$_[2]} : ( $_[2] );
 local $rcpt = join(" ", map { "--recipient \"$_->{'name'}->[0]\"" } @keys);
 &write_entire_file($srcfile, $_[0]);
-local $dstfile = &tempname();
+local $dstfile = &transname();
 local $ascii = $_[3] ? "--armor" : "";
 local $comp = $config{'compress'} eq '' ? "" :
 		" --compress-algo $config{'compress'}";
@@ -139,9 +139,9 @@ else {
 # into &result. Returns an error message or undef on success.
 sub decrypt_data
 {
-local $srcfile = &tempname();
+local $srcfile = &transname();
 &write_entire_file($srcfile, $_[0]);
-local $dstfile = &tempname();
+local $dstfile = &transname();
 local $cmd = "$gpgpath --output $dstfile --decrypt $srcfile";
 local ($fh, $fpid) = &foreign_call("proc", "pty_process_exec", $cmd);
 local ($error, $seen_pass, $pass, $key, $keyid);
@@ -194,9 +194,9 @@ else {
 # mode 1 = ascii signature at end, mode 2 = ascii signature only
 sub sign_data
 {
-local $srcfile = &tempname();
+local $srcfile = &transname();
 &write_entire_file($srcfile, $_[0]);
-local $dstfile = &tempname();
+local $dstfile = &transname();
 local $cmd;
 if ($_[3] == 0) {
 	$cmd = "$gpgpath --output $dstfile --default-key $_[2]->{'key'} --sign $srcfile";
@@ -251,7 +251,7 @@ else {
 # code 4 = verification totally failed, message contains reason
 sub verify_data
 {
-local $datafile = &tempname();
+local $datafile = &transname();
 &write_entire_file($datafile, $_[0]);
 local $cmd;
 local $sigfile;
@@ -259,7 +259,7 @@ if (!$_[1]) {
 	$cmd = "$gpgpath --verify $datafile";
 	}
 else {
-	$sigfile = &tempname();
+	$sigfile = &transname();
 	&write_entire_file($sigfile, $_[1]);
 	$cmd = "$gpgpath --verify $sigfile $datafile";
 	}
@@ -305,9 +305,9 @@ return $rv;
 # write_entire_file(file, data)
 sub write_entire_file
 {
-open(FILE, ">$_[0]");
-print FILE $_[1];
-close(FILE);
+&open_tempfile(FILE, ">$_[0]");
+&print_tempfile(FILE, $_[1]);
+&close_tempfile(FILE);
 }
 
 # get_trust_level(&key)
@@ -330,5 +330,28 @@ close($fh);
 return $tr;
 }
 
+# delete_key(&key)
+# Delete one public or secret key
+sub delete_key
+{
+local ($key) = @_;
+if ($key->{'secret'}) {
+	local $cmd = "$gpgpath --delete-secret-key \"$key->{'name'}->[0]\"";
+	local ($fh, $fpid) = &foreign_call("proc", "pty_process_exec", $cmd);
+	&wait_for($fh, "\\?");
+	syswrite($fh, "y\n");
+	&wait_for($fh, "\\?");
+	syswrite($fh, "y\n");
+	sleep(1);
+	close($fh);
+	}
+local $cmd = "$gpgpath --delete-key \"$key->{'name'}->[0]\"";
+local ($fh, $fpid) = &foreign_call("proc", "pty_process_exec", $cmd);
+&wait_for($fh, "\\?");
+syswrite($fh, "y\n");
+sleep(1);
+close($fh);
+}
+
 1;
 
