@@ -6,7 +6,7 @@ $directive_type_count = 20;
 if ($module_name ne 'htaccess') {
 	do '../web-lib.pl';
 	&init_config();
-	require '../ui-lib.pl';
+	do '../ui-lib.pl';
 	%access = &get_module_acl();
 	@access_types = $access{'types'} eq '*' ? (0 .. $directive_type_count)
 					: split(/\s+/, $access{'types'});
@@ -17,19 +17,25 @@ else {
 map { $access_types{$_}++ } @access_types;
 $site_file = ($config{'webmin_apache'} || $module_config_directory)."/site";
 
-if (!-r $site_file && $module_name ne 'htaccess') {
-	# Need to build list of supported modules
+# Check if a list of supported modules needs to be built
+if ($module_name ne 'htaccess') {
+	local %oldsite;
 	local $httpd = &find_httpd();
-	local ($ver, $mods) = &httpd_info($httpd);
-	if ($ver) {
-		local @mods = map { "$_/$ver" } &configurable_modules();
-		local @st = stat($httpd);
-		local %site = ( 'size' => $st[7],
-				'modules' => join(' ', @mods) );
-		&lock_file($site_file);
-		&write_file($site_file, \%site);
-		chmod(0644, $site_file);
-		&unlock_file($site_file);
+	local @st = stat($httpd);
+	&read_file($site_file, \%oldsite);
+	if ($oldsite{'path'} ne $httpd || $oldsite{'size'} != $st[7]) {
+		# Need to build list of supported modules
+		local ($ver, $mods) = &httpd_info($httpd);
+		if ($ver) {
+			local @mods = map { "$_/$ver" } &configurable_modules();
+			local %site = ( 'size' => $st[7],
+					'path' => $httpd,
+					'modules' => join(' ', @mods) );
+			&lock_file($site_file);
+			&write_file($site_file, \%site);
+			chmod(0644, $site_file);
+			&unlock_file($site_file);
+			}
 		}
 	}
 
@@ -56,9 +62,6 @@ if (&read_file($site_file, \%site)) {
 		$httpd_defines{$d}++;
 		}
 	}
-foreach $d (split(/\s+/, $config{'defines'})) {
-	$httpd_defines{$d}++;
-	}
 
 $apache_docbase = $config{'apache_docbase'} ? $config{'apache_docbase'} :
 		  $httpd_modules{'core'} >= 2.0 ?
@@ -85,6 +88,7 @@ $dummy{'file'} = $_[2];
 $dummy{'type'} = 0;
 $dummy{'name'} = "dummy";
 @rv = (\%dummy);
+local %defs = map { $_, 1 } &get_httpd_defines();
 while($line = <$fh>) {
 	chop;
 	$line =~ s/^\s*#.*$//g;
@@ -94,15 +98,20 @@ while($line = <$fh>) {
 		$_[1]++;
 		last if ($_[3]);
 		}
-	elsif ($line =~ /^\s*<IfModule\s+(\!?)(\S+)\.c>/i) {
+	elsif ($line =~ /^\s*<IfModule\s+(\!?)(\S+)\.c>/i ||
+	       $line =~ /^\s*<IfModule\s+(\!?)(\S+)>/i) {
 		# start of an IfModule block. Read it, and if the module
 		# exists put the directives in this section.
 		local ($not, $mod) = ($1, $2);
 		local $oldline = $_[1];
 		$_[1]++;
 		local @dirs = &parse_config_file($fh, $_[1], $_[2], 1);
+		local $altmod = $mod;
+		$altmod =~ s/^(\S+)_module$/mod_$1/g;
 		if (!$not && $httpd_modules{$mod} ||
-		    $not && !$httpd_modules{$mod}) {
+		    $not && !$httpd_modules{$mod} ||
+		    !$not && $httpd_modules{$altmod} ||
+		    $not && !$httpd_modules{$altmod}) {
 			# use the directives..
 			push(@rv, { 'line', $oldline,
 				    'eline', $oldline,
@@ -122,8 +131,8 @@ while($line = <$fh>) {
 		local $oldline = $_[1];
 		$_[1]++;
 		local @dirs = &parse_config_file($fh, $_[1], $_[2], 1);
-		if (!$not && $httpd_defines{$def} ||
-		    $not && !$httpd_defines{$def}) {
+		if (!$not && $defs{$def} ||
+		    $not && !$defs{$def}) {
 			# use the directives..
 			push(@rv, { 'line', $oldline,
 				    'eline', $oldline,
@@ -136,6 +145,54 @@ while($line = <$fh>) {
 				    'name', "</IfDefine>" });
 			}
 		}
+	elsif ($line =~ /^\s*<IfVersion\s+(\!?)(\S*)\s*(\S+)>/i) {
+		# Start of an IfVersion block. Read it, and if the version
+		# matches put the directives in this section
+		local ($not, $op, $ver) = ($1, $2, $3);
+		local $oldline = $_[1];
+		$_[1]++;
+		local @dirs = &parse_config_file($fh, $_[1], $_[2], 1);
+		$op ||= "=";
+		local $match = 0;
+		local $myver = $httpd_modules{'core'};
+		$myver =~ s/^(\d+)\.(\d)(\d+)$/$1.$2.$3/;
+		if ($op eq "=" || $op eq "==") {
+			if ($ver =~ /^\/(.*)\/$/) {
+				$match = 1 if ($myver =~ /$1/);
+				}
+			else {
+				$match = 1 if ($myver eq $ver);
+				}
+			}
+		elsif ($op eq ">") {
+			$match = 1 if ($myver > $ver);
+			}
+		elsif ($op eq ">=") {
+			$match = 1 if ($myver >= $ver);
+			}
+		elsif ($op eq "<") {
+			$match = 1 if ($myver < $ver);
+			}
+		elsif ($op eq "<=") {
+			$match = 1 if ($myver <= $ver);
+			}
+		elsif ($op eq "~") {
+			$match = 1 if ($myver =~ /$ver/);
+			}
+		$match = !$match if ($not);
+		if ($match) {
+			# use the directives..
+			push(@rv, { 'line', $oldline,
+				    'eline', $oldline,
+				    'file', $_[2],
+				    'name', "<IfVersion $not$op $ver>" });
+			push(@rv, @dirs);
+			push(@rv, { 'line', $_[1]-1,
+				    'eline', $_[1]-1,
+				    'file', $_[2],
+				    'name', "</IfVersion>" });
+			}
+		}
 	elsif ($line =~ /^\s*<(\S+)\s*(.*)>/) {
 		# start of a container directive. The first member is a dummy
 		# directive at the same line as the container
@@ -287,10 +344,10 @@ foreach $v (@virt) {
 	foreach $idn ("ResourceConfig", "AccessConfig", "Include") {
 		local $inc = &find_directive_struct($idn, $mref);
 		next if (!$inc);
-		local $incdir = $inc->{'words'}->[0];
-		if ($incdir !~ /^\//)
-			{ $incdir = "$config{'httpd_dir'}/$incdir"; }
-		push(@$mref, &get_config_file($incdir));
+		local @incs = &expand_apache_include($inc->{'words'}->[0]);
+		foreach my $ginc (@incs) {
+			push(@$mref, &get_config_file($ginc));
+			}
 		}
 	}
 
@@ -321,23 +378,8 @@ else {
 
 # Expand Include directives
 foreach $inc (&find_directive_struct("Include", \@rv)) {
-	local $incdir = $inc->{'words'}->[0];
-	if ($incdir !~ /^\//) { $incdir = "$config{'httpd_dir'}/$incdir"; }
-	if ($incdir =~ /^(.*)\[\^([^\]]+)\](.*)$/) {
-		# A glob like /etc/[^.#]*.conf , which cannot be handled
-		# by Perl's glob function!
-		local $before = $1;
-		local $after = $3;
-		local %reject = map { $_, 1 } split(//, $2);
-		$reject{'*'} = $reject{'?'} = $reject{'['} = $reject{']'} =
-		  $reject{'/'} = $reject{'$'} = $reject{'('} = $reject{')'} =
-		  $reject{'!'} = 1;
-		local $accept = join("", grep { !$reject{$_} } map { chr($_) } (32 .. 126));
-		$incdir = $before."[".$accept."]".$after;
-		}
-	local @incs = glob($incdir);
-	local $ginc;
-	foreach $ginc (@incs) {
+	local @incs = &expand_apache_include($inc->{'words'}->[0]);
+	foreach my $ginc (@incs) {
 		push(@rv, &get_config_file($ginc));
 		}
 	}
@@ -345,6 +387,27 @@ foreach $inc (&find_directive_struct("Include", \@rv)) {
 return @rv;
 }
 
+# expand_apache_include(dir)
+# Given an include directive value, returns a list of matching files
+sub expand_apache_include
+{
+local ($incdir) = @_;
+if ($incdir !~ /^\//) { $incdir = "$config{'httpd_dir'}/$incdir"; }
+if ($incdir =~ /^(.*)\[\^([^\]]+)\](.*)$/) {
+	# A glob like /etc/[^.#]*.conf , which cannot be handled
+	# by Perl's glob function!
+	local $before = $1;
+	local $after = $3;
+	local %reject = map { $_, 1 } split(//, $2);
+	$reject{'*'} = $reject{'?'} = $reject{'['} = $reject{']'} =
+	  $reject{'/'} = $reject{'$'} = $reject{'('} = $reject{')'} =
+	  $reject{'!'} = 1;
+	local $accept = join("", grep { !$reject{$_} } map { chr($_) } (32 .. 126));
+	$incdir = $before."[".$accept."]".$after;
+	}
+return glob($incdir);
+}
+
 # get_virtual_config(index)
 sub get_virtual_config
 {
@@ -465,7 +528,8 @@ foreach $d (@{$_[0]}) {
 # server_root(path, &directives)
 sub server_root
 {
-if ($_[0] =~ /^\//) { return $_[0]; }
+if (!$_[0]) { return undef; }
+elsif ($_[0] =~ /^\//) { return $_[0]; }
 else { return "$config{'httpd_dir'}/$_[0]"; }
 }
 
@@ -533,6 +597,7 @@ foreach $m (keys %httpd_modules) {
 	}
 @rv = grep { $_->{'type'} == $_[0] && $_->{$_[1]} &&
 	     !$done{$_->{'name'}}++ } @rv;
+@rv = grep { &can_edit_directive($_->{'name'}) } @rv;
 @rv = sort { local $td = $a->{'subtype'} <=> $b->{'subtype'};
 	     local $pd = $b->{'priority'} <=> $a->{'priority'};
 	     local $md = $a->{'module'} cmp $b->{'module'};
@@ -541,6 +606,25 @@ foreach $m (keys %httpd_modules) {
 return @rv;
 }
 
+# can_edit_directive(name)
+# Returns 1 if the Apache directive named can be edited by the current user
+sub can_edit_directive
+{
+local ($name) = @_;
+if ($access{'dirsmode'} == 0) {
+	return 1;
+	}
+else {
+	local %dirs = map { lc($_), 1 } split(/\s+/, $access{'dirs'});
+	if ($access{'dirsmode'} == 1) {
+		return $dirs{lc($name)};
+		}
+	else {
+		return !$dirs{lc($name)};
+		}
+	}
+}
+
 # generate_inputs(&editors, &directives)
 # Displays a 2-column list of options, for use inside a table
 sub generate_inputs
@@ -732,6 +816,8 @@ foreach $h (&unique(@hl)) {
 	$rv .= sprintf "<option value=\"$h\" %s>$h\n",
 		$h eq $_[0] ? "selected" : "";
 	}
+$rv .= sprintf "<option value=\"None\" %s>&lt;$text{'core_none'}&gt;\n",
+	$_[0] eq "None" ? "selected" : "";
 $rv .= "</select>\n";
 return $rv;
 }
@@ -1080,7 +1166,7 @@ if ($httpd_modules{'core'} >= 1.301) {
 		if ($config{'httpd_conf'}) {
 			$cmd .= " -f \"$config{'httpd_conf'}\"";
 			}
-		foreach $d (keys %httpd_defines) {
+		foreach $d (&get_httpd_defines()) {
 			$cmd .= " -D$d";
 			}
 		}
@@ -1196,7 +1282,9 @@ if ($config{'apply_cmd'} eq 'restart') {
 	}
 elsif ($config{'apply_cmd'}) {
 	# Use the configured start command
+	&clean_environment();
 	local $out = &backquote_logged("$config{'apply_cmd'} 2>&1");
+	&reset_environment();
 	if ($?) {
 		return "<pre>$out</pre>";
 		}
@@ -1205,13 +1293,17 @@ elsif (-x &translate_filename($config{'apachectl_path'})) {
 	# Use apachectl to restart
 	if ($httpd_modules{'core'} >= 2) {
 		# Do a graceful restart
+		&clean_environment();
 		local $out = &backquote_logged("$config{'apachectl_path'} graceful 2>&1");
+		&reset_environment();
 		if ($?) {
 			return "<pre>$out</pre>";
 			}
 		}
 	else {
+		&clean_environment();
 		local $out = &backquote_logged("$config{'apachectl_path'} restart 2>&1");
+		&reset_environment();
 		if ($out !~ /httpd restarted/) {
 			return "<pre>$out</pre>";
 			}
@@ -1292,7 +1384,7 @@ else {
 	if ($config{'httpd_conf'}) {
 		$cmd .= " -f $config{'httpd_conf'}";
 		}
-	foreach $d (keys %httpd_defines) {
+	foreach $d (&get_httpd_defines()) {
 		$cmd .= " -D$d";
 		}
 	local $temp = &transname();
@@ -1380,5 +1472,73 @@ foreach $l (&find_directive_struct("LoadModule", $conf)) {
 return &unique(@rv);
 }
 
+# get_httpd_defines(automatic-only)
+# Returns a list of defines that need to be passed to Apache
+sub get_httpd_defines
+{
+local ($auto) = @_;
+local @rv;
+if (!$auto) {
+	push(@rv, keys %httpd_defines);
+	}
+if ($config{'defines_file'}) {
+	# Add defines from an environment file
+	local %def;
+	&read_env_file($config{'defines_file'}, \%def);
+	local $var = $def{$config{'defines_name'}};
+	foreach my $v (split(/\s+/, $var)) {
+		if ($v =~ /^-[Dd](\S+)$/) {
+			push(@rv, $1);
+			}
+		else {
+			push(@rv, $v);
+			}
+		}
+	}
+foreach my $md (split(/\t+/, $config{'defines_mods'})) {
+	# Add HAVE_ defines from modules
+	opendir(DIR, $md);
+	while(my $m = readdir(DIR)) {
+		if ($m =~ /^(mod_|lib)(.*).so$/i) {
+			push(@rv, "HAVE_".uc($2));
+			}
+		}
+	closedir(DIR);
+	}
+foreach my $d (split(/\s+/, $config{'defines'})) {
+	push(@rv, $d);
+	}
+return @rv;
+}
+
+# create_webfile_link(file)
+# Creates a link in the debian-style link directory for a new website's file
+sub create_webfile_link
+{
+local ($file) = @_;
+if ($config{'link_dir'}) {
+	local $short = $file;
+	$short =~ s/^.*\///;
+	local $linksrc = "$config{'link_dir'}/$short";
+	&lock_file($linksrc);
+	symlink($file, $linksrc);
+	&unlock_file($linksrc);
+	}
+}
+
+# delete_webfile_link(file)
+sub delete_webfile_link
+{
+local ($file) = @_;
+if ($config{'link_dir'}) {
+	local $short = $file;
+	$short =~ s/^.*\///;
+	local $linksrc = "$config{'link_dir'}/$short";
+	&lock_file($linksrc);
+	unlink($linksrc);
+	&unlock_file($linksrc);
+	}
+}
+
 1;
 
