@@ -869,24 +869,43 @@ sub edit_require
 {
 local($rv, $mode, $list);
 $_[0]->{'value'} =~ /^(\S+)\s*(.*)$/; $mode = $1; $list = $2;
+
+# All users
 $rv = sprintf
       "<input type=radio name=require_mode value=0 %s> $text{'default'}<br>\n",
       $mode ? "" : "checked";
+
+# Only some users
 $rv .= sprintf
       "<input type=radio name=require_mode value=1 %s> $text{'core_users'}:\n",
       $mode eq "user" ? "checked" : "";
 $rv .= sprintf
       "<input name=require_user size=20 value=\"%s\"><br>\n",
       $mode eq "user" ? $list : "";
+
+# Only members of groups
 $rv .= sprintf
       "<input type=radio name=require_mode value=2 %s> $text{'core_groups'}:\n",
       $mode eq "group" ? "checked" : "";
 $rv .= sprintf
       "<input name=require_group size=20 value=\"%s\"><br>\n",
       $mode eq "group" ? $list : "";
+
+# All users
 $rv .= sprintf
-      "<input type=radio name=require_mode value=3 %s> $text{'core_allusers'}:<br>\n",
+      "<input type=radio name=require_mode value=3 %s> $text{'core_allusers'}<br>\n",
       $mode eq "valid-user" ? "checked" : "";
+
+if ($httpd_modules{'mod_authz_owner'} >= 2.2) {
+	# File owner / group matches
+	$rv .= sprintf
+	      "<input type=radio name=require_mode value=4 %s> $text{'core_fileowner'}<br>\n",
+	      $mode eq "file-owner" ? "checked" : "";
+	$rv .= sprintf
+	      "<input type=radio name=require_mode value=5 %s> $text{'core_filegroup'}<br>\n",
+	      $mode eq "file-group" ? "checked" : "";
+
+	}
 return (1, $text{'core_authlog'}, $rv);
 }
 sub save_require
@@ -894,7 +913,10 @@ sub save_require
 if ($in{'require_mode'} == 0) { return ( [ ] ); }
 elsif ($in{'require_mode'} == 1) { return ( [ "user $in{'require_user'}" ] ); }
 elsif ($in{'require_mode'} == 2) { return ( [ "group $in{'require_group'}" ] ); }
-else { return ( [ "valid-user" ] ); }
+elsif ($in{'require_mode'} == 3) { return ( [ "valid-user" ] ); }
+elsif ($in{'require_mode'} == 4) { return ( [ "file-owner" ] ); }
+elsif ($in{'require_mode'} == 5) { return ( [ "file-group" ] ); }
+else { return ( [ ] ); }	# huh?
 }
 
 sub edit_Satisfy
