@@ -3,12 +3,17 @@
 # Install a single module file
 
 # Check arguments
+$nodeps = 0;
+if ($ARGV[0] eq "--nodeps") {
+	shift(@ARGV);
+	$nodeps = 1;
+	}
 while($ARGV[0] eq "--acl") {
 	shift(@ARGV);
 	push(@grant, shift(@ARGV));
 	}
 if (@ARGV > 2 || !@ARGV) {
-	die "usage: install-module.pl [--acl user]* <module.wbm> [config_directory]";
+	die "usage: install-module.pl [--nodeps] [--acl user]* <module.wbm> [config_directory]";
 	}
 $file = $ARGV[0];
 $config = $ARGV[1] ? $ARGV[1] : "/etc/webmin";
@@ -48,7 +53,7 @@ else {
 	$newusers = &webmin::get_newmodule_users();
 	$newusers ||= [ "root", "admin" ];
 	}
-$rv = &webmin::install_webmin_module($file, 0, 0, $newusers);
+$rv = &webmin::install_webmin_module($file, 0, $nodeps, $newusers);
 if (ref($rv)) {
 	for($i=0; $i<@{$rv->[0]}; $i++) {
 		printf "Installed %s in %s (%d kb)\n",
