@@ -2,8 +2,19 @@
 # Functions to parsing user mail files
 
 use POSIX;
+if ($userconfig{'date_tz'} || $config{'date_tz'}) {
+        # Set the timezone for all date calculations, and force a conversion
+        # now as in some cases the first on fails!
+        $ENV{'TZ'} = $userconfig{'date_tz'} ||
+                     $config{'date_tz'};
+        strftime('%H:%M', localtime(time()));
+        }
 use Time::Local;
 
+# Always use DBM indexing
+$config{'index_dbm'} = 2;
+$config{'index_min'} = 1000000;
+
 # list_mails(user|file, [start], [end])
 # Returns a subset of mail from a mbox format file
 sub list_mails
@@ -19,7 +30,7 @@ else {
 	}
 local ($start, $end);
 local $isize = $itype == 0 ? scalar(@index) : $index{'mailcount'};
-if (@_ == 1) {
+if (@_ == 1 || !defined($_[1]) && !defined($_[2])) {
 	$start = 0; $end = $isize-1;
 	}
 elsif ($_[2] < 0) {
@@ -810,7 +821,7 @@ chmod($st[2], $f);
 
 # send_mail(&mail, [file], [textonly], [nocr], [smtp-server],
 #	    [smtp-user], [smtp-pass], [smtp-auth-mode],
-#	    [&notify-flags])
+#	    [&notify-flags], [port])
 # Send out some email message or append it to a file.
 # Returns the number of lines written.
 sub send_mail
@@ -820,25 +831,26 @@ local (%header, $h);
 local $lnum = 0;
 local $sm = $_[4] || $config{'send_mode'};
 local $eol = $_[3] || !$sm ? "\n" : "\r\n";
+local $port = $_[9] || $config{'smtp_port'} || 25;
 foreach $h (@{$_[0]->{'headers'}}) {
 	$header{lc($h->[0])} = $h->[1];
 	}
+local @tm = localtime(time());
+push(@{$_[0]->{'headers'}},
+     [ 'Date', strftime("%a, %d %b %Y %H:%M:%S %z (%Z)", @tm) ])
+	if (!$header{'date'});
 local @from = &address_parts($header{'from'});
 local $esmtp = $_[8] ? 1 : 0;
 if ($_[1]) {
 	# Just append the email to a file using mbox format
-	local @tm = localtime(time());
 	open(MAIL, ">>$_[1]") || &error("Write failed : $!");
 	$lnum++;
 	print MAIL $_[0]->{'fromline'} ? $_[0]->{'fromline'}."\n" :
 		   strftime("From $from[0] %a %b %e %H:%M:%S %Y\n", @tm);
-	push(@{$_[0]->{'headers'}},
-	     [ 'Date', strftime("%a, %d %b %Y %H:%M:%S %Z", @tm) ])
-		if (!$header{'date'});
 	}
 elsif ($sm) {
 	# Connect to SMTP server
-	&open_socket($sm, 25, MAIL);
+	&open_socket($sm, $port, MAIL);
 	&smtp_command(MAIL);
 	if ($esmtp) {
 		&smtp_command(MAIL, "ehlo ".&get_system_hostname()."\r\n");
@@ -966,7 +978,7 @@ if (!$msg_id) {
 	}
 
 # Work out first attachment content type
-local $ftype;
+local ($ftype, $fenc);
 if (@{$_[0]->{'attach'}} >= 1) {
 	local $first = $_[0]->{'attach'}->[0];
 	$ftype = "text/plain";
@@ -974,6 +986,9 @@ if (@{$_[0]->{'attach'}} >= 1) {
 		if (lc($h->[0]) eq "content-type") {
 			$ftype = $h->[1];
 			}
+		if (lc($h->[0]) eq "content-transfer-encoding") {
+			$fenc = $h->[1];
+			}
 		}
 	}
 
@@ -988,7 +1003,8 @@ if (defined($_[0]->{'body'})) {
 	(print MAIL $_[0]->{'body'}) || &error("Write failed : $!");
 	$lnum += ($_[0]->{'body'} =~ tr/\n/\n/);
 	}
-elsif (!$_[2] || $ftype !~ /text\/plain/i) {
+elsif (!$_[2] || $ftype !~ /text\/plain/i ||
+       $fenc =~ /quoted-printable|base64/) {
 	# Sending MIME-encoded email
 	if ($ctype !~ /multipart\/report/i) {
 		$ctype =~ s/;.*$//;
@@ -1303,7 +1319,15 @@ return $data;
 
 sub simplify_date
 {
-if ($_[0] =~ /^(\S+),\s+0*(\d+)\s+(\S+)\s+(\d+)\s+(\d+):(\d+)/) {
+local $u = &parse_mail_date($_[0]);
+if ($u) {
+	local $fmt = $userconfig{'date_fmt'} || $config{'date_fmt'} || "dmy";
+	local $strf = $fmt eq "dmy" ? "%d/%m/%Y" :
+		      $fmt eq "mdy" ? "%m/%d/%Y" :
+				      "%Y/%m/%d";
+	return strftime("$strf %H:%M", localtime($u));
+        }
+elsif ($_[0] =~ /^(\S+),\s+0*(\d+)\s+(\S+)\s+(\d+)\s+(\d+):(\d+)/) {
 	return "$2/$3/$4 $5:$6";
 	}
 elsif ($_[0] =~ /^0*(\d+)\s+(\S+)\s+(\d+)\s+(\d+):(\d+)/) {
@@ -1476,14 +1500,27 @@ sub user_index_file
 {
 local $us = $_[0];
 $us =~ s/\//_/g;
-local $f = $_[0] =~ /^\/.*\/([^\/]+)$/ ?
-	($user_module_config_directory ?
-		"$user_module_config_directory/$1.findex" :
-		"$module_config_directory/$us.findex") :
-       $user_module_config_directory ?
-	"$user_module_config_directory/$_[0].index" :
-	"$module_config_directory/$_[0].index";
+local $f;
 local $hn = &get_system_hostname();
+if ($_[0] =~ /^\/.*\/([^\/]+)$/) {
+	# A file .. the index file is in ~/.usermin/mailbox or
+	# /etc/webmin/mailboxes
+	if ($user_module_config_directory && $config{'shortindex'}) {
+		# Use short name for index file
+		$f = "$user_module_config_directory/$1.findex";
+		}
+	else {
+		$f = $user_module_config_directory ?
+			"$user_module_config_directory/$us.findex" :
+			"$module_config_directory/$us.findex";
+		}
+	}
+else {
+	# A username .. the index file is in /etc/webmin/mailboxes
+        $f = $user_module_config_directory ?
+		"$user_module_config_directory/$_[0].index" :
+		"$module_config_directory/$_[0].index";
+	}
 return -r $f && !-r "$f.$hn" ? $f : "$f.$hn";
 }
 
@@ -1576,17 +1613,9 @@ sub eucconv_and_escape {
 # Returns a subset of mail from a maildir format directory
 sub list_maildir
 {
-local (@rv, $i, $f, @files);
-foreach $d ("$_[0]/cur", "$_[0]/new") {
-	opendir(DIR, $d);
-	while($f = readdir(DIR)) {
-		push(@files, "$d/$f") if ($f !~ /^\./);
-		}
-	closedir(DIR);
-	}
-@files = sort { $a =~ /([^\/]+)$/; local $an = $1;
-		$b =~ /([^\/]+)$/; local $bn = $1;
-		$an cmp $bn } @files;
+local (@rv, $i, $f);
+local @files = &get_maildir_files($_[0]);
+
 local ($start, $end);
 if (!defined($_[1])) {
 	$start = 0;
@@ -1620,14 +1649,7 @@ return @rv;
 # Returns a list of messages with the given indexes, from a maildir directory
 sub select_maildir
 {
-local @files;
-foreach my $d ("$_[0]/cur", "$_[0]/new") {
-	opendir(DIR, $d);
-	while(my $f = readdir(DIR)) {
-		push(@files, "$d/$f") if ($f !~ /^\./);
-		}
-	closedir(DIR);
-	}
+local @files = &get_maildir_files($_[0]);
 local @rv;
 foreach my $i (@{$_[1]}) {
 	local $mail = &read_mail_file($files[$i], $_[2]);
@@ -1637,6 +1659,73 @@ foreach my $i (@{$_[1]}) {
 return @rv;
 }
 
+# Get ordered list of message files (with in-memory and on-disk caching, as
+# this can be slow)
+# get_maildir_files(directory)
+sub get_maildir_files
+{
+# Work out last modified time
+local $newest;
+foreach my $d ("$_[0]/cur", "$_[0]/new") {
+	local @dst = stat($d);
+	$newest = $dst[9] if ($dst[9] > $newest);
+	}
+
+local @files;
+if (defined($main::list_maildir_cache{$_[0]}) &&
+    $main::list_maildir_cache_time{$_[0]} == $newest) {
+	# Use the in-memory cache cache
+	@files = @{$main::list_maildir_cache{$_[0]}};
+	}
+else {
+	# Check the on-disk cache file
+	local $cachefile = &get_maildir_cachefile($_[0]);
+	local @cst = stat($cachefile);
+	if ($cst[9] >= $newest) {
+		# Can read the cache
+		open(CACHE, $cachefile);
+		while(<CACHE>) {
+			chop;
+			push(@files, $_[0]."/".$_);
+			}
+		close(CACHE);
+		}
+	else {
+		# Really read
+		local @shorts;
+		foreach my $d ("cur", "new") {
+			opendir(DIR, "$_[0]/$d");
+			while(my $f = readdir(DIR)) {
+				push(@shorts, "$d/$f")
+					if ($f ne "." && $f ne "..");
+				}
+			closedir(DIR);
+			}
+		@shorts = sort { substr($a, 4) cmp substr($b, 4) } @shorts;
+		@files = map { "$_[0]/$_" } @shorts;
+
+		# Write out the on-disk cache
+		&open_tempfile(CACHE, ">$cachefile", 1);
+		my $err;
+		foreach my $f (@shorts) {
+			my $ok = (print CACHE $f,"\n");
+			$err++ if (!$ok);
+			}
+		&close_tempfile(CACHE) if (!$err);
+		local @st = stat($_[0]);
+		if ($< == 0) {
+			# Cache should have some ownership as directory
+			&set_ownership_permissions($st[4], $st[5], undef,
+						   $cachefile);
+			}
+		}
+	$main::list_maildir_cache{$_[0]} = \@files;
+	$main::list_maildir_cache_time{$_[0]} = $st[7];
+
+	}
+return @files;
+}
+
 # search_maildir(file, field, what)
 # Search for messages in a maildir directory, and return the results
 sub search_maildir
@@ -1666,9 +1755,44 @@ return @rv;
 sub delete_maildir
 {
 local $m;
+
+# Find all maildirs being deleted from
+local %dirs;
+foreach $m (@_) {
+	if ($m->{'file'} =~ /^(.*)\/(cur|new)\/([^\/]+)$/) {
+		$dirs{$1}->{"$2/$3"} = 1;
+		}
+	}
+
+# Delete from caches
+foreach my $dir (keys %dirs) {
+	local $cachefile = &get_maildir_cachefile($dir);
+	local @cst = stat($cachefile);
+	next if (!@cst);
+
+	# Work out last modified time, and don't update cache if too new
+	local $newest;
+	foreach my $d ("$dir/cur", "$dir/new") {
+		local @dst = stat($d);
+		$newest = $dst[9] if ($dst[9] > $newest);
+		}
+	next if ($newest > $cst[9]);
+
+	local $lref = &read_file_lines($cachefile);
+	for(my $i=0; $i<@$lref; $i++) {
+		if ($dirs{$dir}->{$lref->[$i]}) {
+			# Found an entry to remove
+			splice(@$lref, $i--, 1);
+			}
+		}
+	&flush_file_lines($cachefile);
+	}
+
+# Actually delete the files
 foreach $m (@_) {
 	unlink($m->{'file'});
 	}
+
 }
 
 # modify_maildir(&oldmail, &newmail, textonly)
@@ -1683,6 +1807,20 @@ unlink($_[0]->{'file'});
 # Adds some message in maildir format to a directory
 sub write_maildir
 {
+# Work out last modified time, and don't update cache if too new
+local $cachefile = &get_maildir_cachefile($_[1]);
+local @cst = stat($cachefile);
+local $up2date = 0;
+if (@cst) {
+	local $newest;
+	foreach my $d ("$dir/cur", "$dir/new") {
+		local @dst = stat($d);
+		$newest = $dst[9] if ($dst[9] > $newest);
+		}
+	$up2date = 1 if ($newest <= $cst[9]);
+	}
+
+# Select a unique filename and write to it
 local $now = time();
 local $hn = &get_system_hostname();
 local $mf;
@@ -1693,6 +1831,14 @@ do {
 	$now++;
 	} while(-r $mf);
 &send_mail($_[0], $mf, $_[2], 1);
+
+if ($up2date) {
+	# Bring cache up to date
+	$now--;
+	local $lref = &read_file_lines($cachefile);
+	push(@$lref, "cur/$now.$$.$hn");
+	&flush_file_lines($cachefile);
+	}
 }
 
 # empty_maildir(file)
@@ -1708,6 +1854,16 @@ foreach $d ("$_[0]/cur", "$_[0]/new") {
 		}
 	closedir(DIR);
 	}
+unlink(&get_maildir_cachefile($_[0]));
+}
+
+# get_maildir_cachefile(dir)
+# Returns the cache file for a maildir directory
+sub get_maildir_cachefile
+{
+local ($dir) = @_;
+return -r "$dir/maildircache" ? "$dir/maildircache"
+			      : "$dir/.usermin-maildircache";
 }
 
 # count_maildir(dir)
@@ -2029,16 +2185,28 @@ return \%rv;
 # Converts a mail Date: header into a unix time
 sub parse_mail_date
 {
+open(OLDSTDERR, ">&STDERR");	# suppress STDERR from Time::Local
+close(STDERR);
 my $rv = eval {
-	if ($_[0] =~ /^\s*(\S+),\s+(\d+)\s+(\S+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+(\S+)/) {
+	if ($_[0] =~ /^\s*(\S+),\s+(\d+)\s+(\S+)\s+(\d+)\s+(\d+):\s?(\d+):\s?(\d+)\s+(\S+)/) {
 		# Format like Mon, 13 Dec 2004 14:40:41 +0100
 		# or          Mon, 13 Dec 2004 14:18:16 GMT
 		# or	      Tue, 14 Sep 04 02:45:09 GMT
 		local $tm = timegm($7, $6, $5, $2, &month_to_number($3),
 				   $4 < 50 ? $4+100 : $4 < 1000 ? $4 : $4-1900);
-		local $tz = int($8);
-		$tz = $tz/100 if ($tz >= 50 || $tz <= -50);
-		$tm -= $tz*60*60;
+		local $tz = $8;
+		if ($tz =~ /^(\-|\+)?\d+$/) {
+			local $tz = int($tz);
+			$tz = $tz/100 if ($tz >= 50 || $tz <= -50);
+			$tm -= $tz*60*60;
+			}
+		return $tm;
+		}
+	elsif ($_[0] =~ /^\s*(\S+),\s+(\d+)\s+(\S+)\s+(\d+)\s+(\d+):\s?(\d+):\s?(\d+)/) {
+		# Format like Mon, 13 Dec 2004 14:40:41
+		# No timezone, so assume local
+		local $tm = timegm($7, $6, $5, $2, &month_to_number($3),
+				   $4 < 50 ? $4+100 : $4 < 1000 ? $4 : $4-1900);
 		return $tm;
 		}
 	elsif ($_[0] =~ /^\s*(\S+)\s+(\S+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+(\d+)/) {
@@ -2056,21 +2224,46 @@ my $rv = eval {
 		# Format like 30 Jun 2005 21:01:01 -0000
 		local $tm = timegm($6, $5, $4, $1, &month_to_number($2),
 				   $3 < 50 ? $3+100 : $3 < 1000 ? $3 : $3-1900);
-		local $tz = int($7);
-		$tz = $tz/100 if ($tz >= 50 || $tz <= -50);
-		$tm -= $tz*60*60;
+		local $tz = $7;
+		if ($tz =~ /^(\-|\+)?\d+$/) {
+			$tz = int($tz);
+			$tz = $tz/100 if ($tz >= 50 || $tz <= -50);
+			$tm -= $tz*60*60;
+			}
 		return $tm;
 		}
 	else {
-		print STDERR "could not parse $_[0]\n";
 		return undef;
 		}
 	};
+open(STDERR, ">&OLDSTDERR");
+close(OLDSTDERR);
 if ($@) {
-	print STDERR "Error parsing $_[0]\n";
 	return undef;
 	}
 return $rv;
 }
 
+# send_text_mail(from, to, cc, subject, body, [smtp-server])
+# A convenience function for sending a email with just a text body
+sub send_text_mail
+{
+local ($from, $to, $cc, $subject, $body, $smtp) = @_;
+local $cs = &get_charset();
+local $attach = $body =~ /[\177-\377]/ ?
+	{ 'headers' => [ [ 'Content-Type', 'text/plain; charset='.$cs ],
+		         [ 'Content-Transfer-Encoding', 'quoted-printable' ] ],
+          'data' => &quoted_encode($body) } :
+	{ 'headers' => [ [ 'Content-type', 'text/plain' ] ],
+	  'data' => &entities_to_ascii($body) };
+local $mail = { 'headers' =>
+		[ [ 'From', $from ],
+		  [ 'To', $to ],
+		  [ 'Cc', $cc ],
+		  [ 'Subject', $subject ] ],
+		'attach' => [ $attach ] };
+return &send_mail($mail, undef, 1, 0, $smtp);
+}
+
+
 1;
