@@ -326,7 +326,7 @@ elsif ($folder->{'type'} == 2) {
 	local $h = $rv[1];
 	local @uidl = &pop3_uidl($h);
 
-	# For each message in the range, get the headers or body
+	# Work out what we have cached
 	local ($i, $f, %cached, %sizeneed);
 	local @rv;
 	local $cd = "$cache_directory/$_[2]->{'id'}.cache";
@@ -344,6 +344,8 @@ elsif ($folder->{'type'} == 2) {
 	else {
 		mkdir($cd, 0700);
 		}
+
+	# For each requested index, get the headers or body
 	foreach my $i (@$indexes) {
 		local $u = &safe_uidl($uidl[$i]);
 		if ($cached{$u} == 2 || $cached{$u} == 1 && $headersonly) {
@@ -396,8 +398,10 @@ elsif ($folder->{'type'} == 2) {
 			s/\r//g;
 			last if ($_ eq ".\n");
 			if (/^(\d+)\s+(\d+)/ && $sizeneed{$1-1}) {
-				# Add size to the mail cache
-				$rv[$1-1]->{'size'} = $2;
+				# Find mail in results, and set its size
+				local ($ns) = grep { $_->{'idx'} == $1-1 } @rv;
+				next if (!$ns);
+				$ns->{'size'} = $2;
 				local $u = &safe_uidl($uidl[$1-1]);
 				open(CACHE, ">>$cd/$u.headers");
 				print CACHE $2,"\n";
@@ -410,7 +414,39 @@ elsif ($folder->{'type'} == 2) {
 	}
 elsif ($folder->{'type'} == 4) {
 	# IMAP folder
-	# XXX
+
+	# Login and select the specified mailbox
+	local @irv = &imap_login($folder);
+	if ($irv[0] != 1) {
+		# Something went wrong
+		if ($_[4]) {
+			@{$_[4]} = @irv;
+			return ();
+			}
+		elsif ($irv[0] == 0) { &error($irv[1]); }
+		elsif ($irv[0] == 3) { &error(&text('save_emailbox', $irv[1])); }
+		elsif ($irv[0] == 2) { &error(&text('save_elogin2', $irv[1])); }
+		}
+	local $h = $irv[1];
+	local $count = $irv[2];
+	return () if (!$count);
+
+	# Fetch each mail by index
+	local @rv;
+	local $wanted = $headersonly ? "(RFC822.SIZE RFC822.HEADER)"
+				     : "RFC822";
+	local @idxrv = &imap_command($h,
+		"FETCH ".join(",", map { $_+1 } @$indexes)." $wanted");
+	local $i = 0;
+	foreach my $idxrv (@{idxrv->[1]}) {
+		local $mail = &parse_imap_mail($idxrv);
+		if ($mail) {
+			$mail->{'idx'} = $indexes->[$i++];
+			push(@rv, $mail);
+			}
+		}
+
+	return @rv;
 	}
 elsif ($folder->{'type'} == 5) {
 	# Composite folder .. need to convert each index to a position
@@ -447,27 +483,61 @@ elsif ($folder->{'type'} == 6) {
 	local @rv;
 	local $need_save = 0;
 	local @newmems = @$mems;
+
+	# Work out how many different sub-folders we have
+	local $sf;
 	foreach my $i (@$indexes) {
-		local $sf = $mems->[$i]->[0];
-		local $idx = $mems->[$i]->[1];
-		local ($mail) = &mailbox_select_mails($sf, [ $idx ],
-						      $headersonly);
-		if ($mems->[$i]->[2] &&
-		    $mail->{'header'}->{'message-id'} ne $mems->[$i]->[2]) {
-			local $real = &find_by_message_id($sf,$mems->[$i]->[2]);
+		$sf = $mems->[$i]->[0];
+		$sfcount{&folder_name($sf)} = 1;
+		}
+	if ((keys %sfcount) == 1) {
+		# Just one, so we can do one big select
+		local (@sfindexes, @sfmems);
+		foreach my $i (@$indexes) {
+			push(@sfindexes, $mems->[$i]->[1]);
+			push(@sfmems, $i);
+			}
+		local $i = 0;
+		foreach my $mail (&mailbox_select_mails($sf, \@sfindexes,
+							$headersonly)) {
+			$mail->{'mem'} = $mems->[$sfmems[$i]];
+			&push_index($mail, $sf, $sfmems[$i]);
+			push(@rv, $mail);
+			$i++;
+			}
+		}
+	elsif ((keys %sfcount) > 0) {
+		# Several sub-folders, so we need to select from each
+		foreach my $i (@$indexes) {
+			local $sf = $mems->[$i]->[0];
+			local $idx = $mems->[$i]->[1];
+			local ($mail) = &mailbox_select_mails($sf, [ $idx ],
+							      $headersonly);
+			$mail->{'mem'} = $mems->[$i];
+			&push_index($mail, $sf, $i);
+			push(@rv, $mail);
+			}
+		}
+
+	# Fix up any mistmatches between indexes and message IDs
+	foreach $mail (@rv) {
+		local $mem = $mail->{'mem'};
+		if ($mem->[2] &&
+		    $mail->{'header'}->{'message-id'} ne $mem->[2]) {
+			# Message ID mismatch! Fix it..
+			local $real = &find_by_message_id(
+					$mail->{'subfolder'}, $mem->[2]);
 			if ($real) {
-				$mems->[$i]->[1] = $real->{'idx'};
+				$mem->[1] = $real->{'idx'};
 				$mail = $real;
 				$need_save++;
 				}
 			else {
 				# Doesn't exist! Lose from index..
-				@newmems = grep { $_ ne $mems->[$i] } @newmems;
+				@newmems = grep { $_ ne $mem } @newmems;
 				$need_save++;
 				}
 			}
-		&push_index($mail, $sf, $i);
-		push(@rv, $mail);
 		}
 	if ($need_save) {
 		$folder->{'members'} = \@newmems;
@@ -512,8 +582,13 @@ if (!$field) {
 	}
 if (!$field || !$folder->{'sortable'}) {
 	# No sorting .. just return newest first
-	return reverse(&mailbox_list_mails(
+	local @rv = reverse(&mailbox_list_mails(
 		-$start, -$end-1, $folder, $headers, $error));
+	local $i = 0;
+	foreach my $m (@rv) {
+		$m->{'sortidx'} = $i++;
+		}
+	return @rv;
 	}
 
 # Build the appropriate sort index, if missing
@@ -527,7 +602,7 @@ while(my ($k, $v) = each %index) {
 		push(@sorter, [ $1, lc($v) ]);
 		}
 	}
-if ($field eq "size" || $field eq "date") {
+if ($field eq "size" || $field eq "date" || $field eq "x-spam-status") {
 	# Numeric sort
 	@sorter = sort { my $s = $a->[1] <=> $b->[1]; $dir ? $s : -$s } @sorter;
 	}
@@ -543,7 +618,7 @@ local @wantindexes = map { $sorter[$_]->[0] } ($start .. $end);
 local @mails = &mailbox_select_mails($folder, \@wantindexes, $headers);
 for(my $i=0; $i<@mails; $i++) {
 	$rv[$start+$i] = $mails[$i];
-	$mails[$i]->{'sortidx'} = $i;
+	$mails[$i]->{'sortidx'} = $start+$i;
 	}
 return @rv;
 }
@@ -608,11 +683,14 @@ local ($folder, $field, $index) = @_;
 return 0 if (!$folder->{'sortable'});
 local $ifile = &folder_sort_index_file($folder);
 
-dbmopen(%$index, $ifile, 0600);
-if ($index->{'lastchange'} < $folder->{'lastchange'}) {
+&open_dbm_db($index, $ifile, 0600);
+if ($index->{'lastchange'} < $folder->{'lastchange'} ||
+    !$folder->{'lastchange'}) {
 	# The mail file is newer than the index .. add messages not in the index
 	# to it.
-	local $realcount = &mailbox_folder_size($folder);
+	local $realcount = $folder->{'type'} == 6 ?
+			scalar(@{$folder->{'members'}}) :
+			&mailbox_folder_size($folder);
 	local $indexcount = int($index->{'mailcount'});
 	if ($realcount < $indexcount) {
 		# Mail size has decreased! Need total rebuild
@@ -622,7 +700,7 @@ if ($index->{'lastchange'} < $folder->{'lastchange'}) {
 	local @mails = $realcount ? &mailbox_select_mails($folder,
 					[ $indexcount .. $realcount-1 ], 1)
 				  : ( );
-	my @index_fields = ( "subject", "from", "to", "date", "size", "message-id" );
+	my @index_fields = ( "subject", "from", "to", "date", "size", "x-spam-status", "message-id" );
 	foreach my $mail (@mails) {
 		foreach my $f (@index_fields) {
 			if ($f eq "date") {
@@ -640,6 +718,16 @@ if ($index->{'lastchange'} < $folder->{'lastchange'}) {
 				$index->{$mail->{'idx'}."_".$f} =
 					&simplify_from($mail->{'header'}->{$f});
 				}
+			elsif ($f eq "subject") {
+				# Convert subject to display version
+				$index->{$mail->{'idx'}."_".$f} =
+				    &simplify_subject($mail->{'header'}->{$f});
+				}
+			elsif ($f eq "x-spam-status") {
+				# Extract spam score
+				$index->{$mail->{'idx'}."_".$f} =
+					$mail->{'header'}->{$f} =~ /(hits|score)=([0-9\.]+)/ ? $2 : undef;
+				}
 			else {
 				# Just a header
 				$index->{$mail->{'idx'}."_".$f} =
@@ -661,7 +749,7 @@ local ($folder) = @_;
 local $ifile = &folder_sort_index_file($folder);
 
 my %index;
-dbmopen(%index, $ifile, 0600);
+&open_dbm_db(\%index, $ifile, 0600);
 %index = ( );
 }
 
@@ -684,6 +772,7 @@ if (&build_sort_index($folder, undef, \%index)) {
 		if ($k =~ /^(\d+)_message-id$/ && $v eq $mid) {
 			# Got it!
 			local ($rv) = &mailbox_select_mails($folder, [ $1 ]);
+			&set_sort_indexes($folder, [ $rv ]);
 			return $rv;
 			}
 		}
@@ -691,12 +780,81 @@ if (&build_sort_index($folder, undef, \%index)) {
 return undef;
 }
 
+# find_message_by_index(&mails, &folder, index, mid)
+# Finds a message by index (in a sorted folder) or message ID
+sub find_message_by_index
+{
+local ($mails, $folder, $idx, $mid) = @_;
+local $mail = $mails->[$idx];
+if ($mid && (!$mail || $mail->{'header'}->{'message-id'} ne $mid)) {
+	$mail = &find_by_message_id($folder, $mid);
+	}
+return $mail;
+}
+
 # mailbox_search_mail(&fields, andmode, &folder, [&limit])
 # Search a mailbox for multiple matching fields
 sub mailbox_search_mail
 {
+# For folders other than mbox and IMAP, build a sort index and use that for
+# the search, if it is simple enough (Subject, From and To only)
+local @idxfields = grep { $_->[0] eq 'from' || $_->[0] eq 'to' ||
+                          $_->[0] eq 'subject' } @{$_[0]};
+if ($_[2]->{'type'} != 0 && $_[2]->{'type'} != 4 && $_[2]->{'type'} != 5 &&
+    scalar(@idxfields) == scalar(@{$_[0]}) && @idxfields &&
+    &get_product_name() eq 'usermin') {
+	local %index;
+	&build_sort_index($_[2], undef, \%index);
+	local @rv;
+
+	# Work out which mail indexes match the requested headers
+	local %idxmatches = map { ("$_->[0]/$_->[1]", [ ]) } @idxfields;
+	while(my ($k, $v) = each %index) {
+		local ($ki, $kf) = split(/_/, $k, 2);
+		next if (!$kf || $ki eq '');
+
+		# Check all of the fields to see which ones match
+		foreach my $if (@idxfields) {
+			local $iff = $if->[0];
+			local ($neg) = ($iff =~ s/^\!//);
+			next if ($kf ne $iff);
+			if (!$neg && $v =~ /\Q$if->[1]\E/i ||
+			    $neg && $v !~ /\Q$if->[1]\E/i) {
+				push(@{$idxmatches{"$if->[0]/$if->[1]"}}, $ki);
+				}
+			}
+		}
+	local @matches;
+	if ($_[1]) {
+		# Find indexes in all arrays
+		local %icount;
+		foreach my $if (keys %idxmatches) {
+			foreach my $i (@{$idxmatches{$if}}) {
+				$icount{$i}++;
+				}
+			}
+		foreach my $i (keys %icount) {
+			}
+		local $fif = $idxfields[0];
+		@matches = grep { $icount{$_} == scalar(@idxfields) }
+				@{$idxmatches{"$fif->[0]/$fif->[1]"}};
+		}
+	else {
+		# Find indexes in any array
+		foreach my $if (keys %idxmatches) {
+			push(@matches, @{$idxmatches{$if}});
+			}
+		@matches = &unique(@matches);
+		}
+	@matches = sort { $a <=> $b } @matches;
+
+	# Select the actual mails
+	return &mailbox_select_mails($_[2], \@matches, 0);
+	}
+
 if ($_[2]->{'type'} == 0) {
-	# Just search an mbox format file
+	# Just search an mbox format file (which will use its own special
+	# field-level index)
 	return &advanced_search_mail($_[2]->{'file'}, $_[0], $_[1], $_[3]);
 	}
 elsif ($_[2]->{'type'} == 1) {
@@ -828,7 +986,7 @@ sub mailbox_delete_mail
 {
 return undef if (&is_readonly_mode());
 local $f = shift(@_);
-if ($userconfig{'delete_mode'} == 1 && !$f->{'trash'}) {
+if ($userconfig{'delete_mode'} == 1 && !$f->{'trash'} && !$f->{'spam'}) {
 	# Copy to trash folder first
 	local ($trash) = grep { $_->{'trash'} } &list_folders();
 	local $m;
@@ -838,9 +996,11 @@ if ($userconfig{'delete_mode'} == 1 && !$f->{'trash'}) {
 	}
 
 if ($f->{'type'} == 0) {
+	# Delete from mbox
 	&delete_mail($f->{'file'}, @_);
 	}
 elsif ($f->{'type'} == 1) {
+	# Delete from Maildir
 	&delete_maildir(@_);
 	}
 elsif ($f->{'type'} == 2) {
@@ -863,6 +1023,7 @@ elsif ($f->{'type'} == 2) {
 		}
 	}
 elsif ($f->{'type'} == 3) {
+	# Delete from MH dir
 	&delete_mhdir(@_);
 	}
 elsif ($f->{'type'} == 4) {
@@ -892,6 +1053,7 @@ elsif ($f->{'type'} == 5) {
 		}
 	}
 elsif ($f->{'type'} == 6) {
+	# Delete from virtual folder
 	local $sm;
 	if ($f->{'delete'}) {
 		# Delete the underlying messages, and remove from index
@@ -928,9 +1090,27 @@ elsif ($f->{'type'} == 6) {
 	&save_folder($f, $f);
 	}
 
-# Update folder index to remove their entries
+# Update folder index to remove indexes for deleted messages
 if ($f->{'sortable'}) {
-	&delete_sort_index($f);
+	local %index;
+	&build_sort_index($f, undef, \%index);
+	foreach my $m (sort { $b->{'idx'} <=> $a->{'idx'} } @_) {
+		foreach my $k (sort { $a <=> $b } (keys %index)) {
+			local ($ki, $kf) = split(/_/, $k, 2);
+			if ($m->{'idx'} == $ki) {
+				# Delete this actual mail from the index
+				delete($index{$k});
+				}
+			elsif ($m->{'idx'} < $ki) {
+				# Mail is after one being deleted .. shift down
+				$index{($ki-1)."_".$kf} = $index{$k};
+				delete($index{$k});
+				}
+			}
+		}
+	$index{'mailcount'} -= scalar(@_);
+	$index{'lastchange'} = time();
+	dbmclose(%index);
 	}
 }
 
@@ -1002,6 +1182,57 @@ if ($folder->{'sortable'}) {
 	}
 }
 
+# mailbox_copy_folder(&source, &dest)
+# Copy all messages from one folder to another. This is done in an optimized
+# way if possible.
+sub mailbox_copy_folder
+{
+local ($src, $dest) = @_;
+if ($src->{'type'} == 0 && $dest->{'type'} == 0) {
+	# mbox to mbox .. just read and write the files
+	&open_readfile(SOURCE, $src->{'file'});
+	&open_tempfile(DEST, ">>$dest->{'file'}");
+	while(read(SOURCE, $buf, 1024) > 0) {
+		&print_tempfile(DEST, $buf);
+		}
+	&close_tempfile(DEST);
+	close(SOURCE);
+	}
+elsif ($src->{'type'} == 1 && $dest->{'type'} == 1) {
+	# maildir to maildir .. just copy the files
+	local @files = &get_maildir_files($src->{'file'});
+	foreach my $f (@files) {
+		local $fn = $f;
+		$fn =~ s/^.*\///;
+		&copy_source_dest($f, "$dest->{'file'}/$fn");
+		}
+	}
+elsif ($src->{'type'} == 1 && $dest->{'type'} == 0) {
+	# maildir to mbox .. append all the files
+	local @files = &get_maildir_files($src->{'file'});
+	&open_tempfile(DEST, ">>$dest->{'file'}");
+	foreach my $f (@files) {
+		&open_readfile(SOURCE, $f);
+		while(read(SOURCE, $buf, 1024) > 0) {
+			&print_tempfile(DEST, $buf);
+			}
+		close(SOURCE);
+		}
+	&close_tempfile(DEST);
+	}
+else {
+	# read in all mail and write out, in 100 message blocks
+	local $max = &mailbox_folder_size($src);
+	for(my $s=0; $s<$max; $s+=100) {
+		local $e = $s+99;
+		$e = $max-1 if ($e >= $max);
+		local @mail = &mailbox_list_mails($s, $e, $src);
+		local @want = @mail[$s..$e];
+		&mailbox_copy_mail($src, $dest, @want);
+		}
+	}
+}
+
 # mailbox_move_mail(&source, &dest, mail, ...)
 # Move mail from one folder to another
 sub mailbox_move_mail
@@ -1210,7 +1441,7 @@ if ($_[2]->{'sortable'}) {
 	}
 }
 
-# mailbox_folder_size(&folder)
+# mailbox_folder_size(&folder, [estimate])
 # Returns the number of messages in some folder
 sub mailbox_folder_size
 {
@@ -1260,7 +1491,7 @@ elsif ($_[0]->{'type'} == 5) {
 		}
 	return $rv;
 	}
-elsif ($_[0]->{'type'} == 6) {
+elsif ($_[0]->{'type'} == 6 && !$_[1]) {
 	# A virtual folder .. we need to exclude messages that no longer
 	# exist in the parent folders
 	my $rv = 0;
@@ -1271,6 +1502,10 @@ elsif ($_[0]->{'type'} == 6) {
 		}
 	return $rv;
 	}
+elsif ($_[0]->{'type'} == 6 && $_[1]) {
+	# A virtual folder .. but we can just use the last member count
+	return scalar(@{$_[0]->{'members'}});
+	}
 }
 
 # pop3_login(&folder)
@@ -1373,24 +1608,30 @@ foreach $f (keys %imap_login_handle) {
 sub imap_login
 {
 local $h = $imap_login_handle{$_[0]->{'id'}};
-return (1, $h) if ($h);
-$h = time().++$imap_login_count;
-local $error;
-&open_socket($_[0]->{'server'}, $_[0]->{'port'} || $imap_port, $h, \$error);
-return (0, $error) if ($error);
-local $os = select($h); $| = 1; select($os);
+local @rv;
+if (!$h) {
+	# Need to open socket
+	$h = time().++$imap_login_count;
+	local $error;
+	&open_socket($_[0]->{'server'}, $_[0]->{'port'} ||
+		$imap_port, $h, \$error);
+	return (0, $error) if ($error);
+	local $os = select($h); $| = 1; select($os);
 
-# Login normally
-local @rv = &imap_command($h);
-return (0, $rv[3]) if (!$rv[0]);
-@rv = &imap_command($h, "login \"$_[0]->{'user'}\" \"$_[0]->{'pass'}\"");
-return (2, $rv[3]) if (!$rv[0]);
+	# Login normally
+	@rv = &imap_command($h);
+	return (0, $rv[3]) if (!$rv[0]);
+	@rv = &imap_command($h,"login \"$_[0]->{'user'}\" \"$_[0]->{'pass'}\"");
+	return (2, $rv[3]) if (!$rv[0]);
 
-# Select the right folder
+	$imap_login_handle{$_[0]->{'id'}} = $h;
+	}
+
+# Select the right folder (if one was given)
 @rv = &imap_command($h, "select ".($_[0]->{'mailbox'} || "INBOX"));
 return (3, $rv[3]) if (!$rv[0]);
-return (1, $imap_login_handle{$_[0]->{'id'}} = $h,
-	$rv[2] =~ /\*\s+(\d+)\s+EXISTS/i ? $1 : undef);
+local $count = $rv[2] =~ /\*\s+(\d+)\s+EXISTS/i ? $1 : undef;
+return (1, $h, $count);
 }
 
 # imap_command(handle, command)
@@ -1603,10 +1844,11 @@ sub safe_html
 {
 local $html = $_[0];
 local $bodystuff;
-if ($html =~ s/^[\000-\377]*<BODY(.*)>//i) {
+if ($html =~ s/^[\000-\377]*<BODY([^>]*)>//i) {
 	$bodystuff = $1;
 	}
 $html =~ s/<\/BODY>[\000-\377]*$//i;
+$html =~ s/<base[^>]*>//i;
 $html = &filter_javascript($html);
 $html = &safe_urls($html);
 $bodystuff = &safe_html($bodystuff) if ($bodystuff);
@@ -1621,6 +1863,7 @@ local $html = $_[0];
 return undef if ($html !~ /<HEAD[^>]*>/i || $html !~ /<\/HEAD[^>]*>/i);
 $html =~ s/^[\000-\377]*<HEAD[^>]*>//gi || &error("Failed to filter <pre>".&html_escape($html)."</pre>");
 $html =~ s/<\/HEAD[^>]*>[\000-\377]*//gi || &error("Failed to filter <pre>".&html_escape($html)."</pre>");
+$html =~ s/<base[^>]*>//i;
 return &filter_javascript($html);
 }
 
@@ -1661,6 +1904,10 @@ elsif ($url =~ /^(http:|https:)/) {
 		return $before."_unsafe_link_".$after;
 		}
 	}
+elsif ($url =~ /^mailto:([a-z0-9\.\-\_\@]+)/i) {
+	# A mailto link, which we can convert
+	return $before."reply_mail.cgi?new=1&to=".&urlize($1).$after;
+	}
 else {
 	# Relative URL like foo.cgi or /foo.cgi or ../foo.cgi - unsafe!
 	return $before."_unsafe_link_".$after;
@@ -1745,19 +1992,8 @@ foreach $f (@_) {
 		$f->{'size'} = $st[7];
 		}
 	elsif ($f->{'type'} == 1) {
-		# Maildir folder size is that of all mail files
-		local $qd;
-		$f->{'size'} = 0;
-		foreach $qd ('cur', 'new') {
-			local $mf;
-			opendir(QDIR, "$f->{'file'}/$qd");
-			while($mf = readdir(QDIR)) {
-				next if ($mf eq "." || $mf eq "..");
-				local @st = stat("$f->{'file'}/$qd/$mf");
-				$f->{'size'} += $st[7];
-				}
-			closedir(QDIR);
-			}
+		# Maildir folder size is that of all files in it
+		$f->{'size'} = &recursive_disk_usage($f->{'file'});
 		}
 	elsif ($f->{'type'} == 3) {
 		# MH folder size is that of all mail files
@@ -1771,12 +2007,26 @@ foreach $f (@_) {
 			}
 		closedir(MHDIR);
 		}
+	elsif ($f->{'type'} == 4) {
+		# Get size of IMAP folder
+		local ($ok, $h, $count) = &imap_login($f);
+		if ($ok) {
+			$f->{'size'} = 0;
+			local @rv = &imap_command($h,
+				"FETCH 1:$count (RFC822.SIZE)");
+			foreach my $r (@{$rv[1]}) {
+				if ($r =~ /RFC822.SIZE\s+(\d+)/) {
+					$f->{'size'} += $1;
+					}
+				}
+			}
+		}
 	elsif ($f->{'type'} == 5) {
 		# Size of a combined folder is the size of all sub-folders
 		return &folder_size(@{$f->{'subfolders'}});
 		}
 	else {
-		# Cannot get size of a remote folder
+		# Cannot get size of a POP3 folder
 		$f->{'size'} = undef;
 		}
 	$total += $f->{'size'};
@@ -1834,7 +2084,9 @@ foreach $f (@files) {
 		 $f eq "cur" || $f eq "tmp" || $f eq "new" ||
 		 $f =~ /^\.imap/i || $f eq ".customflags" ||
 		 $f eq "dovecot-uidlist" || $f =~ /^courierimap/ ||
-		 $f eq "maildirfolder" || $f eq "maildirsize");
+		 $f eq "maildirfolder" || $f eq "maildirsize" ||
+		 $f eq "maildircache" || $f eq ".subscriptions" ||
+                 $f eq ".usermin-maildircache" || $f =~ /^dovecot\.index/);
 	local $p = "$_[0]/$f";
 	local $added = 0;
 	if ($_[1] || !-d $p || -d "$p/cur") {
@@ -1891,7 +2143,8 @@ local ($plainbody, $htmlbody) = &find_body($mail, $mode);
 local ($quote, $html_edit, $body);
 local $cfg = defined(%userconfig) ? \%userconfig : \%config;
 local @writers = &split_addresses($mail->{'header'}->{'from'});
-local $writer = ($writers[0]->[1] || $writers[0]->[0])." wrote ..";
+local $writer = &decode_mimewords($writers[0]->[1] || $writers[0]->[0]).
+		" wrote ..";
 local $tm;
 if ($cfg->{'reply_date'} &&
     ($tm = &parse_mail_date($_[0]->{'header'}->{'date'}))) {
@@ -2129,7 +2382,9 @@ foreach my $folder (@$folders) {
 		$folder->{'sortable'} = 1;
 		}
 	else {
-		$folder->{'sortable'} = 0;
+		# For POP3 and IMAP folders, we don't know the last change
+		$folder->{'lastchange'} = undef;
+		$folder->{'sortable'} = 1;
 		}
 	}
 }
@@ -2174,5 +2429,20 @@ if ($data =~ /\S/) {
 return undef;
 }
 
+# open_dbm_db(&hash, file, mode)
+# Attempts to open a DBM, first using SDBM_File, and then NDBM_File
+sub open_dbm_db
+{
+local ($hash, $file, $mode) = @_;
+eval "use SDBM_File";
+dbmopen(%$hash, $file, $mode);
+eval { $hash->{'1111111111'} = 'foo bar' };
+if ($@) {
+	dbmclose(%$hash);
+	eval "use NDBM_File";
+	dbmopen(%$hash, $file, $mode);
+	}
+}
+
 1;
 
