@@ -1,13 +1,18 @@
 # mailbox-lib.pl
-# XXX can we cache IMAP messages?
 # XXX don't reply-to our address
 #
 # XXX folder sorting
 #	XXX what about imap/pop3
-#	XXX searching should use sort index, where possible
 #	XXX update mailboxes module too
-#	XXX when deleting mail, update any virtual folders that refer to it
 #	XXX mailbox_list_mails should call mailbox_select_mails for type 6
+#
+# XXX mid in links
+#	XXX display should use sort index
+#	XXX folder list doens't need to really count virtual folder size
+#	    (or it should be faster!)
+#		XXX include in new theme
+#
+# XXX return to mail list after sending (preference)
 
 do '../web-lib.pl';
 &init_config();
@@ -30,6 +35,7 @@ $folders_dir = "$remote_user_info[7]/$userconfig{'mailbox_dir'}";
 %folder_types = map { $_, 1 } (split(/,/, $config{'folder_types'}),
 			       split(/,/, $config{'folder_virts'}));
 $search_folder_id = 1;
+$auto_cmd = "$user_module_config_directory/auto.pl";
 
 # mailbox_file()
 sub mailbox_file
@@ -188,19 +194,115 @@ if ($config{'mail_system'} == 2) {
 		    'type' => 2,
 		    'server' => $config{'pop3_server'} || "localhost",
 		    'mode' => 3,
+		    'remote' => 1,
+		    'nowrite' => 1,
 		    'inbox' => 1,
 		    'index' => 0 });
 	&read_file("$user_module_config_directory/inbox.pop3", $rv[$#rv]);
 	}
 elsif ($config{'mail_system'} == 4) {
 	# IMAP inbox
+	local $imapserver = $config{'pop3_server'} || "localhost";
 	push(@rv, { 'name' => $text{'folder_inbox'},
+		    'id' => 'INBOX',
 		    'type' => 4,
-		    'server' => $config{'pop3_server'} || "localhost",
+		    'server' => $imapserver,
 		    'mode' => 3,
+		    'remote' => 1,
 		    'inbox' => 1,
 		    'index' => 0 });
 	&read_file("$user_module_config_directory/inbox.imap", $rv[$#rv]);
+
+	# Use HTTP username and password, if available
+	if (!defined($rv[0]->{'user'}) && $main::remote_user) {
+		$rv[0]->{'user'} = $main::remote_user;
+		}
+	if (!defined($rv[0]->{'pass'}) && $main::remote_pass) {
+		$rv[0]->{'pass'} = $main::remote_pass;
+		}
+
+	# Get other IMAP folders (if we can)
+	local ($ok, $ih) = &imap_login($rv[0]);
+	if ($ok == 1) {
+		local @irv = &imap_command($ih, "list \"\" \"*\"");
+		if ($irv[0]) {
+			foreach my $l (@{$irv[1]}) {
+				if ($l =~ /LIST\s+\(.*\)\s+"(.*)"\s+"(.*)"/ &&
+				    $2 ne "INBOX") {
+					# Found a folder line
+					push(@rv,
+					  { 'name' => $2,
+					    'id' => $2,
+					    'type' => 4,
+					    'server' => $imapserver,
+					    'user' => $rv[0]->{'user'},
+					    'pass' => $rv[0]->{'pass'},
+					    'mode' => 0,
+					    'remote' => 1,
+					    'imapauto' => 1,
+					    'mailbox' => $2,
+					    'index' => scalar(@rv) });
+					&read_file("$user_module_config_directory/$2.imap", $rv[$#rv]);
+					}
+				}
+			}
+		}
+
+	# Find or create the IMAP sent mail folder
+	local ($sent) = grep { lc($_->{'name'}) eq 'sent'} @rv;
+	if (!$sent) {
+		local @irv = &imap_command($ih, "create \"sent\"");
+		if ($irv[0]) {
+			$sent = { 'id' => "sent",
+			          'type' => 4,
+				  'server' => $imapserver,
+				  'user' => $rv[0]->{'user'},
+				  'pass' => $rv[0]->{'pass'},
+				  'mode' => 2,
+				  'remote' => 1,
+				  'imapauto' => 1,
+				  'mailbox' => 'sent',
+			          'index' => scalar(@rv) };
+			push(@rv, $sent);
+			&read_file("$user_module_config_directory/sent.imap",
+				   $sent);
+			}
+		}
+	if ($sent) {
+		$sent->{'name'} = $text{'folder_sent'};
+		$sent->{'perpage'} = $userconfig{'perpage_sent_mail'};
+		$sent->{'fromaddr'} = $userconfig{'fromaddr_sent_mail'};
+		$sent->{'sent'} = 1;
+		$sent->{'mode'} = 2;
+		}
+
+	# Find or create the IMAP drafts folder
+	local ($drafts) = grep { lc($_->{'name'}) eq 'drafts'} @rv;
+	if (!$drafts) {
+		local @irv = &imap_command($ih, "create \"drafts\"");
+		if ($irv[0]) {
+			$drafts = { 'id' => "drafts",
+			            'type' => 4,
+				    'server' => $imapserver,
+				    'user' => $rv[0]->{'user'},
+				    'pass' => $rv[0]->{'pass'},
+				    'mode' => 3,
+				    'remote' => 1,
+				    'imapauto' => 1,
+				    'mailbox' => 'drafts',
+			            'index' => scalar(@rv) };
+			push(@rv, $drafts);
+			&read_file("$user_module_config_directory/drafts.imap",
+				   $drafts);
+			}
+		}
+	if ($drafts) {
+		$drafts->{'name'} = $text{'folder_drafts'};
+		$drafts->{'drafts'} = 1;
+		$drafts->{'mode'} = 3;
+		}
+
+	goto IMAPONLY;
 	}
 else {
 	# Local mail file inbox
@@ -237,8 +339,10 @@ push(@rv, { 'name' => $text{'folder_sent'},
 	    'index' => scalar(@rv) });
 
 # Add drafts file
-local $df = -r "$folders_dir/Drafts" ? "$folders_dir/Drafts"
-				     : "$folders_dir/drafts";
+local $df = -r "$folders_dir/Drafts" ? "$folders_dir/Drafts" :
+	    -r "$folders_dir/.Drafts" ? "$folders_dir/.Drafts" :
+	    -r "$folders_dir/.drafts" ? "$folders_dir/.drafts" :
+				        "$folders_dir/drafts";
 $done{$df}++;
 push(@rv, { 'name' => $text{'folder_drafts'},
 	    'type' => &folder_type($df),
@@ -265,7 +369,17 @@ if ($folder_types{'local'}) {
 				     $userconfig{'mailbox_recur'})) {
 		local $f = $p;
 		$f =~ s/^\Q$folders_dir\E\///;
-		push(@rv, { 'name' => $f,
+		local $name = $f;
+		if ($folders_dir eq "$remote_user_info[7]/Maildir") {
+			# A sub-folder under Maildir .. remove the . at the
+			# start of the sub-folder name
+			$name =~ s/^\.// || $name =~ s/\/\./\// || next;
+
+			# When in Maildir++ mode, any non-subdirectory
+			# is ignored
+			next if (!-d $p);
+			}
+		push(@rv, { 'name' => $name,
 			    'file' => $p,
 			    'type' => &folder_type($p),
 			    'perpage' => $userconfig{"perpage_$f"},
@@ -284,7 +398,7 @@ if ($inbox->{'type'} == 1) {
 		local $f = $p;
 		$f =~ s/^\Q$inbox->{'file'}\E\///;
 		local $name = $f;
-		$name =~ s/^\.//;
+		$name =~ s/^\.// || $name =~ s/\/\./\//;
 		push(@rv, { 'name' => "$name (Courier)",
 			    'file' => $p,
 			    'type' => &folder_type($p),
@@ -340,6 +454,9 @@ foreach $f (readdir(DIR)) {
 	}
 closedir(DIR);
 
+# When in IMAP inbox mode, we goto this label to skip all local folders
+IMAPONLY:
+
 # Add user-defined composite folders
 local %fcache;
 opendir(DIR, $user_module_config_directory);
@@ -386,6 +503,47 @@ foreach $f (readdir(DIR)) {
 	}
 closedir(DIR);
 
+# Add spam folder as specified in spamassassin module, in case it is
+# outside of the folders we scan
+if (&foreign_check("spam")) {
+	local %suserconfig = &foreign_config("spam", 1);
+	local $file = $suserconfig{'spam_file'};
+	$file =~ s/\.$//;
+	$file =~ s/\/$//;
+	$file = "$remote_user_info[7]/$file" if ($file && $file !~ /^\//);
+	$file =~ s/\~/$remote_user_info[7]/;
+	if ($file) {
+		if ($config{'mail_system'} == 4) {
+			# In IMAP mode, the first folder named spam is marked
+			local ($sf) = grep { lc($_->{'name'}) eq 'spam' } @rv;
+			if ($sf) {
+				$sf->{'spam'} = 1;
+				}
+			}
+		elsif (!$done{$file}) {
+			# Need to add
+			push(@rv, { 'name' => "Spam",
+				    'file' => $file,
+				    'type' => &folder_type($file),
+				    'perpage' => $userconfig{"perpage_$file"},
+				    'fromaddr' => $userconfig{"fromaddr_$file"},
+				    'sent' => $userconfig{"sent_$file"},
+				    'hide' => 0,
+				    'mode' => 0,
+				    'spam' => 1,
+				    'index' => scalar(@rv) } );
+			$done{$file}++;
+			}
+		else {
+			# Mark as spam folder
+			local ($sf) = grep { $_->{'file'} eq $file } @rv;
+			if ($sf) {
+				$sf->{'spam'} = 1;
+				}
+			}
+		}
+	}
+
 # Mark folders are using Notes mail encoding
 foreach $f (@rv) {
 	if ($f->{'file'} && $userconfig{"notes_".$f->{'file'}}) {
@@ -439,11 +597,40 @@ if ($folder->{'type'} == 2) {
 	}
 elsif ($folder->{'type'} == 4) {
 	# An IMAP folder
-	$folder->{'id'} ||= time();
+	local @exclude;
+	if ($folder->{'imapauto'}) {
+		# This type of folder needs to be really created or updated
+		# on the server
+		if (!$folder->{'id'}) {
+			# Need to create
+			local ($ok, $ih) = &imap_login($folder);
+			local @irv = &imap_command($ih,
+					"create \"$folder->{'name'}\"");
+			$irv[0] || &error($irv[2]);
+			$folder->{'id'} = $folder->{'mailbox'} =
+				$folder->{'name'};
+			}
+		elsif ($folder->{'mailbox'} ne $folder->{'name'}) {
+			# Need to rename
+			local ($ok, $ih) = &imap_login($folder);
+			local @irv = &imap_command($ih,
+				"rename \"$folder->{'mailbox'}\" \"$folder->{'name'}\"");
+			$irv[0] || &error($irv[2]);
+			$folder->{'id'} = $folder->{'name'};
+			$folder->{'id'} = $folder->{'mailbox'} =
+				$folder->{'name'};
+			}
+		@exclude = ( "type", "mode", "remote", "nowrite", "index",
+			     "id", "mailbox", "server", "user", "pass" );
+		}
+	else {
+		# Just save details of IMAP folder in a file
+		$folder->{'id'} ||= time();
+		@exclude = ( "type", "mode", "remote", "nowrite", "index" );
+		}
 	local %imap;
-	foreach $k (keys %$folder) {
-		if ($k ne "type" && $k ne "mode" && $k ne "remote" &&
-		    $k ne "nowrite" && $k ne "index") {
+	foreach my $k (keys %$folder) {
+		if (&indexof($k, @exclude) == -1) {
 			$imap{$k} = $folder->{$k};
 			}
 		}
@@ -497,6 +684,10 @@ elsif ($folder->{'type'} == 6) {
 elsif ($folder->{'mode'} == 0) {
 	# Updating a folder in ~/mail .. need to manage file, and config options
 	local $path = "$folders_dir/$folder->{'name'}";
+	if ($folders_dir eq "$remote_user_info[7]/Maildir") {
+		# Maildir sub-folder .. put a . in the name
+		$path =~ s/([^\/]+)$/.$1/;
+		}
 	if ($folder->{'name'} =~ /\//) {
 		local $pp = $path;
 		$pp =~ s/\/[^\/]+$//;
@@ -601,6 +792,13 @@ elsif ($folder->{'type'} == 4) {
 	# An IMAP folder
 	unlink("$user_module_config_directory/$folder->{'id'}.imap");
 	system("rm -rf $user_module_config_directory/$folder->{'id'}.cache");
+
+	if ($folder->{'imapauto'}) {
+		# Remove actual folder from IMAP server too
+		local ($ok, $ih) = &imap_login($folder);
+		local @irv = &imap_command($ih, "delete \"$folder->{'name'}\"");
+		$irv[0] || &error($irv[2]);
+		}
 	}
 elsif ($folder->{'type'} == 5) {
 	# A composite folder
@@ -743,10 +941,10 @@ printf "<input name=perpage size=5 value='%s'></td> </tr>\n",
 
 if ($_[1] != 2) {
 	print "<tr> <td><b>$text{'edit_sentview'}</b></td>\n";
-	printf "<td><input type=radio name=sent value=1 %s> %s\n",
-		$_[0]->{'sent'} ? "checked" : "", $text{'yes'};
-	printf "<input type=radio name=sent value=0 %s> %s</td> </tr>\n",
-		$_[0]->{'sent'} ? "" : "checked", $text{'no'};
+	printf "<td><input type=radio name=show_to value=1 %s> %s\n",
+		$_[0]->{'show_to'} ? "checked" : "", $text{'yes'};
+	printf "<input type=radio name=show_to value=0 %s> %s</td> </tr>\n",
+		$_[0]->{'show_to'} ? "" : "checked", $text{'no'};
 	}
 
 print "<tr> <td><b>$text{'edit_fromaddr'}</b></td>\n";
@@ -773,7 +971,8 @@ else {
 	delete($folder->{'perpage'});
 	}
 if ($mode != 2) {
-	$folder->{'sent'} = $in->{'sent'};
+	$folder->{'show_to'} = $in->{'show_to'};
+	$folder->{'show_from'} = !$in->{'show_to'};
 	}
 if (!$in->{'fromaddr_def'}) {
 	$in->{'fromaddr'} =~ /\S/ || &error($text{'save_efromaddr'});
@@ -1075,35 +1274,20 @@ if (!defined(%read)) {
 	}
 }
 
-# open_dbm_db(&hash, file, mode)
-# Attempts to open a DBM, first using SDBM_File, and then NDBM_File
-sub open_dbm_db
-{
-local ($hash, $file, $mode) = @_;
-eval "use SDBM_File";
-dbmopen(%$hash, $file, $mode);
-eval { $hash->{'1111111111'} = 'foo bar' };
-if ($@) {
-	dbmclose(%$hash);
-	eval "use NDBM_File";
-	dbmopen(%$hash, $file, $mode);
-	}
-}
-
 # spam_report_cmd()
 # Returns a command for reporting spam, or undef if none
 sub spam_report_cmd
 {
 local %sconfig = &foreign_config("spam");
 if ($config{'spam_report'} eq 'sa_learn') {
-	return &has_command($sconfig{'sa_learn'}) ? "$sconfig{'sa_learn'} --spam" : undef;
+	return &has_command($sconfig{'sa_learn'}) ? "$sconfig{'sa_learn'} --spam --mbox" : undef;
 	}
 elsif ($config{'spam_report'} eq 'spamassassin') {
 	return &has_command($sconfig{'spamassassin'}) ? "$sconfig{'spamassassin'} --r" : undef;
 	}
 else {
 	return &has_command($sconfig{'sa_learn'}) ?
-		"$sconfig{'sa_learn'} --spam" :
+		"$sconfig{'sa_learn'} --spam --mbox" :
 	       &has_command($sconfig{'spamassassin'}) ?
 		"$sconfig{'spamassassin'} --r" : undef;
 	}
@@ -1114,7 +1298,7 @@ else {
 sub ham_report_cmd
 {
 local %sconfig = &foreign_config("spam");
-return &has_command($sconfig{'sa_learn'}) ? "$sconfig{'sa_learn'} --ham" : undef;
+return &has_command($sconfig{'sa_learn'}) ? "$sconfig{'sa_learn'} --ham --mbox" : undef;
 }
 
 # can_report_spam(&folder)
@@ -1171,13 +1355,13 @@ elsif ($p == 2) {
 	}
 local $mid = $_[0]->{'header'}->{'message-id'};
 if (!$_[1]) {
-	# Show icon if read by user
+	# Show icon if special
 	if ($read{$mid} == 2) {
 		push(@rv, "<img src=images/special.gif>");
 		}
-	elsif ($read{$mid} == 1) {
-		push(@rv, "<img src=images/read.gif>");
-		}
+	#elsif ($read{$mid} == 1) {
+	#	push(@rv, "<img src=images/read.gif>");
+	#	}
 	}
 else {
 	# Show icons if DSNs received
@@ -1201,10 +1385,6 @@ sub show_mailbox_buttons
 local ($num, $folders, $folder, $mail) = @_;
 local $spacer = "&nbsp;\n";
 if (@$mail) {
-	print "<input type=submit name=delete value=\"$text{'mail_delete'}\" ",
-	      "onClick='return check_clicks(form)'>";
-	print $spacer;
-
 	print "<input type=submit name=mark$_[0] value=\"$text{'mail_mark'}\">";
 	print "<select name=mode$_[0]>\n";
 	print "<option value=1 checked>$text{'mail_mark1'}\n";
@@ -1219,8 +1399,7 @@ if (@$mail) {
 		}
 
 	if ($userconfig{'open_mode'}) {
-		print "<input type=submit name=forward value=\"$text{'mail_forward'}\" ",
-		      "onClick='args = \"\"; for(i=0; i<form.d.length; i++) { if (form.d[i].checked) { args += \"mailforward=\"+form.d[i].value+\"&\"; } } window.open(\"reply_mail.cgi?\"+args, \"compose\", \"toolbar=no,menubar=no,scrollbars=yes,width=1024,height=768\"); return false'>";
+		print "<input type=submit name=forward value=\"$text{'mail_forward'}\" onClick='args = \"folder=$folder->{'index'}\"; for(i=0; i<form.d.length; i++) { if (form.d[i].checked) { args += \"&mailforward=\"+escape(form.d[i].value); } } window.open(\"reply_mail.cgi?\"+args, \"compose\", \"toolbar=no,menubar=no,scrollbars=yes,width=1024,height=768\"); return false'>";
 		}
 	else {
 		# Forward button can just be a normal submit
@@ -1228,8 +1407,13 @@ if (@$mail) {
 		}
 	print $spacer;
 
+	print "<input type=submit name=delete value=\"$text{'mail_delete'}\" ",
+	      "onClick='return check_clicks(form)'>";
+	print $spacer;
+
 	if (&can_report_spam($folder) &&
-	    $userconfig{'spam_buttons'} =~ /list/) {
+	    $userconfig{'spam_buttons'} =~ /list/ ||
+	    $folder->{'spam'}) {
 		print "<input type=submit value=\"$text{'mail_black'}\" name=black>";
 		if ($userconfig{'spam_del'}) {
 			print "<input type=submit value=\"$text{'view_razordel'}\" name=razor>";
@@ -1241,7 +1425,8 @@ if (@$mail) {
 		}
 
 	if (&can_report_ham($folder) &&
-	    $userconfig{'ham_buttons'} =~ /list/) {
+	    $userconfig{'ham_buttons'} =~ /list/ ||
+	    $folder->{'spam'}) {
 		print "<input type=submit value=\"$text{'mail_white'}\" name=white>";
 		print "<input type=submit value=\"$text{'view_ham'}\" name=ham>";
 		print $spacer;
@@ -1252,7 +1437,7 @@ if (@$mail) {
 if ($userconfig{'open_mode'}) {
 	# Show mass open button
 	print "<input type=submit name=new value=\"$text{'mail_open'}\" ",
-	      "onClick='for(i=0; i<form.d.length; i++) { if (form.d[i].checked) { window.open(\"view_mail.cgi?folder=$folder->{'index'}&idx=\"+form.d[i].value, \"view\"+i, \"toolbar=no,menubar=no,scrollbars=yes,width=1024,height=768\"); } } return false'>";
+	      "onClick='for(i=0; i<form.d.length; i++) { if (form.d[i].checked) { window.open(\"view_mail.cgi?folder=$folder->{'index'}&idx=\"+escape(form.d[i].value), \"view\"+i, \"toolbar=no,menubar=no,scrollbars=yes,width=1024,height=768\"); } } return false'>";
 	print $spacer;
 	}
 
@@ -1476,11 +1661,12 @@ else {
 	}
 }
 
-# view_mail_link(&folder, index, start, from-to-text)
+# view_mail_link(&folder, index, start, from-to-text, message-id)
 sub view_mail_link
 {
-local ($folder, $idx, $start, $txt) = @_;
-local $url = "view_mail.cgi?start=$start&idx=$idx&folder=$folder->{'index'}";
+local ($folder, $idx, $start, $txt, $mid) = @_;
+local $qmid = &urlize($mid);
+local $url = "view_mail.cgi?start=$start&idx=$idx&folder=$folder->{'index'}&mid=$qmid";
 if ($userconfig{'open_mode'}) {
 	return "<a href='' onClick='window.open(\"$url\", \"viewmail\", \"toolbar=no,menubar=no,scrollbars=yes,width=1024,height=768\"); return false'>".
 	       &simplify_from($txt)."</a>";
@@ -1512,5 +1698,73 @@ else {
 	}
 }
 
+# messages_from_indexes(&folder, &pairs)
+# Given a list of strings in idx/message-id format, returns the actual emails
+sub messages_from_indexes
+{
+local ($folder, $both) = @_;
+local (@idxs, @mids, @rv);
+foreach my $d (@$both) {
+	local ($di, $dmid) = split(/\//, $d);
+	push(@idxs, $di);
+	push(@mids, $dmid);
+	}
+return ( ) if (!@idxs);
+local @mails = &mailbox_list_mails_sorted($idxs[0], $idxs[@idxs-1], $folder);
+for(my $i=0; $i<@idxs; $i++) {
+	local $mail = &find_message_by_index(\@mails, $folder, $idxs[$i], $mids[$i]);
+	push(@rv, $mail) if ($mail);
+	}
+return @rv;
+}
+
+# get_auto_schedule(&folder)
+# Returns the automatic schedule structure for the given folder
+sub get_auto_schedule
+{
+local ($folder) = @_;
+local $id = $folder->{'id'} || &urlize($folder->{'file'});
+local %rv;
+&read_file("$user_module_config_directory/$id.sched", \%rv) ||
+	return undef;
+return \%rv;
+}
+
+# save_auto_schedule(&folder, &sched)
+# Updates the automatic schedule structure for the given folder
+sub save_auto_schedule
+{
+local ($folder, $sched) = @_;
+local $id = $folder->{'id'} || &urlize($folder->{'file'});
+if ($sched) {
+	&write_file("$user_module_config_directory/$id.sched", $sched);
+	}
+else {
+	unlink("$user_module_config_directory/$id.sched");
+	}
+}
+
+# setup_auto_cron()
+# Creates the Cron job that runs auto.pl
+sub setup_auto_cron
+{
+&foreign_require("cron", "cron-lib.pl");
+local @jobs = &cron::list_cron_jobs();
+local ($job) = grep { $_->{'command'} eq $auto_cmd &&
+		      $_->{'user'} eq $remote_user } @jobs;
+if (!$job) {
+	$job = { 'command' => $auto_cmd,
+		 'active' => 1,
+		 'user' => $remote_user,
+		 'mins' => int(rand()*60),
+		 'hours' => '*',
+		 'days' => '*',
+		 'months' => '*',
+		 'weekdays' => '*' };
+	&cron::create_cron_job($job);
+	}
+&cron::create_wrapper($auto_cmd, $module_name, "auto.pl");
+}
+
 1;
 
