@@ -205,19 +205,24 @@ if ($charset) {
           "content=\"text/html; Charset=$charset\">\n";
     }
 if (@_ > 0) {
+    local $title;
     if ($gconfig{'sysinfo'} == 1) {
-        printf "<title>%s : %s on %s (%s %s)</title>\n",
+        $title = sprintf "%s : %s on %s (%s %s)\n",
             $_[0], $remote_user, &get_display_hostname(),
             $os_type, $os_version;
         }
     elsif ($gconfig{'sysinfo'} == 4) {
-        printf "<title>%s on %s (%s %s)</title>\n",
-            $_[0], $remote_user, &get_display_hostname(),
+        $title = sprintf "%s on %s (%s %s)\n",
+            $remote_user, &get_display_hostname(),
             $os_type, $os_version;
         }
     else {
-        print "<title>$_[0]</title>\n";
+        $title = $_[0];
         }
+    if ($gconfig{'showlogin'} && $remote_user) {
+        $title = $remote_user." : ".$title;
+        }
+    print "<title>$title</title>\n";
     print $_[7] if ($_[7]);
     if ($gconfig{'sysinfo'} == 0 && $remote_user) {
         print "<SCRIPT LANGUAGE=\"JavaScript\">\n";
@@ -255,7 +260,8 @@ if ($remote_user && @_ > 1) {
 	print qq~<table width="100%" border="0" cellspacing="0" cellpadding="0" background="$gconfig{'webprefix'}/images/top_bar/bg.jpg" height="32">
 	  <tr>
 	    <td width="4" nowrap><img src="$gconfig{'webprefix'}/images/top_bar/left.jpg" width="4" height="32"></td>
-	    <td width="100%" nowrap><a href="http://www.usermin.com/"><img src="$gconfig{'webprefix'}/usermin_logo.gif" width="99" height="32" border="0" alt="Webmin home page"></td>~;
+	    <td width="100%" nowrap><a href="http://www.usermin.com/"><img src="$gconfig{'webprefix'}/usermin_logo.gif" width="99" height="32" border="0" alt="Webmin home page"></td>
+            <td width="100%" nowrap><a href="http://www.opencountry.com/webmin/" target=_new><img src="$gconfig{'webprefix'}/images/top_bar/oc_logo.gif" width="268" height="32" border="0" alt="OpenCountry home page"></a></td>~;
         if ($gconfig{'feedback'}) {
 		print qq~<td><a href='$gconfig{'webprefix'}/feedback_form.cgi?module=$module_name'><img src=$gconfig{'webprefix'}/images/top_bar/feedback.jpg width=97 height=32 alt="$text{'main_feedback'}" border=0></a></td>~;
 		if (!$ENV{'LOCAL_USER'}) {
@@ -281,28 +287,8 @@ if (@_ > 1 && !$one && $remote_user && !$gconfig{'notabs'}) {
   </tr>
 </table>~;
 
-    &read_file("$config_directory/webmin.catnames", \%catnames);
-    foreach $m (@msc_modules) {
-        $c = $m->{'category'};
-        next if ($cats{$c});
-        if (defined($catnames{$c})) {
-            $cats{$c} = $catnames{$c};
-            }
-        elsif ($text{"category_$c"}) {
-            $cats{$c} = $text{"category_$c"};
-            }
-        else {
-            # try to get category name from module ..
-            local %mtext = &load_language($m->{'dir'});
-            if ($mtext{"category_$c"}) {
-                $cats{$c} = $mtext{"category_$c"};
-                }
-            else {
-                $c = $m->{'category'} = "";
-                $cats{$c} = $text{"category_$c"};
-                }
-            }
-        }
+    # Get categories
+    %cats = &list_categories(\@msc_modules);
     @cats = sort { $b cmp $a } keys %cats;
     $cats = @cats;
     $per = $cats ? 100.0 / $cats : 100;
@@ -441,8 +427,7 @@ if (@_ > 1 && (!$_[5] || $ENV{'HTTP_WEBMIN_SERVERS'})) {
 	if ($_[3]) {
 		local %access = &get_module_acl();
 		if (!$access{'noconfig'} && !$config{'noprefs'}) {
-			local $cprog = $user_module_config_directory ?
-					"uconfig.cgi" : "config.cgi";
+			local $cprog = "uconfig.cgi";
 			&tab_start; print "<a href=\"$gconfig{'webprefix'}/$cprog?$module_name\">",
 			      $text{'header_config'},"</a>"; &tab_end;
 			}
@@ -659,7 +644,12 @@ sub theme_error
 &header($text{'error'}, "");
 print "<p>\n";
 print "<h3>",($main::whatfailed ? "$main::whatfailed : " : ""),@_,"</h3>\n";
-&footer($ENV{'HTTP_REFERER'}, $text{'error_previous'});
+if ($ENV{'HTTP_REFERER'} && $main::completed_referers_check) {
+	&footer($ENV{'HTTP_REFERER'}, $text{'error_previous'});
+	}
+else {
+	&footer();
+	}
 }
 
 sub theme_ui_print_unbuffered_header
