@@ -3,7 +3,7 @@
 
 do '../web-lib.pl';
 &init_config();
-require '../ui-lib.pl';
+do '../ui-lib.pl';
 if ($config{'mysql_libs'}) {
 	$ENV{$gconfig{'ld_env'}} .= ':' if ($ENV{$gconfig{'ld_env'}});
 	$ENV{$gconfig{'ld_env'}} .= $config{'mysql_libs'};
@@ -22,13 +22,16 @@ if ($module_info{'usermin'}) {
 	            'bpath', '/',
 	            'buser', 'root',
 	            'edonly', 0 );
-	$mysql_login = $userconfig{'login'};
-	$mysql_pass = $userconfig{'pass'};
+	if ($config{'useident'} ne 'yes') {
+		$mysql_login = $userconfig{'login'};
+		$mysql_pass = $userconfig{'pass'};
+		}
 	if (open(VERSION, "$user_module_config_directory/version")) {
 		chop($mysql_version = <VERSION>);
 		close(VERSION);
 		}
 	$max_dbs = $userconfig{'max_dbs'};
+	$commands_file = "$user_module_config_directory/commands";
 	}
 else {
 	# Webmin connects to the database as the user specified in the per-user
@@ -48,6 +51,7 @@ else {
 		}
 	$cron_cmd = "$module_config_directory/backup.pl";
 	$max_dbs = $config{'max_dbs'};
+	$commands_file = "$module_config_directory/commands";
 	}
 $authstr = &make_authstr();
 $master_db = 'mysql';
@@ -59,7 +63,7 @@ $password_func = $config{'passwd_mode'} ? "old_password" : "password";
 	    'blob', 'text', 'mediumblob', 'mediumtext', 'longblob', 'longtext',
 	    'enum', 'set');
 
-@priv_cols = ('Host', 'User', 'Password', 'Select_priv', 'Insert_priv', 'Update_priv', 'Delete_priv', 'Create_priv', 'Drop_priv', 'Reload_priv', 'Shutdown_priv', 'Process_priv', 'File_priv', 'Grant_priv', 'References_priv', 'Index_priv', 'Alter_priv', 'Show_db_priv', 'Super_priv', 'Create_tmp_table_priv', 'Lock_tables_priv', 'Execute_priv', 'Repl_slave_priv', 'Repl_client_priv');
+@priv_cols = ('Host', 'User', 'Password', 'Select_priv', 'Insert_priv', 'Update_priv', 'Delete_priv', 'Create_priv', 'Drop_priv', 'Reload_priv', 'Shutdown_priv', 'Process_priv', 'File_priv', 'Grant_priv', 'References_priv', 'Index_priv', 'Alter_priv', 'Show_db_priv', 'Super_priv', 'Create_tmp_table_priv', 'Lock_tables_priv', 'Execute_priv', 'Repl_slave_priv', 'Repl_client_priv', 'Create_view_priv', 'Show_view_priv', 'Create_routine_priv', 'Alter_routine_priv', 'Create_user_priv');
 
 if (!$config{'nodbi'}) {
 	# Check if we have DBI::mysql
@@ -69,10 +73,14 @@ use DBI;
 EOF
 }
 
-
 $old_user_priv_cols = $mysql_version >= 4 ? 21 : 14;
 $old_db_priv_cols = $mysql_version >= 4 ? 12 : 10;
 
+@mysql_set_variables = ( "key_buffer", "max_allowed_packet",
+			 "sort_buffer", "net_buffer_length",
+			 "myisam_sort_buffer_size" );
+@mysql_number_variables = ( "table_cache", "max_connections" );
+
 # make_authstr([login], [pass], [host], [port], [sock])
 sub make_authstr
 {
@@ -85,17 +93,21 @@ return ($sock ? " -S $sock" : "").
        ($host ? " -h $host" : "").
        ($port ? " -P $port" : "").
        ($login ? " -u ".quotemeta($login) : "").
-       ($pass ? " -p".quotemeta($pass) : "");
+       ($pass && $mysql_version >= 4.1 ? " --password=".quotemeta($pass) :
+        $pass ? " -p".quotemeta($pass) : "");
 }
 
 # is_mysql_running()
 # Returns 1 if mysql is running, 0 if not, or -1 if running but
-# inaccessible without a password
+# inaccessible without a password. When called in an array context, also
+# returns the full error message
 sub is_mysql_running
 {
 local $out = `"$config{'mysqladmin'}" $authstr status 2>&1`;
-return $out =~ /uptime/i ? 1 :
-       $out =~ /denied|password/i ? -1 : 0;
+local $rv = $out =~ /uptime/i ? 1 :
+            $out =~ /denied|password/i ? -1 : 0;
+$out =~ s/^.*\Q$config{'mysqladmin'}\E\s*:\s*//;
+return wantarray ? ($rv, $out) : $rv;
 }
 
 # list_databases()
@@ -131,7 +143,12 @@ if ($t =~ /access denied/i) {
 elsif (!ref($t)) {
 	&error("<tt>$t</tt>");
 	}
-return map { $_->[0] } @{$t->{'data'}};
+# Filter out views
+local %views;
+if (&supports_views()) {
+	%views = map { $_, 1 } &list_views($_[0]);
+	}
+return grep { !$views{$_} } map { $_->[0] } @{$t->{'data'}};
 }
 
 # table_structure(database, table)
@@ -362,24 +379,30 @@ local $mid = int((@{$_[1]}+1) / 2);
 local ($i, $j);
 print "<table width=100%><tr>\n";
 foreach $s ([0, $mid-1], [$mid, @{$_[1]}-1]) {
-	print "<td width=50% valign=top><table border width=100%>\n";
+	print "<td width=50% valign=top>\n";
+
+	# Header
+	local @tds = ( "width=5" );
 	if ($s->[0] <= $s->[1]) {
-		print "<tr $tb> ";
+		local @hcols;
 		foreach $t (@{$_[0]}) {
-			print "<td><b>$t</b></td> ";
+			push(@hcols, $t);
 			}
-		print "</tr>\n";
+		print &ui_columns_start(\@hcols, 100, 0, \@tds);
 		}
+
 	for($i=$s->[0]; $i<=$s->[1]; $i++) {
-		print "<tr $cb>\n";
-		print "<td width=5%>$_[1]->[$i]</td>\n";
-		print "<td><a href='$_[2]->[$i]'>$_[3]->[$i]</a></td>\n";
+		local @cols;
+		push(@cols, "<a href='$_[2]->[$i]'>$_[3]->[$i]</a>");
 		for($j=4; $j<@_; $j++) {
-			print "<td>$_[$j]->[$i]</td>\n";
+			push(@cols, $_[$j]->[$i]);
 			}
-		print "</tr>\n";
+		print &ui_checked_columns_row(\@cols, \@tds, "d", $_[1]->[$i]);
 		}
-	print "</table></td>\n";
+	if ($s->[0] <= $s->[1]) {
+		print &ui_columns_end();
+		}
+	print "</td>\n";
 	}
 print "</tr></table>\n";
 }
@@ -460,6 +483,39 @@ else {
 	}
 }
 
+# supports_views()
+# Returns 1 if this MySQL install supports views
+sub supports_views
+{
+if (!defined($supports_views_cache)) {
+	if ($mysql_version < 5) {
+		$supports_views_cache = 0;
+		}
+	else {
+		local @dbs = &list_databases();
+		$supports_views_cache =
+			&indexof("information_schema", @dbs) >= 0 ? 1 : 0;
+		}
+	}
+return $supports_views_cache;
+}
+
+# supports_variables()
+# Returns 1 if running mysql version 4.0.3 or later
+sub supports_variables
+{
+if ($mysql_version =~ /^(\d+)\.(\d+)\.(\d+)$/ &&
+    ($1 > 4 || ($1 == 4 && $2 > 0) || ($1 == 4 && $2 == 0 && $3 >= 3))) {
+	return 1;
+	}
+elsif ($mysql_version > 4) {
+	return 1;
+	}
+else {
+	return 0;
+	}
+}
+
 # user_priv_cols()
 # Returns the number of columns used for privileges in the user table
 sub user_priv_cols
@@ -488,6 +544,20 @@ if (!$db_priv_cols) {
 return $db_priv_cols;
 }
 
+# host_priv_cols()
+# Returns the number of columns used for privileges in the db table
+sub host_priv_cols
+{
+if (!$host_priv_cols) {
+	local @str = &table_structure("mysql", "host");
+	local $s;
+	foreach $s (@str) {
+		$host_priv_cols++ if ($s->{'field'} =~ /_priv/i);
+		}
+	}
+return $host_priv_cols;
+}
+
 sub is_blob
 {
 return $_[0]->{'type'} =~ /(text|blob)$/i;
@@ -527,6 +597,8 @@ else {
 }
 
 # execute_before(db, handle, escape, path, db-for-config)
+# Executes the before-backup command for some DB, and sends output to the
+# given file handle. Returns 1 if the command suceeds, or 0 on failure
 sub execute_before
 {
 local $cmd = $config{'backup_before_'.$_[4]};
@@ -534,11 +606,13 @@ if ($cmd) {
 	$ENV{'BACKUP_FILE'} = $_[3];
 	local $h = $_[1];
 	local $out;
-	&execute_command($cmd, undef, \$out, \$out);
+	local $rv = &execute_command($cmd, undef, \$out, \$out);
 	if ($h && $out) {
 		print $h $_[2] ? "<pre>".&html_escape($out)."</pre>" : $out;
 		}
+	return !$rv;
 	}
+return 1;
 }
 
 # execute_after(db, handle, escape, path, db-for-config)
@@ -549,11 +623,13 @@ if ($cmd) {
 	$ENV{'BACKUP_FILE'} = $_[3];
 	local $h = $_[1];
 	local $out;
-	&execute_command($cmd, undef, \$out, \$out);
+	local $rv = &execute_command($cmd, undef, \$out, \$out);
 	if ($h && $out) {
 		print $h $_[2] ? "<pre>".&html_escape($out)."</pre>" : $out;
 		}
+	return !$rv;
 	}
+return 1;
 }
 
 # show_table_form(count)
@@ -704,5 +780,382 @@ if ($? || $out =~ /failed/i) {
 return undef;
 }
 
+# split_enum(type)
+# Returns a list of allowed values for an enum
+sub split_enum
+{
+local ($type) = @_;
+if ($type =~ /^(enum|set)\((.*)\)$/) {
+	$type = $2;
+	}
+local $esize = $type;
+local @ev;
+while($esize =~ /^'([^']*)'(,?)(.*)$/) {
+	push(@ev, $1);
+	$esize = $3;
+	}
+return @ev;
+}
+
+# Returns 1 if the mysql server being managed is on this system
+sub is_mysql_local
+{
+return $config{'host'} eq '' || $config{'host'} eq 'localhost' ||
+       $config{'host'} eq &get_system_hostname() ||
+       &to_ipaddress($config{'host'}) eq &to_ipaddress(&get_system_hostname());
+}
+
+# get_mysql_config()
+# Returns the parsed my.cnf file
+sub get_mysql_config
+{
+if (!defined(@mysql_config_cache)) {
+	local $sect;
+	local $lnum = 0;
+	open(CONF, $config{'my_cnf'}) || return undef;
+	while(<CONF>) {
+		s/\r|\n//g;
+		s/#.*$//;
+		s/\s+$//;
+		if (/^\s*\[(\S+)\]$/) {
+			# Start of a section
+			$sect = { 'name' => $1,
+				  'members' => [ ],
+				  'line' => $lnum,
+				  'eline' => $lnum };
+			push(@mysql_config_cache, $sect);
+			}
+		elsif (/^\s*(\S+)\s*=\s*(.*)$/ && $sect) {
+			# Variable in a section
+			push(@{$sect->{'members'}},
+			     { 'name' => $1,
+			       'value' => $2,
+			       'line' => $lnum });
+			$sect->{'eline'} = $lnum;
+			}
+		elsif (/^\s*(\S+)$/ && $sect) {
+			# Single directive in a section
+			push(@{$sect->{'members'}},
+			     { 'name' => $1,
+			       'line' => $lnum });
+			$sect->{'eline'} = $lnum;
+			}
+		$lnum++;
+		}
+	close(CONF);
+	}
+return \@mysql_config_cache;
+}
+
+# find(name, &conf)
+sub find
+{
+local ($name, $conf) = @_;
+local @rv = grep { lc($_->{'name'}) eq lc($name) } @$conf;
+return wantarray ? @rv : $rv[0];
+}
+
+# find_value(name, &conf)
+sub find_value
+{
+local ($name, $conf) = @_;
+local @rv = map { $_->{'value'} } &find($name, $conf);
+return wantarray ? @rv : $rv[0];
+}
+
+# save_directive(&conf, &section, name, &values)
+# Updates one or multiple lines in a my.cnf section
+sub save_directive
+{
+local ($conf, $sect, $name, $values) = @_;
+local @old = &find($name, $sect->{'members'});
+local $lref = &read_file_lines($config{'my_cnf'});
+
+for(my $i=0; $i<@old || $i<@$values; $i++) {
+	local $old = $old[$i];
+	local $line = $values->[$i] eq "" ? $name :
+			"$name = $values->[$i]";
+	if ($old && defined($values->[$i])) {
+		# Updating
+		$lref->[$old->{'line'}] = $line;
+		$old->{'value'} = $values->[$i];
+		}
+	elsif (!$old && defined($values->[$i])) {
+		# Adding
+		splice(@$lref, $sect->{'eline'}+1, 0, $line);
+		&renumber($conf, $sect->{'eline'}, 1);
+		push(@{$sect->{'members'}},
+			{ 'name' => $name,
+			  'value' => $values->[$i],
+			  'line' => $sect->{'eline'}+1 });
+		}
+	elsif ($old && !defined($values->[$i])) {
+		# Deleting
+		splice(@$lref, $old->{'line'}, 1);
+		&renumber($conf, $old->{'line'}, -1);
+		@{$sect->{'members'}} = grep { $_ ne $old }
+					     @{$sect->{'members'}};
+		}
+	}
+}
+
+sub renumber
+{
+local ($conf, $line, $offset) = @_;
+foreach my $sect (@$conf) {
+	$sect->{'line'} += $offset if ($sect->{'line'} >= $line);
+	$sect->{'eline'} += $offset if ($sect->{'eline'} >= $line);
+	foreach my $m (@{$sect->{'members'}}) {
+		$m->{'line'} += $offset if ($m->{'line'} >= $line);
+		}
+	}
+}
+
+# parse_set_variables(value, ...)
+# Returns a hash of variable mappings
+sub parse_set_variables
+{
+local %vars;
+foreach my $v (@_) {
+	if ($v =~ /^(\S+)=(\S+)$/) {
+		$vars{$1} = $2;
+		}
+	}
+return %vars;
+}
+
+sub mysql_size_input
+{
+local ($name, $value) = @_;
+local $units = "b";
+if ($value =~ /^(\d+)([a-z])$/i) {
+	$value = $1;
+	$units = $2;
+	}
+return &ui_textbox($name, $value, 8)."\n".
+       &ui_select($name."_units", $units,
+		  [ [ "b", "bytes" ], [ "K", "kB" ],
+		    [ "M", "MB" ], [ "G", "GB" ] ]);
+}
+
+# list_indexes(db)
+# Returns the names of all indexes in some database
+sub list_indexes
+{
+local ($db) = @_;
+local (@rv, $r);
+foreach my $table (&list_tables($db)) {
+	local $s = &execute_sql_safe($db, "show index from ".&quotestr($table));
+	local (%tp, $i);
+	for($i=0; $i<@{$s->{'titles'}}; $i++) {
+		$tp{lc($s->{'titles'}->[$i])} = $i;
+		}
+	foreach $r (@{$s->{'data'}}) {
+		if ($r->[$tp{'key_name'}] ne 'PRIMARY') {
+			push(@rv, $r->[$tp{'key_name'}]);
+			}
+		}
+	}
+return &unique(@rv);
+}
+
+# index_structure(db, indexname)
+# Returns information on an index
+sub index_structure
+{
+local ($db, $index) = @_;
+local (@rv, $r);
+local $info;
+foreach my $table (&list_tables($db)) {
+	local $s = &execute_sql_safe($db, "show index from ".&quotestr($table));
+	local (%tp, $i);
+	for($i=0; $i<@{$s->{'titles'}}; $i++) {
+		$tp{lc($s->{'titles'}->[$i])} = $i;
+		}
+	foreach $r (@{$s->{'data'}}) {
+		if ($r->[$tp{'key_name'}] eq $index) {
+			# Found some info
+			$info->{'table'} = $r->[$tp{'table'}];
+			$info->{'name'} = $index;
+			$info->{'type'} = lc($r->[$tp{'index_type'}]) ||
+					  lc($r->[$tp{'comment'}]);
+			push(@{$info->{'cols'}}, $r->[$tp{'column_name'}]);
+			}
+		}
+	}
+return $info;
+}
+
+# list_views(db)
+# Returns the names of all views in some database
+sub list_views
+{
+local ($db) = @_;
+local @rv;
+local $d = &execute_sql($db, "select table_schema,table_name from information_schema.views");
+foreach $r (@{$d->{'data'}}) {
+	push(@rv, $r->[1]);
+	}
+return @rv;
+}
+
+# view_structure(db, viewname)
+# Returns information about a view
+sub view_structure
+{
+local ($db, $view) = @_;
+local $info = { 'name' => $view };
+local $d = &execute_sql($db, "show create view $view");
+local $c = $d->{'data'}->[0]->[1];
+if ($c =~ /algorithm\s*=\s*(\S+)/i) {
+	$info->{'algorithm'} = lc($1);
+	}
+if ($c =~ /definer\s*=\s*`(\S+)`\@`(\S+)`/i) {
+	$info->{'definer'} = "$1\@$2";
+	}
+elsif ($c =~ /definer\s*=\s*(\S+)/i) {
+	$info->{'definer'} = $1;
+	}
+if ($c =~ /sql\s+security\s+(\S+)/i) {
+	$info->{'security'} = lc($1);
+	}
+if ($c =~ s/\s+with\s+(cascaded|local)\s+check\s+option//i) {
+	$info->{'check'} = lc($1);
+	}
+if ($c =~ /view\s+(`\S+`|\S+)\s+as\s+(.*)/i) {
+	$info->{'query'} = $2;
+	}
+return $info;
+}
+
+# list_character_sets([db])
+# Returns a list of supported character sets. Each row is an array ref of
+# a code and name
+sub list_character_sets
+{
+local @rv;
+local $db = $_[0] || $master_db;
+if ($mysql_version < 4.1) {
+	local $d = &execute_sql($db, "show variables like 'character_sets'");
+	@rv = map { [ $_, $_ ] } split(/\s+/, $d->{'data'}->[0]->[1]);
+	}
+else {
+	local $d = &execute_sql($db, "show character set");
+	@rv = map { [ $_->[0], $_->[1] ] } @{$d->{'data'}};
+	}
+return sort { lc($a) cmp lc($b) } @rv;
+}
+
+# list_system_variables()
+# Returns a list of all system variables, and their default values
+sub list_system_variables
+{
+local $mysqld = $config{'mysqld'};
+if (!$mysqld) {
+	# Mysqld path not in config .. guess from mysql path
+	$mysqld = $config{'mysql'};
+	$mysqld =~ s/mysql$/mysqld/g;
+	$mysqld =~ s/bin/sbin/g;
+	if (!-x $mysqld) {
+		$mysqld = $config{'mysql'};
+		$mysqld =~ s/mysql$/mysqld/g;
+		$mysqld =~ s/bin/libexec/g;
+		if (!-x $mysqld) {
+			# Look in Webmin path
+			&error($mysqld);
+			$mysqld = &has_command("mysqld");
+			}
+		}
+	}
+return ( ) if (!$mysqld);
+
+# Read supported variables
+local @rv;
+&open_execute_command(MYSQLD, "$mysqld --verbose --help", 1, 1);
+while(<MYSQLD>) {
+	s/\r|\n//g;
+	if (/^(\S+)\s+current\s+value:\s+(\S*)/) {
+		push(@rv, [ $1, $2 ]);
+		}
+	elsif (/^\-\-\-\-/) {
+		$started = 1;
+		}
+	elsif ($started && /^(\S+)\s+(.*)/) {
+		push(@rv, [ $1, $2 eq "(No default value)" ? undef : $2 ]);
+		}
+	}
+close(MYSQL);
+return @rv;
+}
+
+# list_compatible_formats()
+# Returns a list of two-element arrays, containing compatability format
+# codes and descriptions
+sub list_compatible_formats
+{
+return map { [ $_, $text{'compat_'.$_} ] }
+	   ( "ansi", "mysql323", "mysql40", "postgresql", "oracle", "mssql",
+	     "db2", "maxdb" );
+}
+
+# list_compatible_options()
+# Returns a list of two-element arrays, containing compatability options
+sub list_compatible_options
+{
+return map { [ $_, $text{'compat_'.$_} ] }
+	   ( "no_key_options", "no_table_options", "no_field_options" );
+}
+
+# compression_format(file)
+# Returns 0 if uncompressed, 1 for gzip, 2 for compress, 3 for bzip2 or
+# 4 for zip
+sub compression_format
+{
+open(BACKUP, $_[0]);
+local $two;
+read(BACKUP, $two, 2);
+close(BACKUP);
+return $two eq "\037\213" ? 1 :
+       $two eq "\037\235" ? 2 :
+       $two eq "PK" ? 4 :
+       $two eq "BZ" ? 3 : 0;
+}
+
+# backup_database(db, dest-file, compress-mode, drop-flag, where-clause,
+#                 charset, &compatible, &only-tables, run-as-user)
+# Backs up a database to the given file, optionally with compression. Returns
+# undef on success, or an error message on failure.
+sub backup_database
+{
+local ($db, $file, $compress, $drop, $where, $charset, $compatible,
+       $tables, $user) = @_;
+if ($compress == 0) {
+	$writer = ">$file";
+	}
+elsif ($compress == 1) {
+	$writer = "| gzip -c >$file";
+	}
+elsif ($compress == 2) {
+	$writer = "| bzip2 -c >$file";
+	}
+local $dropsql = $drop ? "--add-drop-table" : "";
+local $wheresql = $where ? "\"--where=$in{'where'}\"" : "";
+local $charsetsql = $charset ?
+	"--default-character-set=".quotemeta($charset) : "";
+local $compatiblesql = @$compatible ?
+	"--compatible=".join(",", @$compatible) : "";
+local $quotingsql = &supports_quoting() ? "--quote-names" : "";
+local $tablessql = join(" ", map { quotemeta($_) } @$tables);
+local $cmd = "$config{'mysqldump'} $authstr $dropsql $wheresql $charsetsql $compatiblesql $quotingsql ".quotemeta($db)." $tablessql 2>&1 $writer";
+if ($user && $user ne "root") {
+	$cmd = &command_as_user($user, undef, $cmd);
+	}
+local $out = &backquote_logged("$cmd 2>&1");
+if ($? || $out) {
+	return $out;
+	}
+return undef;
+}
+
 1;
 
