@@ -24,7 +24,9 @@ if ($module_info{'usermin'}) {
 	$postgres_pass = $userconfig{'pass'};
 	$postgres_sameunix = 0;
 	%access = ( 'backup' => 1,
-		    'restore' => 1 );
+		    'restore' => 1,
+		    'tables' => 1,
+		    'cmds' => 1, );
 	$max_dbs = $userconfig{'max_dbs'};
 	}
 else {
@@ -54,7 +56,8 @@ EOF
 
 # is_postgresql_running()
 # Returns 1 if yes, 0 if no, -1 if the login is invalid, -2 if there
-# is a library problem
+# is a library problem. When called in an array context, returns the full error
+# message too.
 sub is_postgresql_running
 {
 local $temp = &transname();
@@ -86,39 +89,47 @@ open(OUT, "$cmd 2>&1 |");
 while(<OUT>) { $out .= $_; }
 close(OUT);
 unlink($temp);
+local $rv;
 if ($out =~ /setuserid:/i || $out =~ /no\s+password\s+supplied/i ||
     $out =~ /no\s+postgres\s+username/i || $out =~ /authentication\s+failed/i ||
     $out =~ /password:.*password:/i || $out =~ /database.*does.*not/i ||
     $out =~ /user.*does.*not/i) {
-	return -1;
+	$rv = -1;
 	}
 elsif ($out =~ /connect.*failed/i || $out =~ /could not connect to server:/) {
-	return 0;
+	$rv = 0;
 	}
 elsif ($out =~ /lib\S+\.so/i) {
-	return -2;
+	$rv = -2;
 	}
 else {
-	return 1;
+	$rv = 1;
 	}
+return wantarray ? ($rv, $out) : $rv;
 }
 
-# get_postgresql_version()
+# get_postgresql_version([from-command])
 sub get_postgresql_version
 {
+local ($fromcmd) = @_;
 local $main::error_must_die = 1;
+return $postgresql_version_cache if (defined($postgresql_version_cache));
 local $rv;
-eval {
-	local $v = &execute_sql_safe($config{'basedb'}, 'select version()');
-	$v = $v->{'data'}->[0]->[0];
-	if ($v =~ /postgresql\s+([0-9\.]+)/i) {
-		$rv = $1;
-		}
-	};
+if (!$fromcmd) {
+	eval {
+		local $v = &execute_sql_safe($config{'basedb'},
+					     'select version()');
+		$v = $v->{'data'}->[0]->[0];
+		if ($v =~ /postgresql\s+([0-9\.]+)/i) {
+			$rv = $1;
+			}
+		};
+	}
 if (!$rv || $@) {
 	local $out = `$config{'psql'} -V 2>&1 </dev/null`;
-	$rv = $out =~ /\s([0-9\.]+)$/ ? $1 : undef;
+	$rv = $out =~ /\s([0-9\.]+)/ ? $1 : undef;
 	}
+$postgresql_version_cache = $rv;
 return $rv;
 }
 
@@ -199,11 +210,27 @@ if (&supports_schemas($_[0])) {
 						  : $r->[2]."($sz)",
 			    'null' => $r->[5] =~ /f|0/ ? 'YES' : 'NO' } );
 		}
+
+	# Work out which fields are the primary key
+	if (&supports_indexes()) {
+		local ($keyidx) = grep { $_ eq $_[1]."_pkey" }
+				       &list_indexes($_[0]);
+		if ($keyidx) {
+			local $istr = &index_structure($_[0], $keyidx);
+			foreach my $r (@rv) {
+				if (&indexof($r->{'field'},
+					     @{$istr->{'cols'}}) >= 0) {
+					$r->{'key'} = 'PRI';
+					}
+				}
+			}
+		}
+
 	return @rv;
 	}
 else {
 	# Just look by table name
-	local $t = &execute_sql_safe($_[0], "select a.attnum, a.attname, t.typname, a.attlen, a.atttypmod, a.attnotnull, a.atthasdef FROM pg_class c, pg_attribute a, pg_type t WHERE c.relname = '$_[1]'    and a.attnum > 0     and a.attrelid = c.oid     and a.atttypid = t.oid order by attnum");
+	local $t = &execute_sql_safe($_[0], "select a.attnum, a.attname, t.typname, a.attlen, a.atttypmod, a.attnotnull, a.atthasdef FROM pg_class c, pg_attribute a, pg_type t WHERE c.relname = '$_[1]' and a.attnum > 0 and a.attrelid = c.oid     and a.atttypid = t.oid order by attnum");
 	local (@rv, $r);
 	foreach $r (@{$t->{'data'}}) {
 		local $arr;
@@ -351,17 +378,18 @@ else {
 		}
 
 	# Call the psql program
-	local $temp = &transname();
-	open(TEMP, ">$temp");
-	print TEMP "$postgres_login\n$postgres_pass\n";
-	close(TEMP);
 	local $host = $config{'host'} ? "-h $config{'host'}" : "";
 	$host .= " -p $config{'port'}" if ($config{'port'});
-	local $cmd = "$config{'psql'} -u -c ".quotemeta($sql)." $host $_[0]";
+	local $cmd = $config{'psql'}.
+		     (!&supports_pgpass() ? " -u" : " -U $postgres_login").
+		     " -c ".quotemeta($sql)." $host $_[0]";
 	if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
 		$cmd = "su $postgres_login -c ".quotemeta($cmd);
 		}
+	$cmd = &command_with_login($cmd);
 
+	delete($ENV{'LANG'});		# to force output to english
+	delete($ENV{'LANGUAGE'});
         if ( $break_f == 0 ) {
 		# Running a normal SQL command, not one with a \
 		#$ENV{'PAGER'} = "cat";
@@ -373,7 +401,7 @@ else {
 				&proc::close_controlling_pty();
 				}
 			}
-		open(OUT, "$cmd <$temp 2>&1 |");
+		open(OUT, "$cmd 2>&1 |");
 		local ($line, $rv, @data);
 		do {
 			$line = <OUT>;
@@ -386,7 +414,8 @@ else {
 			}
 		elsif (!defined($line)) {
 			# Un-expected end of output ..
-			&error(&text('esql', "<tt>$sql</tt>", "<tt>$config{'psql'} failed</tt>"));
+			&error(&text('esql', "<tt>$sql</tt>",
+				     "<tt>$config{'psql'} failed</tt>"));
 			}
 		else {
 			local $dash = <OUT>;
@@ -436,7 +465,7 @@ else {
 		local ( @titles, @row, @data, $rc, $emsgf, $emsg ) ;
 
 		$emsgf = &transname();
-		$rc = &system_logged ( "$cmd < $temp >$emsgf 2>&1");
+		$rc = &system_logged ( "$cmd >$emsgf 2>&1");
 		$emsg  = `cat $emsgf` ;
 		unlink ( $emsgf ) ;
 		if ($rc) {
@@ -463,20 +492,39 @@ return &execute_sql(@_);
 sub can_edit_db
 {
 if ($module_info{'usermin'}) {
-	foreach $l (split(/\t/, $config{'access'})) {
+	# Check access control list in configuration
+	local $rv;
+	DB: foreach $l (split(/\t/, $config{'access'})) {
 		if ($l =~ /^(\S+):\s*(.*)$/ &&
 		    ($1 eq $remote_user || $1 eq '*')) {
 			local @dbs = split(/\s+/, $2);
 			foreach $d (@dbs) {
 				$d =~ s/\$REMOTE_USER/$remote_user/g;
-				return 1 if ($d eq '*' || $_[0] =~ /^$d$/);
+				if ($d eq '*' || $_[0] =~ /^$d$/) {
+					$rv = 1;
+					last DB;
+					}
 				}
-			return 0;
+			$rv = 0;
+			last DB;
 			}
 		}
-	return 0;
+	if ($rv && $config{'access_own'}) {
+		# Check ownership on DB - first get login ID
+		if (!defined($postgres_login_id)) {
+			local $d = &execute_sql($config{'basedb'}, "select usesysid from pg_user where usename = ?", $postgres_login);
+			$postgres_login_id = $d->{'data'}->[0]->[0];
+			}
+		# Get database owner
+		local $d = &execute_sql($config{'basedb'}, "select datdba from pg_database where datname = ?", $_[0]);
+		if ($d->{'data'}->[0]->[0] != $postgres_login_id) {
+			$rv = 0;
+			}
+		}
+	return $rv;
 	}
 else {
+	# Check Webmin ACL
 	local $d;
 	return 1 if ($access{'dbs'} eq '*');
 	foreach $d (split(/\s+/, $access{'dbs'})) {
@@ -674,11 +722,13 @@ if ($cmd) {
 	$ENV{'BACKUP_FILE'} = $_[3];
 	local $h = $_[1];
 	local $out;
-	&execute_command($cmd, undef, \$out, \$out);
+	local $rv = &execute_command($cmd, undef, \$out, \$out);
 	if ($h && $out) {
 		print $h $_[2] ? "<pre>".&html_escape($out)."</pre>" : $out;
 		}
+	return !$rv;
 	}
+return 1;
 }
 
 # execute_after(db, handle, escape, path, db-for-config)
@@ -689,11 +739,13 @@ if ($cmd) {
 	$ENV{'BACKUP_FILE'} = $_[3];
 	local $h = $_[1];
 	local $out;
-	&execute_command($cmd, undef, \$out, \$out);
+	local $rv = &execute_command($cmd, undef, \$out, \$out);
 	if ($h && $out) {
 		print $h $_[2] ? "<pre>".&html_escape($out)."</pre>" : $out;
 		}
+	return !$rv;
 	}
+return 1;
 }
 
 sub quote_table
@@ -715,24 +767,17 @@ if (&is_readonly_mode()) {
 	return (0, undef);
 	}
 local ($db, $file, $user, $pass) = @_;
-local $cmd = "$config{'psql'} -u -f $file".
+if (!defined($user)) {
+	$user = $postgres_login;
+	$pass = $postgres_pass;
+	}
+local $cmd = "$config{'psql'} -f $file".
+	     (&supports_pgpass() ? " -U $user" : " -u").
 	     ($config{'host'} ? " -h $config{'host'}" : "").
 	     ($config{'port'} ? " -h $config{'port'}" : "").
 	     " $db";
-local $ltemp = &transname();
-open(TEMP, ">$ltemp");
-if ($user) {
-	print TEMP "$user\n$pass\n";
-	}
-else {
-	print TEMP "$postgres_login\n$postgres_pass\n";
-	}
-close(TEMP);
-if ($postgres_sameunix && defined(getpwnam($postgres_login)) && !$user) {
-	$cmd = "su $postgres_login -c ".quotemeta($cmd);
-	}
-local $out = &backquote_logged("$cmd <$ltemp 2>&1");
-unlink($ltemp);
+$cmd = &command_with_login($cmd, $user, $pass);
+local $out = &backquote_logged("$cmd 2>&1");
 return ($out =~ /ERROR/i ? 1 : 0, $out);
 }
 
@@ -744,24 +789,30 @@ local $mid = int((@{$_[1]}+1) / 2);
 local ($i, $j);
 print "<table width=100%><tr>\n";
 foreach $s ([0, $mid-1], [$mid, @{$_[1]}-1]) {
-	print "<td width=50% valign=top><table border width=100%>\n";
+	print "<td width=50% valign=top>\n";
+
+	# Header
+	local @tds = ( "width=5" );
 	if ($s->[0] <= $s->[1]) {
-		print "<tr $tb> ";
+		local @hcols;
 		foreach $t (@{$_[0]}) {
-			print "<td><b>$t</b></td> ";
+			push(@hcols, $t);
 			}
-		print "</tr>\n";
+		print &ui_columns_start(\@hcols, 100, 0, \@tds);
 		}
+
 	for($i=$s->[0]; $i<=$s->[1]; $i++) {
-		print "<tr $cb>\n";
-		print "<td width=5%>$_[1]->[$i]</td>\n";
-		print "<td><a href='$_[2]->[$i]'>$_[3]->[$i]</a></td>\n";
+		local @cols;
+		push(@cols, "<a href='$_[2]->[$i]'>$_[3]->[$i]</a>");
 		for($j=4; $j<@_; $j++) {
-			print "<td>$_[$j]->[$i]</td>\n";
+			push(@cols, $_[$j]->[$i]);
 			}
-		print "</tr>\n";
+		print &ui_checked_columns_row(\@cols, \@tds, "d", $_[1]->[$i]);
 		}
-	print "</table></td>\n";
+	if ($s->[0] <= $s->[1]) {
+		print &ui_columns_end();
+		}
+	print "</td>\n";
 	}
 print "</tr></table>\n";
 }
@@ -830,6 +881,7 @@ return undef;
 # or undef on success.
 sub setup_postgresql
 {
+return undef if (!$config{'setup_cmd'});
 local $temp = &transname();
 local $rv = &system_logged("($config{'setup_cmd'}) >$temp 2>&1");
 local $out = `cat $temp`;
@@ -840,5 +892,264 @@ if ($rv) {
 return undef;
 }
 
+# list_indexes(db)
+# Returns the names of all indexes in some database
+sub list_indexes
+{
+local ($db) = @_;
+local (@rv, $r);
+local %tables = map { $_, 1 } &list_tables($db);
+if (&supports_schemas($db)) {
+	local $t = &execute_sql_safe($db, "select schemaname,indexname,tablename from pg_indexes");
+	return map { ($_->[0] eq "public" ? "" : $_->[0].".").$_->[1] }
+		grep { $tables{($_->[0] eq "public" ? "" : $_->[0].".").$_->[2]} }
+		   @{$t->{'data'}};
+	}
+else {
+	local $t = &execute_sql_safe($db, "select indexname,tablename from pg_indexes");
+	return map { $_->[0] } grep { $tables{$t->[1]} } @{$t->{'data'}};
+	}
+}
+
+# index_structure(db, indexname)
+# Returns information on an index
+sub index_structure
+{
+local ($db, $index) = @_;
+local $info = { 'name' => $index };
+if (&supports_schemas($db)) {
+	local ($sn, $in);
+	if ($index =~ /^(\S+)\.(\S+)$/) {
+		($sn, $in) = ($1, $2);
+		}
+	else {
+		($sn, $in) = ("public", $index);
+		}
+	local $t = &execute_sql_safe($db, "select schemaname,tablename,indexdef from pg_indexes where indexname = '$in' and schemaname = '$sn'");
+	local $r = $t->{'data'}->[0];
+	if ($r->[0] eq "public") {
+		$info->{'table'} = $r->[1];
+		}
+	else {
+		$info->{'table'} = $r->[0].".".$r->[1];
+		}
+	$info->{'create'} = $r->[2];
+	}
+else {
+	local $t = &execute_sql_safe($db, "select tablename,indexdef from pg_indexes where indexname = '$index'");
+	local $r = $t->{'data'}->[0];
+	$info->{'table'} = $r->[0];
+	$info->{'create'} = $r->[1];
+	}
+
+# Parse create expression
+if ($info->{'create'} =~ /^create\s+unique/i) {
+	$info->{'type'} = 'unique';
+	}
+if ($info->{'create'} =~ /using\s+(\S+)\s/) {
+	$info->{'using'} = lc($1);
+	}
+if ($info->{'create'} =~ /\((.*)\)/) {
+	$info->{'cols'} = [ split(/\s*,\s*/, $1) ];
+	}
+
+return $info;
+}
+
+sub supports_indexes
+{
+return &get_postgresql_version() >= 7.3;
+}
+
+# list_views(db)
+# Returns the names of all views in some database
+sub list_views
+{
+local ($db) = @_;
+local (@rv, $r);
+if (&supports_schemas($db)) {
+	local $t = &execute_sql_safe($db, "select schemaname,viewname from pg_views where schemaname != 'pg_catalog' and schemaname != 'information_schema'");
+	return map { ($_->[0] eq "public" ? "" : $_->[0].".").$_->[1] }
+		   @{$t->{'data'}};
+	}
+else {
+	local $t = &execute_sql_safe($db, "select viewname from pg_indexes");
+	return map { $_->[0] } @{$t->{'data'}};
+	}
+}
+
+# view_structure(db, viewname)
+# Returns information about a view
+sub view_structure
+{
+local ($db, $view) = @_;
+local $info = { 'name' => $view };
+if (&supports_schemas($db)) {
+	local ($sn, $in);
+	if ($view =~ /^(\S+)\.(\S+)$/) {
+		($sn, $in) = ($1, $2);
+		}
+	else {
+		($sn, $in) = ("public", $view);
+		}
+	local $t = &execute_sql_safe($db, "select schemaname,viewname,definition from pg_views where viewname = '$in' and schemaname = '$sn'");
+	local $r = $t->{'data'}->[0];
+	$info->{'query'} = $r->[2];
+	}
+else {
+	local $t = &execute_sql_safe($db, "select viewname,definition from pg_views where viewname = '$index'");
+	local $r = $t->{'data'}->[0];
+	$info->{'query'} = $r->[1];
+	}
+
+$info->{'query'} =~ s/;$//;
+
+return $info;
+}
+
+sub supports_views
+{
+return &get_postgresql_version() >= 7.3;
+}
+
+# list_sequences(db)
+# Returns the names of all sequences in some database
+sub list_sequences
+{
+local ($db) = @_;
+local (@rv, $r);
+if (&supports_schemas($db)) {
+	local $t = &execute_sql_safe($db, "select schemaname,relname from pg_statio_user_sequences");
+	return map { ($_->[0] eq "public" ? "" : $_->[0].".").$_->[1] }
+		   @{$t->{'data'}};
+	}
+else {
+	local $t = &execute_sql_safe($db, "select relname from pg_statio_user_sequences");
+	return map { $_->[0] } @{$t->{'data'}};
+	}
+}
+
+# sequence_structure(db, name)
+# Returns details of a sequence
+sub sequence_structure
+{
+local ($db, $seq) = @_;
+local $info = { 'name' => $seq };
+
+local $t = &execute_sql_safe($db, "select * from ".&quote_table($seq));
+local $r = $t->{'data'}->[0];
+local $i = 0;
+foreach my $c (@{$t->{'titles'}}) {
+	$info->{$c} = $r->[$i++];
+	}
+
+return $info;
+}
+
+sub supports_sequences
+{
+return &get_postgresql_version() >= 7.4 ? 1 :
+       &get_postgresql_version() >= 7.3 ? 2 : 0;
+}
+
+# Returns 1 if the postgresql server being managed is on this system
+sub is_postgresql_local
+{
+return $config{'host'} eq '' || $config{'host'} eq 'localhost' ||
+       $config{'host'} eq &get_system_hostname() ||
+       &to_ipaddress($config{'host'}) eq &to_ipaddress(&get_system_hostname());
+}
+
+# backup_database(database, dest-path, format, [&only-tables])
+# Executes the pg_dump command to backup the specified database to the
+# given destination path. Returns undef on success, or an error message
+# on failure.
+sub backup_database
+{
+local ($db, $path, $format, $tables) = @_;
+local $tablesarg = join(" ", map { " -t ".quotemeta($_) } @$tables);
+local $cmd = $config{'dump_cmd'}.
+	     (!$postgres_login ? "" :
+	      &supports_pgpass() ? " -U $postgres_login" : " -u").
+	     ($config{'host'} ? " -h $config{'host'}" : "").
+	     ($format eq 'p' ? "" : " -b").
+	     $tablesarg.
+	     " -F$format -f ".quotemeta($path)." $db";
+if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
+	$cmd = &command_as_user($postgres_login, 0, $cmd);
+	}
+$cmd = &command_with_login($cmd);
+local $out = &backquote_logged("$cmd 2>&1");
+if ($? || $out =~ /could not|error|failed/i) {
+	return $out;
+	}
+return undef;
+}
+
+# restore_database(database, source-path, only-data, clear-db)
+# Restores the contents of a PostgreSQL backup into the specified database.
+# Returns undef on success, or an error message on failure.
+sub restore_database
+{
+local ($db, $path, $only, $clean) = @_;
+local $cmd = $config{'rstr_cmd'}.
+	     (!$postgres_login ? "" :
+	      &supports_pgpass() ? " -U $postgres_login" : " -u").
+	     ($config{'host'} ? " -h $config{'host'}" : "").
+	     ($only ? " -a" : "").
+	     ($clean ? " -c" : "").
+	     " -d $in{'db'} ".quotemeta($path);
+if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
+	$cmd = &command_as_user($postgres_login, 0, $cmd);
+	}
+$cmd = &command_with_login($cmd);
+local $out = &backquote_logged("$cmd 2>&1");
+if ($? || $out =~ /could not|error|failed/i) {
+	return $out;
+	}
+return undef;
+}
+
+sub supports_pgpass
+{
+return &get_postgresql_version(1) >= 7.3;
+}
+
+# command_with_login(command, [user, pass])
+# Given a command that talks to postgresql (like psql or pg_dump), sets up
+# the environment so that it can login to the database. Returns a modified
+# command to execute.
+sub command_with_login
+{
+local ($cmd, $user, $pass) = @_;
+if (!defined($user)) {
+	$user = $postgres_login;
+	$pass = $postgres_pass;
+	}
+local $loginfile;
+if (&supports_pgpass()) {
+	# Can use .pgpass file
+	local $temphome = &transname();
+	&make_dir($temphome, 0755);
+	local $pgpass = "$temphome/.pgpass";
+	open(PGPASS, ">$pgpass");
+	print PGPASS "*:*:*:$user:$pass\n";
+	close(PGPASS);
+	$ENV{'HOME'} = $temphome;
+	&set_ownership_permissions(
+		$postgres_sameunix ? $user : undef,
+		undef, 0600, $pgpass);
+	}
+else {
+	# Need to put login and password in temp file
+	$loginfile = &transname();
+	open(TEMP, ">$loginfile");
+	print TEMP "$user\n$pass\n";
+	close(TEMP);
+	$cmd .= " <$loginfile";
+	}
+return $cmd;
+}
+
 1;
 
