@@ -1,6 +1,13 @@
 # linux-lib.pl
 # Quota functions for all linux version
 
+# Tell the mount module not to check which filesystems are supported,
+# as we don't care for the calls made by this module
+$mount::no_check_support = 1;
+
+# Pass UIDs and GIDs to edquota instead of names
+$edquota_use_ids = 1;
+
 # quotas_init()
 sub quotas_init
 {
@@ -154,6 +161,9 @@ if ($out =~ /\s(\d+\.\d+)/) {
 	$version = $1;
 	}
 
+# Force load of quota kernel modules
+&system_logged("modprobe quota_v2 >/dev/null 2>&1");
+
 if ($_[1]%2 == 1) {
 	# turn on user quotas
 	local $qf = $version >= 2 ? "aquota.user" : "quota.user";
@@ -320,6 +330,7 @@ sub parse_repquota_output
 {
 local($rep, @rep, $n, $what, $u, @uinfo);
 $what = $_[1];
+$$what = ( );
 $rep = `$_[0] 2>&1`;
 if ($?) { return -1; }
 local $st = &supports_status($_[0], $what);
@@ -378,10 +389,10 @@ return $nn;
 # edit_quota_file(data, filesys, sblocks, hblocks, sfiles, hfiles)
 sub edit_quota_file
 {
-local($rv, $line, %mtab, @m, @line, $i);
+local($rv, $line, %mtab, @m, @line);
 %mtab = &get_mtab_map();
 @line = split(/\n/, $_[0]);
-for($i=0; $i<@line; $i++) {
+for(my $i=0; $i<@line; $i++) {
 	if ($line[$i] =~ /^(\S+): blocks in use: (\d+), limits \(soft = (\d+), hard = (\d+)\)$/ && $mtab{&resolve_links("$1")} eq $_[1]) {
 		# Found old-style lines to change
 		$rv .= "$1: blocks in use: $2, limits (soft = $_[2], hard = $_[3])\n";
@@ -412,23 +423,17 @@ return $rv;
 # Runs quotacheck on some filesystem
 sub quotacheck
 {
-local $out = &backquote_logged("$config{'quotacheck_command'} $_[0] 2>&1");
+local $out;
+local $cmd = $config{'quotacheck_command'};
+$cmd =~ s/\s+-[ug]//g;
+local $flag = $_[1] ? "-u" : "-g";
+$out = &backquote_logged("$cmd $flag $_[0] 2>&1");
 if ($?) {
 	# Try with the -f and -m options
-	if ($_[1] == 1) {
-		$out = &backquote_logged("$config{'quotacheck_command'} -u -f -m $_[0] 2>&1");
-		}
-	else {
-		$out = &backquote_logged("$config{'quotacheck_command'} -g -f -m $_[0] 2>&1");
-		}
+	$out = &backquote_logged("$cmd $flag -f -m $_[0] 2>&1");
 	if ($?) {
 		# Try with the -F option
-		if ($_[1] == 1) {
-			$out = &backquote_logged("$config{'quotacheck_command'} -u -F $_[0] 2>&1");
-			}
-		else {
-			$out = &backquote_logged("$config{'quotacheck_command'} -g -F $_[0] 2>&1");
-			}
+		$out = &backquote_logged("$config{'quotacheck_command'} $flag -F $_[0] 2>&1");
 		}
 	return $out if ($?);
 	}
@@ -509,19 +514,19 @@ return @rv;
 # edit_grace_file(data, filesystem, btime, bunits, ftime, funits)
 sub edit_grace_file
 {
-local($rv, $line, @m, %mtab, @line, $i);
+local($rv, $line, @m, %mtab, @line);
 %mtab = &get_mtab_map();
 @line = split(/\n/, $_[0]);
-for($i=0; $i<@line; $i++) {
+for(my $i=0; $i<@line; $i++) {
 	$line = $line[$i];
+	print STDERR "line = $line\n";
 	if ($line =~ /^(\S+): block grace period: (\d+) (\S+), file grace period: (\d+) (\S+)/ && $mtab{&resolve_links("$1")} eq $_[1]) {
 		# replace this line
 		$line = "$1: block grace period: $_[2] $unit_to_name{$_[3]}, file grace period: $_[4] $unit_to_name{$_[5]}";
 		}
 	elsif ($line =~ /^\s+(\S+)\s+(\d+)(\S+)\s+(\d+)(\S+)/ && $mtab{&resolve_links("$1")} eq $_[1]) {
 		# replace new-style line
-		$line = "  $1  $_[2]$unit_to_name{$_[3]} $_[4]$unit_to_name{$_[5
-]}";
+		$line = "  $1 $_[2]$unit_to_name{$_[3]} $_[4]$unit_to_name{$_[5]}";
 		}
 	elsif ($line =~ /^device\s+(\S+)\s+\((\S+)\):/i && $2 eq $_[1]) {
 		# replace even newer-style line
@@ -529,6 +534,7 @@ for($i=0; $i<@line; $i++) {
 		$line = "Block grace: $_[2] $unit_to_name{$_[3]} Inode grace: $_[4] $unit_to_name{$_[5]}";
 		$i++;
 		}
+	print STDERR "new = $line\n";
 	$rv .= "$line\n";
 	}
 return $rv;
