@@ -170,11 +170,25 @@ while(<AUTHS>) {
 	s/\r|\n//g;
 	s/#.*$//;
 	if (/^((.*)\s+)?(\d+)\s+(\d+)\s+(\d+)\s+(\S+)$/) {
+		# SSH1 style line
 		local $auth = { 'name' => $6,
 				'key' => $5,
 				'exp' => $4,
 				'bits' => $3,
 				'opts' => $2,
+				'type' => 1,
+				'line' => $lnum,
+				'index' => scalar(@rv) };
+		$auth->{'opts'} =~ s/\s+$//;
+		push(@rv, $auth);
+		}
+	elsif (/^((.*)\s+)?([a-z]\S+)\s+(\S+)\s+(\S+)/) {
+		# SSH2 style line
+		local $auth = { 'name' => $5,
+				'key' => $4,
+				'opts' => $2,
+				'keytype' => $3,
+				'type' => 2,
 				'line' => $lnum,
 				'index' => scalar(@rv) };
 		$auth->{'opts'} =~ s/\s+$//;
@@ -241,8 +255,16 @@ return join(",", @rv);
 
 sub auth_line
 {
-return join(" ", $_[0]->{'opts'} ? ( $_[0]->{'opts'} ) : ( ),
-		 $_[0]->{'bits'}, $_[0]->{'exp'}, $_[0]->{'key'}, $_[0]->{'name'} );
+if ($_[0]->{'type'} == 2) {
+	# SSH 2 format line
+	return join(" ", $_[0]->{'opts'} ? ( $_[0]->{'opts'} ) : ( ),
+			 $_[0]->{'keytype'}, $_[0]->{'key'}, $_[0]->{'name'} );
+	}
+else {
+	# SSH 1 format line
+	return join(" ", $_[0]->{'opts'} ? ( $_[0]->{'opts'} ) : ( ),
+			 $_[0]->{'bits'}, $_[0]->{'exp'}, $_[0]->{'key'}, $_[0]->{'name'} );
+	}
 }
 
 # list_knowns()
@@ -306,7 +328,36 @@ return join(" ", join(",", @{$_[0]->{'hosts'}}), $_[0]->{'bits'}, $_[0]->{'exp'}
 sub get_ssh_type
 {
 local $out = `ssh -V 2>&1 </dev/null`;
-return $out =~ /openssh/i ? 'openssh' : 'ssh';
+return $out =~ /openssh/i ? 'openssh' :
+       $out =~ /Sun_SSH/i ? 'openssh' :
+			    'ssh';
+}
+
+# get_ssh_version()
+# Returns the SSH version number
+sub get_ssh_version
+{
+local $out = `ssh -V 2>&1 </dev/null`;
+if ($out =~ /(sshd\s+version\s+([0-9\.]+))/i ||
+    $out =~ /(ssh\s+secure\s+shell\s+([0-9\.]+))/i) {
+	# Classic commercial SSH
+	return $2;
+	}
+elsif ($out =~ /(OpenSSH.([0-9\.]+))/i) {
+	# OpenSSH .. assume all versions are supported
+	return $2;
+	}
+elsif ($out =~ /(Sun_SSH_([0-9\.]+))/i) {
+	# Solaris 9 SSH is actually OpenSSH 2.x
+	return 2.0;
+	}
+elsif (($out = $config{'sshd_version'}) && ($out =~ /(Sun_SSH_([0-9\.]+))/i)) {
+	# Probably Solaris 10 SSHD that didn't display version.  Use it.
+	return 2.0;
+	}
+else {
+	return undef;
+	}
 }
 
 1;
