@@ -66,6 +66,79 @@ $rv .= "</tr>\n";
 return $rv;
 }
 
+# ui_columns_header(&columns, &tdtags)
+# Returns HTML for a row in a multi-column table, with a header background
+sub ui_columns_header
+{
+return &theme_ui_columns_header(@_) if (defined(&theme_ui_columns_header));
+local ($cols, $tdtags) = @_;
+local $rv;
+$rv .= "<tr $tb>\n";
+local $i;
+for($i=0; $i<@$cols; $i++) {
+	$rv .= "<td ".$tdtags->[$i]."><b>".
+	       ($cols->[$i] eq "" ? "<br>" : $cols->[$i])."</b></td>\n";
+	}
+$rv .= "</tr>\n";
+return $rv;
+}
+
+# ui_checked_columns_row(&columns, &tdtags, checkname, checkvalue, [checked?])
+# Returns HTML for a row in a multi-column table, in which the first
+# column is a checkbox
+sub ui_checked_columns_row
+{
+return &theme_ui_checked_columns_row(@_) if (defined(&theme_ui_checked_columns_row));
+local ($cols, $tdtags, $checkname, $checkvalue, $checked) = @_;
+local $rv;
+$rv .= "<tr $cb>\n";
+$rv .= "<td ".$tdtags->[0].">".
+       &ui_checkbox($checkname, $checkvalue, undef, $checked)."</td>\n";
+local $i;
+for($i=0; $i<@$cols; $i++) {
+	$rv .= "<td ".$tdtags->[$i+1].">";
+	if ($cols->[$i] !~ /<a\s+href|<input|<select|<textarea/) {
+		$rv .= "<label for=\"".
+			&quote_escape("${checkname}_${checkvalue}")."\">";
+		}
+	$rv .= ($cols->[$i] eq "" ? "<br>" : $cols->[$i]);
+	if ($cols->[$i] !~ /<a\s+href|<input|<select|<textarea/) {
+		$rv .= "</label>";
+		}
+	$rv .= "</td>\n";
+	}
+$rv .= "</tr>\n";
+return $rv;
+}
+
+# ui_radio_columns_row(&columns, &tdtags, checkname, checkvalue)
+# Returns HTML for a row in a multi-column table, in which the first
+# column is a radio button
+sub ui_radio_columns_row
+{
+return &theme_ui_radio_columns_row(@_) if (defined(&theme_ui_radio_columns_row));
+local ($cols, $tdtags, $checkname, $checkvalue) = @_;
+local $rv;
+$rv .= "<tr $cb>\n";
+$rv .= "<td ".$tdtags->[0].">".
+       &ui_oneradio($checkname, $checkvalue, "", 0)."</td>\n";
+local $i;
+for($i=0; $i<@$cols; $i++) {
+	$rv .= "<td ".$tdtags->[$i+1].">";
+	if ($cols->[$i] !~ /<a\s+href|<input|<select|<textarea/) {
+		$rv .= "<label for=\"".
+			&quote_escape("${checkname}_${checkvalue}")."\">";
+		}
+	$rv .= ($cols->[$i] eq "" ? "<br>" : $cols->[$i]);
+	if ($cols->[$i] !~ /<a\s+href|<input|<select|<textarea/) {
+		$rv .= "</label>";
+		}
+	$rv .= "</td>\n";
+	}
+$rv .= "</tr>\n";
+return $rv;
+}
+
 # ui_columns_end()
 # Returns HTML to end a table started by ui_columns_start
 sub ui_columns_end
@@ -76,18 +149,20 @@ return "</table>\n";
 
 ####################### form generation functions
 
-# ui_form_start(script, method)
+# ui_form_start(script, method, [target])
 # Returns HTML for a form that submits to some script
 sub ui_form_start
 {
 return &theme_ui_form_start(@_) if (defined(&theme_ui_form_start));
-local ($script, $method) = @_;
+local ($script, $method, $target) = @_;
 local $rv;
 $rv .= "<form action='$script' ".
-		($method eq "post" ? "method=post" :
-		 $method eq "form-data" ?
-			"method=post enctype=multipart/form-data" :
-			"method=get").">\n";
+	($method eq "post" ? "method=post" :
+	 $method eq "form-data" ?
+		"method=post enctype=multipart/form-data" :
+		"method=get").
+	($target ? " target=$target" : "").
+       ">\n";
 return $rv;
 }
 
@@ -107,7 +182,7 @@ if ($buttons && @$buttons) {
 				      $b eq $buttons->[0] ? " align=left" :
 				      $b eq $buttons->[@$buttons-1] ?
 					" align=right" : " align=center").">".
-			       &ui_submit($b->[1], $b->[0]).
+			       &ui_submit($b->[1], $b->[0], $b->[3]).
 			       ($b->[2] ? " ".$b->[2] : "")."</td>\n";
 			}
 		elsif ($b) {
@@ -123,15 +198,17 @@ $rv .= "</form>\n";
 return $rv;
 }
 
-# ui_textbox(name, value, size, [disabled?])
+# ui_textbox(name, value, size, [disabled?], [maxlength])
 # Returns HTML for a text input
 sub ui_textbox
 {
 return &theme_ui_textbox(@_) if (defined(&theme_ui_textbox));
-local ($name, $value, $size, $dis) = @_;
+local ($name, $value, $size, $dis, $max) = @_;
 return "<input name=\"".&quote_escape($name)."\" ".
        "value=\"".&quote_escape($value)."\" ".
-       "size=$size ".($dis ? "disabled=true" : "").">";
+       "size=$size ".($dis ? "disabled=true" : "").
+       ($max ? " maxlength=$max" : "").
+       ">";
 }
 
 # ui_bytesbox(name, bytes, [size], [disabled?])
@@ -234,17 +311,54 @@ local $rv;
 local $o;
 foreach $o (@$opts) {
 	local $id = &quote_escape($name."_".$o->[0]);
+	local $label = $o->[1] || $o->[0];
+	local $after;
+	if ($label =~ /^([^<]*)(<[\000-\377]*)$/) {
+		$label = $1;
+		$after = $2;
+		}
 	$rv .= "<input type=radio name=\"".&quote_escape($name)."\" ".
                "value=\"".&quote_escape($o->[0])."\"".
 	       ($o->[0] eq $value ? " checked" : "").
 	       ($dis ? " disabled=true" : "").
 	       " id=\"$id\"".
 	       " $o->[2]> <label for=\"$id\">".
-	       ($o->[1] || $o->[0])."</label>\n";
+	       $label."</label>".$after."\n";
 	}
 return $rv;
 }
 
+# ui_radio_table(name, value, &options, [disabled?])
+# Returns HTML for a series of radio buttons, one per line in a table
+sub ui_radio_table
+{
+return &theme_ui_radio_table(@_) if (defined(&theme_ui_radio_table));
+local ($name, $value, $opts, $dis) = @_;
+local $rv = "<table>";
+local $o;
+foreach $o (@$opts) {
+	local $id = &quote_escape($name."_".$o->[0]);
+	local $label = $o->[1] || $o->[0];
+	local $after;
+	if ($label =~ /^([^<]*)(<[\000-\377]*)$/) {
+		$label = $1;
+		$after = $2;
+		}
+	$rv .= "<tr> <td>";
+	$rv .= "<input type=radio name=\"".&quote_escape($name)."\" ".
+               "value=\"".&quote_escape($o->[0])."\"".
+	       ($o->[0] eq $value ? " checked" : "").
+	       ($dis ? " disabled=true" : "").
+	       " id=\"$id\"".
+	       " $o->[3]> <label for=\"$id\">".
+	       $label."</label>".$after."</td>\n";
+	$rv .= "<td>$o->[2]</td>\n";
+	$rv .= "</tr>\n";
+	}
+$rv .= "</table>\n";
+return $rv;
+}
+
 # ui_yesno_radio(name, value, [yes], [no], [disabled?])
 # Like ui_yesno, but always displays just two inputs (yes and no)
 sub ui_yesno_radio
@@ -263,12 +377,19 @@ sub ui_checkbox
 {
 return &theme_ui_checkbox(@_) if (defined(&theme_ui_checkbox));
 local ($name, $value, $label, $sel, $tags, $dis) = @_;
-local $id = &quote_escape($name);
+local $after;
+if ($label =~ /^([^<]*)(<[\000-\377]*)$/) {
+	$label = $1;
+	$after = $2;
+	}
 return "<input type=checkbox name=\"".&quote_escape($name)."\" ".
        "value=\"".&quote_escape($value)."\" ".
        ($sel ? " checked" : "").($dis ? " disabled=true" : "").
-       " id=\"$id\"".
-       " $tags> <label for=\"$id\">$label</label>\n";
+       " id=\"".&quote_escape("${name}_${value}")."\"".
+       " $tags> ".
+       ($label eq "" ? $after :
+	 "<label for=\"".&quote_escape("${name}_${value}").
+	 "\">$label</label>$after")."\n";
 }
 
 # ui_oneradio(name, value, label, selected?, [tags], [disabled?])
@@ -277,23 +398,29 @@ sub ui_oneradio
 {
 return &theme_ui_oneradio(@_) if (defined(&theme_ui_oneradio));
 local ($name, $value, $label, $sel, $tags, $dis) = @_;
-local $id = &quote_escape($name);
+local $id = &quote_escape("${name}_${value}");
+local $after;
+if ($label =~ /^([^<]*)(<[\000-\377]*)$/) {
+	$label = $1;
+	$after = $2;
+	}
 return "<input type=radio name=\"".&quote_escape($name)."\" ".
        "value=\"".&quote_escape($value)."\" ".
        ($sel ? " checked" : "").($dis ? " disabled=true" : "").
        " id=\"$id\"".
-       " $tags> <label for=\"$id\">$label</label>\n";
+       " $tags> <label for=\"$id\">$label</label>$after\n";
 }
 
-# ui_textarea(name, value, rows, cols, [wrap], [disabled?])
+# ui_textarea(name, value, rows, cols, [wrap], [disabled?], [tags])
 # Returns HTML for a multi-line text input
 sub ui_textarea
 {
 return &theme_ui_textarea(@_) if (defined(&theme_ui_textarea));
-local ($name, $value, $rows, $cols, $wrap, $dis) = @_;
+local ($name, $value, $rows, $cols, $wrap, $dis, $tags) = @_;
 return "<textarea name=\"".&quote_escape($name)."\" ".
        "rows=$rows cols=$cols".($wrap ? " wrap=$wrap" : "").
-       ($dis ? " disabled=true" : "").">".
+       ($dis ? " disabled=true" : "").
+       ($tags ? " $tags" : "").">".
        &html_escape($value).
        "</textarea>";
 }
@@ -304,7 +431,7 @@ sub ui_user_textbox
 {
 return &theme_ui_user_textbox(@_) if (defined(&theme_ui_user_textbox));
 return &ui_textbox($_[0], $_[1], 13, $_[3])." ".
-       &user_chooser_button($_[0], 1, $_[2]);
+       &user_chooser_button($_[0], 0, $_[2]);
 }
 
 # ui_group_textbox(name, value, [form], [disabled?])
@@ -313,16 +440,16 @@ sub ui_group_textbox
 {
 return &theme_ui_group_textbox(@_) if (defined(&theme_ui_group_textbox));
 return &ui_textbox($_[0], $_[1], 13, $_[3])." ".
-       &group_chooser_button($_[0], 1, $_[2]);
+       &group_chooser_button($_[0], 0, $_[2]);
 }
 
 # ui_opt_textbox(name, value, size, option1, [option2], [disabled?],
-#		 [&extra-fields])
+#		 [&extra-fields], [max])
 # Returns HTML for a text field that is optional
 sub ui_opt_textbox
 {
 return &theme_ui_opt_textbox(@_) if (defined(&theme_ui_opt_textbox));
-local ($name, $value, $size, $opt1, $opt2, $dis, $extra) = @_;
+local ($name, $value, $size, $opt1, $opt2, $dis, $extra, $max) = @_;
 local $dis1 = &js_disable_inputs([ $name, @$extra ], [ ]);
 local $dis2 = &js_disable_inputs([ ], [ $name, @$extra ]);
 local $rv;
@@ -334,7 +461,8 @@ $rv .= "<input type=radio name=\"".&quote_escape($name."_def")."\" ".
        ($dis ? " disabled=true" : "")." onClick='$dis2'> ".$opt2."\n";
 $rv .= "<input name=\"".&quote_escape($name)."\" ".
        "size=$size value=\"".&quote_escape($value)."\" ".
-       ($value eq "" || $dis ? "disabled=true" : "").">\n";
+       ($value eq "" || $dis ? "disabled=true" : "").
+       ($max ? " maxlength=$max" : "").">\n";
 return $rv;
 }
 
@@ -387,11 +515,20 @@ return &theme_ui_table_row(@_) if (defined(&theme_ui_table_row));
 local ($label, $value, $cols) = @_;
 $cols ||= 1;
 local $rv;
+if ($main::ui_table_pos+$cols+1 > $main::ui_table_cols) {
+	# If the requested number of cols won't fit in the number
+	# remaining, start a new row
+	$rv .= "</tr>\n";
+	$main::ui_table_pos = 0;
+	}
 $rv .= "<tr>\n" if ($main::ui_table_pos%$main::ui_table_cols == 0);
 $rv .= "<td valign=top $_[3]->[0]><b>$label</b></td>\n" if (defined($label));
 $rv .= "<td valign=top colspan=$cols $_[3]->[1]>$value</td>\n";
 $main::ui_table_pos += $cols+(defined($label) ? 1 : 0);
-$rv .= "</tr>\n" if ($main::ui_table_pos%$main::ui_table_cols == 0);
+if ($main::ui_table_pos%$main::ui_table_cols == 0) {
+	$rv .= "</tr>\n";
+	$main::ui_table_pos = 0;
+	}
 return $rv;
 }
 
@@ -399,8 +536,28 @@ return $rv;
 sub ui_table_hr
 {
 return &theme_ui_table_hr(@_) if (defined(&theme_ui_table_hr));
-$ui_table_pos = 0;
-return "<tr> <td colspan=$main::ui_table_cols><hr></td> </tr>\n";
+local $rv;
+if ($ui_table_pos) {
+	$rv .= "</tr>\n";
+	$ui_table_pos = 0;
+	}
+$rv .= "<tr> <td colspan=$main::ui_table_cols><hr></td> </tr>\n";
+return $rv;
+}
+
+# ui_table_span(text)
+# Outputs a table row that spans the whole table, and contains the given text
+sub ui_table_span
+{
+local ($text) = @_;
+return &theme_ui_table_hr(@_) if (defined(&theme_ui_table_hr));
+local $rv;
+if ($ui_table_pos) {
+	$rv .= "</tr>\n";
+	$ui_table_pos = 0;
+	}
+$rv .= "<tr> <td colspan=$main::ui_table_cols>$text</td> </tr>\n";
+return $rv;
 }
 
 # ui_buttons_start()
@@ -425,17 +582,23 @@ return &theme_ui_buttons_row(@_) if (defined(&theme_ui_buttons_row));
 local ($script, $label, $desc, $hiddens, $after, $before) = @_;
 return "<form action=$script>\n".
        $hiddens.
-       "<tr> <td nowrap width=20%>".($before ? $before." " : "").
+       "<tr> <td nowrap width=20% valign=top>".($before ? $before." " : "").
        &ui_submit($label).($after ? " ".$after : "")."</td>\n".
-       "<td valign=top width=80%>$desc</td> </tr>\n".
+       "<td valign=top width=80% valign=top>$desc</td> </tr>\n".
        "</form>\n";
 }
 
-# ui_buttons_hr()
+# ui_buttons_hr([title])
 sub ui_buttons_hr
 {
+local ($title) = @_;
 return &theme_ui_buttons_hr(@_) if (defined(&theme_ui_buttons_hr));
-return "<tr> <td colspan=2><hr></td> </tr>\n";
+if ($title) {
+	return "<tr> <td colspan=2><table cellpadding=0 cellspacing=0 width=100%><tr> <td width=50%><hr></td> <td nowrap>$title</td> <td width=50%><hr></td> </tr></table></td> </tr>\n";
+	}
+else {
+	return "<tr> <td colspan=2><hr></td> </tr>\n";
+	}
 }
 
 ####################### header and footer functions
@@ -448,7 +611,9 @@ return &theme_ui_post_header(@_) if (defined(&theme_ui_post_header));
 local ($text) = @_;
 local $rv;
 $rv .= "<center><font size=+1>$text</font></center>\n" if (defined($text));
-$rv .= "<hr>\n";
+if (!$tconfig{'nohr'} && !$tconfig{'notophr'}) {
+	$rv .= "<hr>\n";
+	}
 return $rv;
 }
 
@@ -457,7 +622,11 @@ return $rv;
 sub ui_pre_footer
 {
 return &theme_ui_pre_footer(@_) if (defined(&theme_ui_pre_footer));
-return "<hr>\n";
+local $rv;
+if (!$tconfig{'nohr'} && !$tconfig{'nobottomhr'}) {
+	$rv .= "<hr>\n";
+	}
+return $rv;
 }
 
 # ui_print_header(subtext, args...)
@@ -536,10 +705,10 @@ sub js_disable_inputs
 local $rv;
 local $f;
 foreach $f (@{$_[0]}) {
-	$rv .= "form.$f.disabled = true; ";
+	$rv .= "form.elements[\"$f\"].disabled = true; ";
 	}
 foreach $f (@{$_[1]}) {
-	$rv .= "form.$f.disabled = false; ";
+	$rv .= "form.elements[\"$f\"].disabled = false; ";
 	}
 return $_[2] ? "$_[2]='$rv'" : $rv;
 }
@@ -550,10 +719,10 @@ sub js_checkbox_disable
 local $rv;
 local $f;
 foreach $f (@{$_[1]}) {
-	$rv .= "form.$f.disabled = $_[0].checked; ";
+	$rv .= "form.elements[\"$f\"].disabled = $_[0].checked; ";
 	}
 foreach $f (@{$_[2]}) {
-	$rv .= "form.$f.disabled = !$_[0].checked; ";
+	$rv .= "form.elements[\"$f\"].disabled = !$_[0].checked; ";
 	}
 return $_[3] ? "$_[3]='$rv'" : $rv;
 }
