@@ -1,45 +1,52 @@
 # linux-lib.pl
 # Mount table functions for linux
 
-if (&has_command("amd")) {
-	local $amd = &read_amd_conf();
-	$amd_support = $amd =~ /\[\s*global\s*\]/i ? 2 : 1;
+if (!$no_check_support) {
+	if (&has_command("amd")) {
+		local $amd = &read_amd_conf();
+		$amd_support = $amd =~ /\[\s*global\s*\]/i ? 2 : 1;
+		}
+	$autofs_support = &has_command("automount");
+	if (&has_command("mount.cifs")) {
+		$cifs_support = 4;
+		}
+	if (&has_command("mount.smbfs")) {
+		$smbfs_support = &backquote_command("mount.smbfs -v", 1) =~ /username=/i ? 4 : 3;
+		$smbfs_fs = "smbfs";
+		}
+	elsif (&has_command("mount.smb")) {
+		$smbfs_support = &backquote_command("mount.smb -v", 1) =~ /username=/i ? 4 : 3;
+		$smbfs_fs = "smb";
+		}
+	elsif (&has_command("smbmount")) {
+		$smbfs_support = &backquote_command("smbmount -v", 1) =~ /Version\s+2/i ? 2 : 1;
+		$smbfs_fs = "smbfs";
+		}
+	$swaps_support = -r "/proc/swaps";
+	if (`uname -r` =~ /^(\d+\.\d+)/ && $1 >= 2.4) {
+		$tmpfs_support = 1;
+		$ext3_support = 1;
+		$no_mount_check = 1;
+		$bind_support = 1;	# XXX which version?
+		}
+	if (&has_command("mkfs.xfs")) {
+		$xfs_support = 1;
+		}
+	if (&has_command("mkfs.jfs")) {
+		$jfs_support = 1;
+		}
 	}
-$autofs_support = &has_command("automount");
-if (&has_command("mount.cifs")) {
-	$cifs_support = 4;
-	}
-if (&has_command("mount.smbfs")) {
-	$smbfs_support = &backquote_command("mount.smbfs -v", 1) =~ /username=/i ? 4 : 3;
-	$smbfs_fs = "smbfs";
-	}
-elsif (&has_command("mount.smb")) {
-	$smbfs_support = &backquote_command("mount.smb -v", 1) =~ /username=/i ? 4 : 3;
-	$smbfs_fs = "smb";
-	}
-elsif (&has_command("smbmount")) {
-	$smbfs_support = &backquote_command("smbmount -v", 1) =~ /Version\s+2/i ? 2 : 1;
-	$smbfs_fs = "smbfs";
-	}
-$swaps_support = -r "/proc/swaps";
+
+# We always need to check this, to fix up LABEL= mounts
 if (&has_command("e2label") &&
     &backquote_command("mount --version 2>&1", 1) =~ /(\d+)\.(\d+)/ && ($1 > 2 || $1==2 && $2 >= 10)) {
 	$has_e2label = 1;
 	}
-if (`uname -r` =~ /^(\d+\.\d+)/ && $1 >= 2.4) {
-	$tmpfs_support = 1;
-	$ext3_support = 1;
-	$no_mount_check = 1;
-	$bind_support = 1;	# XXX which version?
-	}
-if (&has_command("mkfs.xfs")) {
-	$xfs_support = 1;
-	}
 if (&has_command("xfs_db")) {
 	$has_xfs_db = 1;
 	}
-if (&has_command("mkfs.jfs")) {
-	$jfs_support = 1;
+if (&has_command("vol_id")) {
+	$has_volid = 1;
 	}
 
 # Return information about a filesystem, in the form:
@@ -469,6 +476,22 @@ while(<MTAB>) {
 				}
 			}
 		}
+	if (!$_[0] && $has_volid && $p[0] =~ /^\/dev\//) {
+		# Check for a UUID on this partition, and if there is one
+		# and the filesystem is in fstab with the label, change
+		# the device.
+		local $out = &backquote_command("vol_id $p[0] 2>&1", 1);
+		if ($out =~ /ID_FS_UUID=(\S+)/) {
+			local $uuid = $1;
+			foreach $m (@mounts) {
+				if ($m->[0] eq $p[1] &&
+				    $m->[1] eq "UUID=$uuid") {
+					$p[0] = "UUID=$uuid";
+					last;
+					}
+				}
+			}
+		}
 
 	# check fstab for a mount on the same dir which is a symlink
 	# to the device
@@ -503,6 +526,20 @@ while(<SWAPS>) {
 					}
 				}
 			}
+		if (!$_[0] && $has_volid && $sf =~ /^\/dev\//) {
+			# Similarly, fix up UUID= entrys for swap in /etc/fstab
+			local $out = &backquote_command("vol_id $sf 2>&1", 1);
+			if ($out =~ /ID_FS_UUID=(\S+)/) {
+				local $uuid = $1;
+				foreach $m (@mounts) {
+					if ($m->[0] eq "swap" &&
+					    $m->[1] eq "UUID=$uuid") {
+						$sf = "UUID=$uuid";
+						last;
+						}
+					}
+				}
+			}
 		push(@rv, [ "swap", $sf, "swap", "-" ]);
 		}
 	}
@@ -628,10 +665,13 @@ else {
 	# some filesystem supported by mount
 	local $fs = $_[2] eq "*" ? "auto" : $_[2];
 	if ($_[1] =~ /LABEL=(.*)/) {
-		$cmd = "mount -t $fs -L $1 $opts $_[0]";
+		$cmd = "mount -t $fs -L ".quotemeta($1)." $opts ".quotemeta($_[0]);
+		}
+	elsif ($_[1] =~ /UUID=(\S+)/) {
+		$cmd = "mount -t $fs -U ".quotemeta($1)." $opts ".quotemeta($_[0]);
 		}
 	else {
-		$cmd = "mount -t $fs $opts $_[1] $_[0]";
+		$cmd = "mount -t $fs $opts ".quotemeta($_[1])." ".quotemeta($_[0]);
 		}
 	$out = &backquote_logged("$cmd 2>&1 </dev/null");
 	if ($?) { return "<pre>$out</pre>"; }
@@ -661,6 +701,9 @@ else {
 	if ($_[1] =~ /LABEL=(.*)/) {
 		$cmd = "mount -t $fs -L $1 $opts $_[0]";
 		}
+	elsif ($_[1] =~ /UUID=(\S+)/) {
+		$cmd = "mount -t $fs -U $1 $opts $_[0]";
+		}
 	else {
 		$cmd = "mount -t $fs $opts $_[1] $_[0]";
 		}
@@ -671,7 +714,7 @@ else {
 	}
 }
 
-# unmount_dir(directory, device, type)
+# unmount_dir(directory, device, type, options, [force])
 # Unmount a directory that is currently mounted. Returns 0 if ok,
 # or an error string if failed
 sub unmount_dir
@@ -698,7 +741,8 @@ elsif ($_[2] eq "autofs") {
 	sleep(2);
 	}
 else {
-	$out = &backquote_logged("umount $_[0] 2>&1");
+	local $fflag = $_[4] ? "-l" : "";
+	$out = &backquote_logged("umount $fflag $_[0] 2>&1");
 	if ($?) { return "<pre>$out</pre>"; }
 	if ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
 		# remove options from list
@@ -713,6 +757,18 @@ undef(@list_mounted_cache);
 return 0;
 }
 
+# can_force_unmount_dir(directory, device, type)
+# Returns 1 if some directory can be forcibly un-mounted
+sub can_force_unmount_dir
+{
+if ($_[2] ne "swap" && $_[2] ne "auto" && $_[2] ne "autofs") {
+	# All filesystems using the normal 'mount' command can be
+	return 1;
+	}
+else {
+	return 0;
+	}
+}
 
 # mount_modes(type)
 # Given a filesystem type, returns 4 numbers that determine how the file
@@ -763,8 +819,8 @@ sub disk_space
 if (&get_mounted($_[1], "*") < 0) { return (); }
 if ($_[0] eq "proc" || $_[0] eq "swap" ||
     $_[0] eq "auto" || $_[0] eq "autofs") { return (); }
-if (&backquote_command("LC_ALL='' LANG='' df -k $_[1]", 1) =~
-    /Mounted on\n\S+\s+(\S+)\s+\S+\s+(\S+)/) {
+local $out = &backquote_command("LC_ALL='' LANG='' df -k $_[1]", 1);
+if ($out =~ /Mounted on\n\S+\s+(\S+)\s+\S+\s+(\S+)/) {
 	return ($1, $2);
 	}
 return ( );
@@ -790,7 +846,7 @@ sub list_fstypes
 {
 local @sup = ("ext2", "minix", "msdos", "nfs", "nfs4", "iso9660", "ext",
 	      "xiafs", "hpfs", "fat", "vfat", "umsdos", "sysv", "reiserfs",
-	      "ntfs");
+	      "ntfs", "hfs", "fatx");
 push(@sup, $smbfs_fs) if ($smbfs_support);
 push(@sup, "cifs") if ($cifs_support);
 push(@sup, "auto") if ($amd_support);
@@ -823,7 +879,7 @@ local(%fsmap);
 	  "xiafs","Old XIAFS Linux Filesystem",
 	  "hpfs","OS/2 Filesystem",
 	  "fat","MS-DOS Filesystem",
-	  "vfat","Windows 95 Filesystem",
+	  "vfat","Windows Filesystem",
 	  "umsdos","Linux on top of MS-DOS Filesystem",
 	  "sysv","System V Filesystem",
 	  "swap","Virtual Memory",
@@ -838,7 +894,9 @@ local(%fsmap);
 	  "xfs","SGI Filesystem",
 	  "jfs","IBM Journalling Filesystem",
 	  "ntfs","Windows NT Filesystem",
-	  "bind","Loopback Filesystem"
+	  "bind","Loopback Filesystem",
+	  "hfs","Apple Filesystem",
+	  "fatx","XBOX Filesystem",
 	  );
 return $config{long_fstypes} && $fsmap{$_[0]} ? $fsmap{$_[0]} : uc($_[0]);
 }
@@ -888,15 +946,29 @@ elsif ($_[0] eq "swap") {
 	&foreign_require("fdisk", "fdisk-lib.pl");
 	printf "<tr> <td valign=top><b>$text{'linux_swapfile'}</b></td>\n";
 	print "<td colspan=3>\n";
-	local $found;
+	local ($found, $ufound);
+
+	# Show partitions input
 	local $sel = &foreign_call("fdisk", "partition_select", "lnx_disk",
 				   $_[1], 3, \$found);
 	printf "<input type=radio name=lnx_dev value=0 %s> %s %s<br>\n",
 		$found ? "checked" : "", $text{'linux_disk'}, $sel;
+
+	# Show UUID input
+	if ($has_volid) {
+		local $u = $_[1] =~ /UUID=(\S+)/ ? $1 : undef;
+		local $usel = &fdisk::volid_select("lnx_volid", $u, \$ufound);
+		if ($usel) {
+			printf "<input type=radio name=lnx_dev value=5 %s> %s %s<br>\n", $ufound ? "checked" : "", $text{'linux_usel'}, $usel;
+			}
+		}
+
+	# Show other file input
 	printf "<input type=radio name=lnx_dev value=1 %s> %s\n",
-		$found ? "" : "checked", $text{'linux_swapfile'};
+		$found || $ufound ? "" : "checked", $text{'linux_swapfile'};
 	printf "<input name=lnx_other size=35 value='%s'> %s<br>\n",
-		$found ? "" : $_[1], &file_chooser_button("lnx_other");
+		$found || $ufound ? "" : $_[1],
+		&file_chooser_button("lnx_other");
 	print "</td> </tr>\n";
 	}
 elsif ($_[0] eq $smbfs_fs || $_[0] eq "cifs") {
@@ -926,12 +998,15 @@ else {
 	&foreign_require("fdisk", "fdisk-lib.pl");
 	printf "<tr> <td valign=top><b>%s</b></td>\n", &fstype_name($_[0]);
 	print "<td colspan=3>\n";
-	local ($found, $rfound, $lfound, $vfound, $rsel, $c);
+	local ($found, $rfound, $lfound, $vfound, $ufound, $rsel, $c);
+
+	# Show regular partition input
 	local $sel = &foreign_call("fdisk", "partition_select", "lnx_disk",
 				   $_[1], 0, \$found);
 	printf "<input type=radio name=lnx_dev value=0 %s> %s %s<br>\n",
 		$found ? "checked" : "", $text{'linux_disk'}, $sel;
 
+	# Show RAID input
 	if (&foreign_check("raid")) {
 		&foreign_require("raid", "raid-lib.pl");
 		local $conf = &foreign_call("raid", "get_raidtab");
@@ -951,6 +1026,7 @@ else {
 			}
 		}
 
+	# Show LVM input
 	if (&foreign_check("lvm")) {
 		&foreign_require("lvm", "lvm-lib.pl");
 		local @vgs = &foreign_call("lvm", "list_volume_groups");
@@ -962,9 +1038,13 @@ else {
 		if (@lvs) {
 			local $lsel;
 			foreach $l (@lvs) {
+				local $sf = &same_file($_[1], $l->{'device'});
 				$lsel .= sprintf "<option value=%s %s>%s\n",
-					$l->{'device'}, $_[1] eq $l->{'device'} ? 'selected' : '', &text('linux_ldev', $l->{'vg'}, $l->{'name'});
-				$vfound++ if ($_[1] eq $l->{'device'});
+					$l->{'device'},
+					$sf ? 'selected' : '',
+					&text('linux_ldev', $l->{'vg'},
+							    $l->{'name'});
+				$vfound++ if ($sf);
 				}
 			printf "<input type=radio name=lnx_dev value=4 %s> %s\n",
 				$vfound ? "checked" : " ", $text{'linux_lvm'};
@@ -972,20 +1052,31 @@ else {
 			}
 		}
 
+	# Show label input
 	if ($has_e2label || $has_xfs_db) {
 		local $l = $_[1] =~ /LABEL=(.*)/ ? $1 : undef;
-		local $esel = &foreign_call("fdisk", "label_select",
-					    "lnx_label", $l, \$lfound);
+		local $esel = &fdisk::label_select("lnx_label", $l, \$lfound);
 		if ($esel) {
 			printf "<input type=radio name=lnx_dev value=3 %s> %s %s<br>\n", $lfound ? "checked" : "", $text{'linux_lsel'}, $esel;
 			}
 		}
 
+	# Show UUID input
+	if ($has_volid) {
+		local $u = $_[1] =~ /UUID=(\S+)/ ? $1 : undef;
+		local $usel = &fdisk::volid_select("lnx_uuid", $u, \$ufound);
+		if ($usel) {
+			printf "<input type=radio name=lnx_dev value=5 %s> %s %s<br>\n", $ufound ? "checked" : "", $text{'linux_usel'}, $usel;
+			}
+		}
+
+	# Show other device input
+	local $anyfound = $found || $rfound || $lfound || $vfound || $ufound;
 	printf "<input type=radio name=lnx_dev value=1 %s> %s\n",
-		$found || $rfound || $lfound || $vfound ? "" : "checked",
+		$anyfound ? "" : "checked",
 		$text{'linux_other'};
 	printf "<input name=lnx_other size=35 value='%s'> %s<br>\n",
-		$found || $rfound || $lfound || $vfound ? "" : $_[1],
+		$anyfound ? "" : $_[1],
 		&file_chooser_button("lnx_other");
 	print "</td> </tr>\n";
 	}
@@ -1224,7 +1315,8 @@ elsif (($_[0] eq "nfs") || ($_[0] eq "nfs4")) {
 	print "   }\n";
 	print "</script>\n";
 	}
-elsif ($_[0] eq "fat" || $_[0] eq "vfat" || $_[0] eq "msdos" || $_[0] eq "umsdos"){
+elsif ($_[0] eq "fat" || $_[0] eq "vfat" || $_[0] eq "msdos" ||
+       $_[0] eq "umsdos" || $_[0] eq "fatx"){
 	# All dos-based filesystems share some options
         print "<tr $tb> <td colspan=4><b>$text{'edit_dos_opt'}</b></td> </tr>\n";
 	print "<tr> <td><b>$text{'linux_uid'}</b></td>\n";
@@ -1505,6 +1597,17 @@ elsif ($_[0] eq $smbfs_fs || $_[0] eq "cifs") {
 			print "</tr>\n";
 			}
 		}
+
+	if ($_[0] eq "cifs") {
+		# Show cifs-only options
+		print "<tr> <td><b>$text{'linux_codepage'}</b></td>\n";
+		print "<td>",&ui_opt_textbox("smbfs_codepage",
+		    $options{'codepage'}, 10, $text{'default'}),"</td>\n";
+
+		print "<td><b>$text{'linux_iocharset'}</b></td>\n";
+		print "<td>",&ui_opt_textbox("smbfs_iocharset",
+		    $options{'iocharset'}, 10, $text{'default'}),"</td> </tr>\n";
+		}
 	}
 elsif ($_[0] eq "reiserfs") {
 	# Reiserfs is a new super-efficient filesystem
@@ -1699,6 +1802,9 @@ else {
 	elsif ($in{'lnx_dev'} == 4) {
 		$dv = $in{'lnx_lvm'};
 		}
+	elsif ($in{'lnx_dev'} == 5) {
+		$dv = "UUID=".$in{'lnx_uuid'};
+		}
 	else {
 		$dv = $in{'lnx_other'};
 		$dv || &error($text{'linux_edev'});
@@ -1713,7 +1819,7 @@ else {
 	#	}
 
 	# Check if the device actually exists and uses the right filesystem
-	if (!-r $dv && $dv !~ /LABEL=/) {
+	if (!-r $dv && $dv !~ /LABEL=/ && $dv !~ /UUID=/) {
 		if ($_[0] eq "swap" && $dv !~ /\/dev/) {
 			&swap_form($dv);
 			}
@@ -1836,7 +1942,7 @@ elsif ($_[0] eq "ext2" || $_[0] eq "ext3") {
 		{ $options{'usrquota'} = $options{'grpquota'} = ""; }
 	}
 elsif ($_[0] eq "fat" || $_[0] eq "vfat" ||
-       $_[0] eq "msdos" || $_[0] eq "umsdos") {
+       $_[0] eq "msdos" || $_[0] eq "umsdos" || $_[0] eq "fatx") {
 	# All dos-based filesystems have similar options
 	delete($options{"uid"}); delete($options{"gid"});
 	if ($in{fat_uid} ne "") { $options{"uid"} = getpwnam($in{'fat_uid'}); }
@@ -2024,6 +2130,23 @@ elsif ($_[0] eq $smbfs_fs || $_[0] eq "cifs") {
 			if ($in{'smbfs_user2'}) { $options{'user'} = ''; }
 			}
 		}
+
+	if ($_[0] eq "cifs") {
+		# Parse CIFS-specific options
+		delete($options{'codepage'});
+		if (!$in{'smbfs_codepage_def'}) {
+			$in{'smbfs_codepage'} =~ /^\S+$/ ||
+				&error($text{'linux_ecodepage'});
+			$options{'codepage'} = $in{'smbfs_codepage'};
+			}
+
+		delete($options{'iocharset'});
+		if (!$in{'smbfs_iocharset_def'}) {
+			$in{'smbfs_iocharset'} =~ /^\S+$/ ||
+				&error($text{'linux_eiocharset'});
+			$options{'iocharset'} = $in{'smbfs_iocharset'};
+			}
+		}
 	}
 elsif ($_[0] eq "reiserfs") {
 	# Save reiserfs options
@@ -2338,6 +2461,8 @@ return $_[0] =~ /^\/dev\/(s|h)d([a-z])(\d+)$/ ?
 	&text('linux_ldev', "$1", "$2") :
        $_[0] =~ /LABEL=(\S+)/i ?
 	&text('linux_label', "$1") :
+       $_[0] =~ /UUID=(\S+)/i ?
+	&text('linux_uuid', "$1") :
        $_[0] eq '/dev/cdrom' ?
 	$text{'linux_cddev'} :
        $_[0] eq '/dev/burner' ?
